Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq rels ordtype finmap.
Require Import Eqdep prelude domains heaps.
Require Import pcm resources8 proc9 step9 hstep9 sbisim9 wbisim9.
Require Import locksafe9 prog11 progsep11.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(***************************)
(* Lock is a loop over CAS *)
(***************************)

Section LockingExample.
Variable R : resource.
Notation U := (univ_of R).
Notation x := (lock_of R).
Notation Inv := (Inv (R:=R)).
(* Notation shareable := (shareable x). *)
Notation procHeap := (procHeap U).
Notation pre := (@pre R).
Notation post := (@post R).
Notation cont := (@cont R).
Notation spec := (spec (R:=R)).

(* loop invariant is the type of the program *)
Definition lock_s : spec unit := 
  (fun gE lp => hp lp = empty /\ ownsno (mx lp),
   fun _ gE1 gE2 lp1 lp2 =>
     exists inv, 
       lp2 = [mgh true, gh lp1, inv :+ lp1] /\
       Inv (gh lp2 \+ gE2) inv).

(* Locking program is a loop over CAS *)

Program Definition lock : progsep lock_s :=
  ffix (fun (self : unit -> progsep lock_s) (_ : unit) =>
    Do (b <-- cas;
        If b (ret tt) (self tt))) tt.
Next Obligation.
move=>lp gE /= [E O1]; rewrite (mgh_eta lp) E (eqP O1) /=.
apply: bnd_cas=>/= ep1; last first. 
- rewrite -[empty]unh0.
  by apply: val_do=>// x gE' mL' gL' i'; rewrite !unh0.
move=>inv I; apply: val_ret=>ep2 /= O2 _. 
by exists inv; rewrite O2.
Qed.

Lemma val_lock gL gE i (r : cont unit) : 
        (forall gE' inv, 
          def (inv :+ i) -> valid (gL \+ gE') -> Inv (gL \+ gE') inv -> 
          r tt  gE' [mgh true, gL, inv :+ i]) -> 
        verify [mgh false, gL, i] gE lock_s r.
Proof.
move=>H; rewrite -[i]un0h; apply: val_do=>//=.
by case=>gE' mL' gL' i' [inv][[->->->]] I V; rewrite unh0=>D; apply: H.
Qed.

Lemma bnd_lock B s2 gL gE i (r : cont B) : 
        (forall gE' inv, 
          Inv (gL \+ gE') inv -> 
          verify [mgh true, gL, inv :+ i] gE' (s2 tt) r) -> 
        verify [mgh false, gL, i] gE (bind_s lock_s s2) r.
Proof.
move=>H; rewrite -[i]un0h; apply: bnd_do=>//=.
by case=>gE' _ _ _  [inv][[->->->]] I; rewrite unh0; apply: H.
Qed.

End LockingExample.

Implicit Arguments lock [R].


(***********************)
(* Concurrent counters *)
(***********************)

Module NatPCM.
Section Def.

Definition valid := [fun x : nat => true].
Definition join := [fun x1 x2 => x1 + x2].
Definition unit := 0.

Lemma joinC x y : x + y = y + x.
Proof. by rewrite addnC. Qed.

Lemma joinA x y z : x + (y + z) = x + y + z.
Proof. by rewrite addnA. Qed.

Lemma validL x y : valid (x + y) -> valid x.
Proof. by []. Qed.

Lemma unitL x : unit + x = x.
Proof. by rewrite add0n. Qed.

End Def.
End NatPCM.

Definition natPCMMixin := PCMMixin NatPCM.joinC NatPCM.joinA 
                            NatPCM.unitL NatPCM.validL.
Canonical natPCM := Eval hnf in PCM nat natPCMMixin.


Module Counters.
Section Counters.
Variable y : ptr. (* the shared pointer *)

Definition invariant (n : nat) h := h = y :-> n.

Lemma precise_inv : forall n, precise (invariant n).
Proof. by move=>n h h1 h2 Sh1 -> Sh2 ->. Qed.

Definition R := Resource y precise_inv.
Notation U := (univ_of R).
Notation x := (lock_of R).
Notation Inv := (Inv (R:=R)).
Notation procHeap := (procHeap U).
Notation pre := (@pre R).
Notation post := (@post R).
Notation cont := (@cont R).
Notation spec := (spec (R:=R)).

Let incr_s (n : nat) : spec unit := 
  (fun gE (lp : procHeap) => hp lp = empty /\ ownsno (mx lp),
   fun _ gE1 gE2 (lp1 lp2 : procHeap) => 
     lp2 = [mgh false, (gh lp1) \+ n, empty]).

Definition phi (n : nat) := fun k => Some (k + n).

Lemma phi_local (n : nat) : annot_axiom (phi n).
Proof. by move=>/= k _ m [<-] V; rewrite -joinAC. Qed.

Definition phi_annot (n : nat) := Annot (@phi_local n).

Program
Definition incr (n : nat) : progsep (incr_s n) :=
  Do (lock;;
      k <-- read _ y;
      y ::= k + n;;
      unlock (phi_annot n)).
Next Obligation.
case=>mL gL hL gE /= [->] O; rewrite (eqP O) -[empty]unh0.
apply: bnd_lock=>gE1 _ [->]; apply: bnd_read=>gE2 ->//.
apply: bnd_write=>gE3 ->//.
by apply: val_unlock=>//; last by rewrite joinAC.
Qed.

Program
Definition incr2_par (k1 k2 : nat) : progsep (incr_s (k1 \+ k2)) :=
  Do (par (incr k1) (incr k2);; ret tt).
Next Obligation.
case=>mL gL hL gE [/= ->{hL}]; move/eqP=>->{mL}.
rewrite -[empty]unh0.
apply: (bnd_hoare (r1 := fun _ _ =>  
           eq^~ [mgh false, gL \+ k1 \+ k2, empty])); last first.
- move=>_ gE' _ _ _ [->->->]. 
  by apply: val_ret=>//=; rewrite joinA.
rewrite -[empty]unh0 -[gL]unitR -[[mx false]]unitR.
apply: (par_do (fun _ _ => eq^~ [mgh false, gL \+ k1, empty])
               (fun _ _ => eq^~ [mgh false, k2, empty]));
try by rewrite -[empty]unh0; apply: val_do=>//; case=>gE' _ _ _ [->->->].
by move=>?????????? [->->->][->->->]; rewrite !unitR unh0.
Qed.

Program
Definition incr2_seq (k1 k2 : nat) : progsep (incr_s (k1 \+ k2)) :=
  Do (incr k1 ;; incr k2).
Next Obligation.
case=>mL gL hL gE [/= ->{hL}]; move/eqP=>->{mL}; rewrite -[empty]unh0.
apply: bnd_do=>// [][gE1 _ _ _][->->->].
apply: val_do=>// [][gE2 _ _ _][->->->].
by rewrite joinA.
Qed.

(* Now a function incr3 that is parametrized wrt incr, 
so that we can plug different implementations *)

Section Incr3.
Variables (k1 k2 k3 : U).
Variables (incr2 : progsep (incr_s (k1 \+ k2))).

Program
Definition incr3_par : progsep (incr_s (k1 \+ k2 \+ k3)) :=
  Do (par incr2 (incr k3);; ret tt).
Next Obligation.
case=>mL gL hL gE [/= ->{hL}]; move/eqP=>->{mL}.
rewrite -[empty]unh0.
apply: (bnd_hoare (r1 := fun _ _ =>  
           eq^~ [mgh false, gL \+ k1 \+ k2 \+ k3, empty])); last first.
- move=>_ gE' _ _ _ [->->->]. 
  by apply: val_ret=>//=; rewrite !joinA.
rewrite -[empty]unh0 -[gL]unitR -[[mx false]]unitR.
apply: (par_do (fun _ _ => eq^~ [mgh false, gL \+ k1 \+ k2, empty])
               (fun _ _ => eq^~ [mgh false, k3, empty]));
try by rewrite -[empty]unh0; apply: val_do=>//;
       case=>gE' _ _ _ [->->->]; rewrite ?joinA.
by move=>?????????? [->->->][->->->]; rewrite !unitR unh0.
Qed.

End Incr3.

(* End now, plug in the different implementations of incr2 *)
(* the specs are the same, showing that we're summing up k1 + k2 + k3 *)

Definition incr3a k1 k2 k3 := incr3_par k3 (incr2_par k1 k2).
Definition incr3b k1 k2 k3 := incr3_par k3 (incr2_seq k1 k2).

End Counters.
End Counters.

(************)
(* Iterator *)
(************)

Module ListIter.
Section Def.
Variables (R : resource) (A : Type).
Notation U := (univ_of R).
Notation lk := (lock_of R).
Notation Inv := (Inv (R:=R)).
(* Notation shareable := (shareable x). *)
Notation procHeap := (procHeap U).
Notation pre := (@pre R).
Notation post := (@post R).
Notation cont := (@cont R).
Notation spec := (spec (R:=R)).

Fixpoint join_seq (F : A -> U) (xs : seq A) :=
  if xs is x::xs' then F x \+ join_seq F xs' else Unit.

Variable (phi psi : A -> U).

Definition f_s (n : A) : spec unit := 
  (fun gE lp => lp = [mgh false, phi n, empty], 
   fun _ gE1 gE2 lp1 lp2 => lp2 = [mgh false, psi n, empty]).
     
Variable (f : forall n, progsep (f_s n)).

Definition iter_s (ns : seq A) : spec unit := 
  (fun gE lp => lp = [mgh false, join_seq phi ns, empty],
   fun _ gE1 gE2 lp1 lp2 => 
     lp2 = [mgh false, join_seq psi ns, empty]).

Let iterT := forall xs, progsep (iter_s xs).

Program
Definition iter : iterT := 
  ffix (fun (self : iterT) xs =>
    Do (Match_list xs
        (ret tt)
        (fun x' xs' => par (f x') (self xs');; ret tt))). 
Next Obligation.
move=>lp ep /= ->; case: xs=>[|x xs] /=; first by apply: val_ret.
rewrite -[empty]unh0.
apply: (bnd_hoare (r1 := fun _ _ =>  
           eq^~ [mgh false, psi x \+ join_seq psi xs, empty])); last first.
- by move=>_ gE' _ _ _ [->->->]; apply: val_ret.
rewrite -[empty]unh0 -[[mx false]]unitR.
apply: (par_do (fun _ _ => eq^~ [mgh false, psi x, empty])
               (fun _ _ => eq^~ [mgh false, join_seq psi xs, empty])).
- by rewrite -[empty]unh0; apply: val_do=>// _ gE' _ _ _ [->->->].  
- by rewrite -[empty]unh0; apply: val_do=>// _ gE' _ _ _ [->->->]. 
by move=>?????????? [->->->][->->->].
Qed.

End Def.

End ListIter.


(* Applying iter *)

Module ApplyingIter.
Section ApplyingIter.
Variable lk : ptr. (* the lock pointer *)

(* Let's choose the counter's PCM and invariant *)
Definition R := Counters.R lk. 
Notation U := (univ_of R).
Notation spec := (spec (R:=R)).

(* we want to start with 0 *)
Let phi (n : nat) : U := 0.
(* and sum up list elements as they appear 
   (i.e., we don't modify them first) *)
Let psi (n : nat) : U := n.

(* incr is what we want to iterate *)
Program 
Definition incr_by_n (n : nat) : progsep (ListIter.f_s phi psi n) := 
  Do (Counters.incr lk n).
Next Obligation.
case=>/= _ _ _ gE [->->->]; rewrite -[empty]unh0.
apply: val_do=>/=; first by rewrite unh0.
by move=>_ gE' _ _ _ [->->->]; rewrite unh0 unitL.
Qed.

(* let's sum up the numbers 1 - 6 *)
(* we expect 21 *)

Definition sum_s : spec unit := 
  (fun gE lp => lp = [mgh false, 0, empty], 
   fun _ gE1 gE2 lp1 lp2 => lp2 = [mgh false, 21, empty]).

Program Definition sum_up_in_parallel : progsep sum_s := 
  Do (ListIter.iter incr_by_n [:: 0; 1; 2; 3; 4; 5; 6]).
Next Obligation.
case=>_ _ _ gE [->->->]; rewrite -[empty]unh0.
by apply: val_do=>//; case=>? _ _ _ [->->->].
Qed.

End ApplyingIter.
End ApplyingIter.


(*******************************)
(* Coarse-grained set examples *)
(*******************************)

(* finite maps PCM *)

Inductive boolmap (V : ordType) := Undef | Def of {finMap V -> bool}.

Definition inmap (V : ordType) (v : V) := locked (Def (ins v true (@nil V bool))).

Definition outmap (V : ordType) (v : V) := locked (Def (ins v false (@nil V bool))).

Definition dom V (f : boolmap V) : pred V := 
  if f is Def m then fun v => v \in supp m else pred0.

Definition upd V (f : boolmap V) v b := nosimpl
  if f is Def m then Def (ins v b m) else @Undef V.

Module BoolMapPCM.
Section Def.
Variables (V : ordType).

Definition valid (f : boolmap V) := 
  if f is Def _ then true else false.

Definition join (f1 f2 : boolmap V) := 
  if (f1, f2) is (Def m1, Def m2) then 
    if disj m1 m2 then Def (fcat m1 m2)
    else @Undef V
  else @Undef V.

Definition unit := Def (@nil V bool).

Lemma joinC f1 f2 : join f1 f2 = join f2 f1. 
Proof. 
case: f1 f2=>[|m1][|m2] //; rewrite /join.
by case: ifP=>E; rewrite disjC E // fcatC.
Qed.

Lemma joinCA f1 f2 f3 : join f1 (join f2 f3) = join f2 (join f1 f3).
Proof. 
case: f1 f2 f3=>[|m1][|m2][|m3] //.
rewrite /join; case E1: (disj m2 m3); last first.
- by case E2: (disj m1 m3)=>//; rewrite disj_fcat E1 andbF.
rewrite disj_fcat andbC.
case E2: (disj m1 m3)=>//; rewrite disj_fcat (disjC m2) E1 /= andbT.
by case E3: (disj m1 m2)=>//; rewrite fcatAC // E1 E2 E3.
Qed.

Lemma joinA f1 f2 f3 : join f1 (join f2 f3) = join (join f1 f2) f3.
Proof. by rewrite (joinC f2) joinCA joinC. Qed.

Lemma validL f1 f2 : valid (join f1 f2) -> valid f1.
Proof. by case: f1. Qed.

Lemma unitL f : join unit f = f.
Proof. by case: f=>[//|m]; rewrite /join/unit disjC disj_nil fcat0s. Qed.

End Def.
End BoolMapPCM.

Definition boolmapPCMMixin V := 
  PCMMixin (@BoolMapPCM.joinC V) (@BoolMapPCM.joinA V) 
           (@BoolMapPCM.unitL V) (@BoolMapPCM.validL V).
Canonical boolMapPCM V := Eval hnf in PCM (boolmap V) (boolmapPCMMixin V).

Lemma dom_join V (f1 f2 : boolmap V) :
        dom (f1 \+ f2) =i 
        [pred x | valid (f1 \+ f2) && (x \in [predU dom f1 & dom f2])].
Proof.
case: f1 f2=>[|m1] // [|m2] // v; rewrite /dom /PCM.join /=. 
by case: ifP=>// E; rewrite supp_fcat. 
Qed.

Lemma upd_joinL V (f1 f2 : boolmap V) v b : 
        upd (f1 \+ f2) v b = 
        if v \in dom f1 then upd f1 v b \+ f2 else f1 \+ upd f2 v b.
Proof.
case: f1 f2=>[|m1][|m2] //; case: ifP=>H //; 
rewrite /upd /PCM.join /=; case: ifP=>D //.
- rewrite disjC disj_ins disjC D.
  case: disjP D=>//; move/(_ _ H)=>H4 _; rewrite H4.
  by rewrite fcat_inss.
- by rewrite disjC disj_ins disjC D andbF.
- rewrite disj_ins D H /=.
  rewrite (@fcatC _ _ m1) // (@fcatC _ _ m1).
  - by rewrite fcat_inss // H.
  by rewrite disj_ins H D.
by rewrite disj_ins D andbF.
Qed.

CoInductive defUn_spec V (h1 h2 : boolmap V) : bool -> Type :=
| def_false1 of ~~ valid h1 : defUn_spec h1 h2 false
| def_false2 of ~~ valid h2 : defUn_spec h1 h2 false
| def_false3 x of x \in dom h1 & x \in dom h2 : defUn_spec h1 h2 false
| def_true of valid h1 & valid h2 & 
    (forall x, x \in dom h1 -> x \notin dom h2) : defUn_spec h1 h2 true.

Lemma valid_join V (h1 h2 : boolmap V) : defUn_spec h1 h2 (valid (h1 \+ h2)).
Proof.
case: h1 h2=>[|h1][|h2] /=; try by [apply: def_false1 | apply: def_false2].
rewrite /PCM.join /=; case: ifP=>E.
- by apply: def_true=>// x H; case: disjP E=>//; move/( _ _ H). 
by case: disjP E=>// x T1 T2 _; apply: (def_false3 (x:=x)).
Qed.

Lemma dom_inmap (V : ordType) (v : V) : dom (inmap v) =i pred1 v. 
Proof. by move=>x; rewrite /inmap -ssreflect.lock inE. Qed.

Lemma dom_outmap (V : ordType) (v : V) : dom (outmap v) =i pred1 v.
Proof. by move=>x; rewrite /outmap -ssreflect.lock inE. Qed.

Lemma valid_inmap (V : ordType) (v : V) : valid (inmap v).
Proof. by rewrite /inmap -ssreflect.lock. Qed.

Lemma valid_outmap (V : ordType) (v : V) : valid (outmap v).
Proof. by rewrite /outmap -ssreflect.lock. Qed.

Lemma valid_inmap_join V v (f : boolmap V) : 
         valid (inmap v \+ f) = [&& valid f & v \notin dom f]. 
Proof.
case: valid_join; last 1 first.
- by move=>_ -> H; rewrite (H v) //; rewrite dom_inmap inE /=. 
- by rewrite valid_inmap. 
- by case: (valid f).
by move=>x; rewrite dom_inmap andbC; move/eqP=>->->. 
Qed.

Lemma valid_outmap_join V v (f : boolmap V) : 
         valid (outmap v \+ f) = [&& valid f & v \notin dom f]. 
Proof.
case: valid_join; last 1 first.
- by move=>_ -> H; rewrite (H v) //; rewrite dom_outmap inE /=. 
- by rewrite valid_outmap. 
- by case: (valid f).
by move=>x; rewrite dom_outmap andbC; move/eqP=>->->. 
Qed.

Definition invert V (f : boolmap V) := 
  if f is Def m then [pred v | fnd v m == Some true] else pred0.

Lemma invert_join V (f g : boolmap V) : 
        valid (f \+ g) -> 
        invert (f \+ g) = [predU invert f & invert g].
Proof.
case: f g=>[|f][|g] //= Df; rewrite /invert.   
case E: (Def f \+ Def g)=>[|n]; first by rewrite E in Df. 
congr (ssrbool.SimplPred _); apply: fext=>x; rewrite !inE /=.
case: valid_join Df=>// _ _ H _. 
rewrite /PCM.join /= in E.
case: ifP E=>// Df [<-{n}].
rewrite fnd_fcat.
case: ifP=>E; last by move/fnd_supp: (negbT E)=>->; rewrite orbF.
move: E; rewrite -[x \in _]negbK.
by move/contra; move/(_ _ (H x))/fnd_supp=>->. 
Qed.

Lemma invert_meet V (f g : boolmap V) : 
         valid (f \+ g) ->
         [predI invert f & invert g] =i pred0.
Proof.
case: f g=>[|f][|g] //= Df x; rewrite !inE /=.
case: valid_join Df=>// _ _ H _.
case: eqP=>// E.
have: x \in supp f by case: suppP E=>// ->.
by move/H/fnd_supp=>->. 
Qed.

Lemma invert_dom V (f : boolmap V) v : 
        v \in invert f -> v \in dom f.
Proof. by case: f=>[|f] //=; rewrite inE /=; case: suppP=>// ->. Qed.

Lemma invert_inmap (V : ordType) (v : V) : 
        invert (inmap v) = pred1 v.
Proof.
rewrite /inmap -ssreflect.lock.
congr (ssrbool.SimplPred _); apply: fext=>x. 
by rewrite fnd_ins; case: (x =P v).
Qed.

Lemma invert_outmap (V : ordType) (v : V) : 
        invert (outmap v) = pred0.
Proof.
rewrite /outmap -ssreflect.lock.
congr (ssrbool.SimplPred _); apply: fext=>x.
by rewrite fnd_ins; case: (x =P v). 
Qed.


(*****************************************)
(* Implementation of coarse-grained sets *)
(* in terms of sequential library        *)
(*****************************************)

Module Sets.
Section Sets.
Variables (V : ordType) (x : ptr) (setshape : ptr -> pred V -> Pred heap). 
Variable prec : forall x S, precise (setshape x S).

Definition invariant (f : boolmap V) h := h \In setshape x (invert f).

Lemma precise_inv : forall f, precise (invariant f).
Proof. by move=>f; apply: prec. Qed.

Variable (lk : ptr). 

Definition R := Resource lk precise_inv.
Notation U := (univ_of R).
Notation x := (lock_of R).
Notation Inv := (Inv (R:=R)).
Notation procHeap := (procHeap U).
Notation pre := (@pre R).
Notation post := (@post R).
Notation cont := (@cont R).
Notation spec := (spec (R:=R)).

Definition sadd_s (v : V) : spec unit := 
  (fun gE (lp : procHeap) => exists S : pred V, 
     [/\ hp lp \In setshape x S, v \notin S & owns (mx lp)],
   fun _ gE1 gE2 (lp1 lp2 : procHeap) => 
     forall S, hp lp1 \In setshape x S -> 
       [/\ mg lp2 = mg lp1, gE2 = gE1 &
           hp lp2 \In setshape x [predU S & pred1 v]]).
 
Definition sremove_s (v : V) : spec unit := 
  (fun gE (lp : procHeap) => exists S : pred V, 
     [/\ hp lp \In setshape x [predU S & pred1 v], 
         v \notin S & owns (mx lp)],
   fun _ gE1 gE2 (lp1 lp2 : procHeap) => 
     forall S, hp lp1 \In setshape x [predU S & pred1 v] -> 
       [/\ mg lp2 = mg lp1, gE2 = gE1 & hp lp2 \In setshape x S]).

Variables (sadd : forall v, progsep (sadd_s v))
          (sremove : forall v, progsep (sremove_s v)).

Lemma bnd_sadd B (s2 : unit -> spec B) (S : pred V) v gL gE i j r : 
       v \notin S ->
       i \In setshape x S ->
        (forall i', i' \In setshape x [predU S & pred1 v] ->
          verify [mgh true, gL, i' :+ j] gE (s2 tt) r) ->
       verify [mgh true, gL, i :+ j] gE (bind_s (sadd_s v) s2) r.
Proof.
move=>H1 H2 H3.
apply: bnd_do=>[|[] gE' mL' gL' i']; first by exists S.
by case/(_ _ H2)=>[[->->->]]; move/H3. 
Qed.

Lemma bnd_sremove B (s2 : unit -> spec B) (S : pred V) v gL gE i j r : 
       v \notin S ->
       i \In setshape x [predU S & pred1 v] ->
        (forall i', i' \In setshape x S ->
          verify [mgh true, gL, i' :+ j] gE (s2 tt) r) ->
       verify [mgh true, gL, i :+ j] gE (bind_s (sremove_s v) s2) r.
Proof.
move=>H1 H2 H3.
apply: bnd_do=>[|[] gE' mL' gL' i']; first by exists S.
by case/(_ _ H2)=>[[->->->]]; move/H3. 
Qed.

Definition cadd_s (v : V) : spec unit := 
  (fun gE (lp : procHeap) => 
     exists f, lp = [mgh false, f \+ outmap v, empty], 
   fun _ gE1 gE2 (lp1 lp2 : procHeap) => 
     forall f, lp1 = [mgh false, f \+ outmap v, empty] ->
               lp2 = [mgh false, f \+ inmap v, empty]).

Definition cremove_s (v : V) : spec unit := 
  (fun gE (lp : procHeap) => 
     exists f, lp = [mgh false, f \+ inmap v, empty], 
   fun _ gE1 gE2 (lp1 lp2 : procHeap) => 
     forall f, lp1 = [mgh false, f \+ inmap v, empty] ->
               lp2 = [mgh false, f \+ outmap v, empty]).

Definition phi_add (v : V) (f : boolmap V) := 
  if v \in dom f then Some (upd f v true) else None.

Definition phi_remove (v : V) (f : boolmap V) := 
  if v \in dom f then Some (upd f v false) else None.

Lemma phi_add_local (v : V) : annot_axiom (phi_add v).
Proof. 
move=>f1 f2 f3; rewrite /phi_add. 
case: ifP=>Dm // [<-{f2}] Df.
by rewrite dom_join !inE /= Df Dm /= upd_joinL Dm. 
Qed.

Lemma phi_remove_local (v : V) : annot_axiom (phi_remove v).
Proof. 
move=>f1 f2 f3; rewrite /phi_remove. 
case: ifP=>Dm // [<-{f2}] Df.
by rewrite dom_join !inE /= Df Dm /= upd_joinL Dm. 
Qed.

Definition phi_add_annot (v : V) := Annot (@phi_add_local v).
Definition phi_remove_annot (v : V) := Annot (@phi_remove_local v).

Program
Definition cadd (v : V) : progsep (cadd_s v) :=
  Do (lock;;
      sadd v;;
      unlock (phi_add_annot v)).
Next Obligation.
apply: ghE=>//; case=>_ _ _ gE f [->->->] _ _. 
apply: bnd_lock=>{gE} gE inv I; apply: vrfV=>/= Df. 
rewrite /Inv /= /invariant joinAC in Df I.
rewrite invert_join // invert_outmap in I.
apply: bnd_sadd I _.
- rewrite inE /=. rewrite negb_or. rewrite andbT.  
  rewrite joinC in Df.
  case: valid_join Df=>//.
  move=>V1 V2. move=>H _.
  move/(_ v): H. rewrite dom_outmap. move/(_ (eq_refl _)).
  apply: contra. apply: invert_dom.
have Df' : valid (f \+ gE \+ inmap v).
- rewrite joinC valid_outmap_join in Df.
  by rewrite joinC valid_inmap_join.
move=>i' H.
apply: val_unlock=>//=; last first.
- rewrite /Inv /= /invariant /= joinAC.
  rewrite invert_join // invert_inmap. 
  suff L : forall S : simpl_pred V, [predU S & pred0] = S by rewrite L in H.
  case=>S; congr (ssrbool.SimplPred _).
  by apply: fext=>y; rewrite orbF. 
- by rewrite joinAC.
rewrite joinAC in Df.
rewrite /phi_add dom_join !inE (validL Df) /=.
rewrite dom_outmap inE /= eq_refl orbT.
rewrite joinC upd_joinL dom_outmap inE /= eq_refl.
rewrite /upd /inmap /outmap -!ssreflect.lock. 
by rewrite joinC ins_ins eq_refl. 
Qed.


Program
Definition cremove (v : V) : progsep (cremove_s v) :=
  Do (lock;;
      sremove v;;
      unlock (phi_remove_annot v)).
Next Obligation.
apply: ghE=>//; case=>_ _ _ gE f [->->->] _ _. 
apply: bnd_lock=>{gE} gE inv I; apply: vrfV=>/= Df. 
rewrite /Inv /= /invariant joinAC in Df I.
rewrite invert_join // invert_inmap in I.
apply: bnd_sremove I _.
- move/invert_meet/(_ v): Df; rewrite inE /=. 
  case E: (v \in invert _)=>//=.
  by rewrite invert_inmap !inE /= eq_refl. 
have Df' : valid (f \+ gE \+ outmap v).
- rewrite joinC valid_inmap_join in Df.
  by rewrite joinC valid_outmap_join.
move=>i' H.
apply: val_unlock=>//=; last first.
- rewrite /Inv /= /invariant /= joinAC.
  rewrite invert_join // invert_outmap. 
  suff -> : forall S : simpl_pred V, [predU S & pred0] = S by [].
  case=>S; congr (ssrbool.SimplPred _).
  by apply: fext=>y; rewrite orbF. 
- by rewrite joinAC.
rewrite joinAC in Df.
rewrite /phi_remove dom_join !inE (validL Df) /=.
rewrite dom_inmap inE /= eq_refl orbT.
rewrite joinC upd_joinL dom_inmap inE /= eq_refl.
rewrite /upd /inmap /outmap -!ssreflect.lock. 
by rewrite joinC ins_ins eq_refl. 
Qed.

End Sets.
End Sets.

