Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq rels ordtype.
Require Import Eqdep prelude heaps.
Require Import pcm resources8 proc9 step9 sbisim9 wbisim9.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(************************************)
(* Action safety relative to a heap *)
(************************************)

Definition asafe A (a : action A) : Pred heap :=
  fun i => 
    match a with
      IdleAct pf => True
    | AllocAct pf v => True
    | DeallocAct pf x => x \in dom i
    | ReadAct x => x \in dom i /\ Dyn.typ (look x i) = A
    | WriteAct pf x v => x \in dom i
    | CASAct B1 B2 pf x v1 v2 => x \in dom i /\ Dyn.typ (look x i) = B1
    end.

(* views for asafe *)

Lemma asf_idleP i : i \In asafe Idle. 
Proof. by []. Qed. 

Lemma asf_allocP A i (v : A) : i \In asafe (Alloc v).
Proof. by []. Qed. 

Definition InE' := (Mem_Seq1, In_cons, In_Simpl, Simpl_PredE). (* to be extended *)

Lemma asf_deallocP i (x : ptr) : 
        def i -> 
         (i \In asafe (Dealloc x) <->  
          exists j, exists A, exists v : A, i = x :-> v :+ j).
Proof.
move=>D; split; last first.
- case=>j [A][v]; rfe1 D.
  by rewrite -toPredE /= domPtUn inE /= eq_refl D.
move=>H; case/heap_etaP: H D=>-> D.
by exists (free x i), (Dyn.typ (look x i)), (Dyn.val (look x i)).
Qed.

Lemma asf_deallocX A i (x : ptr) (v : A) : 
        def (x :-> v :+ i) -> 
        x :-> v :+ i \In asafe (Dealloc x). 
Proof. by move=>D; rewrite -toPredE /= domPtUn inE /= eq_refl D. Qed.

Lemma asf_readP A i (x : ptr) : 
        def i -> 
        (i \In asafe (@Read A x) <-> exists j, exists v : A, i = x :-> v :+ j).
Proof.
move=>D; split; last first.
- case=>j [v] E; rewrite {i}E in D *.
  by rewrite -toPredE /= domPtUn inE /= eq_refl lookPtUn // D.
case=>H <-; case/heap_etaP: H D=>-> D.
by exists (free x i); rewrite lookPtUn //=; exists (Dyn.val (look x i)).
Qed.

Lemma asf_readPQ A B i (x : ptr) (v : B) : 
        def (x :-> v :+ i) -> 
        (x :-> v :+ i \In asafe (@Read A x) <-> A = B).
Proof.
move=>D; rewrite (asf_readP A x D); split.
- by case=>j [w]; case/(cancelT D)=>->.
by rbe2 v D; exists i, v.
Qed.

Lemma asf_readX A i (x : ptr) (v : A) : 
        def (x :-> v :+ i) -> x :-> v :+ i \In asafe (@Read A x). 
Proof. by move=>D; rewrite -toPredE /= domPtUn inE /= eq_refl D lookPtUn. Qed.

Lemma asf_writeP A i (x : ptr) (v : A) : 
        def i -> 
         (i \In asafe (Write x v) <-> 
          exists j, exists B, exists w : B, i = x :-> w :+ j).
Proof.
move=>D; split; last first.
- by case=>j [B][w]; rfe1 D; rewrite -toPredE /= domPtUn inE /= eq_refl D.
move=>H; case/heap_etaP: H D=>-> D.
by exists (free x i), (Dyn.typ (look x i)), (Dyn.val (look x i)).
Qed.

Lemma asf_writeX A B i (x : ptr) (v : A) (w : B) : 
        def (x :-> w :+ i) -> x :-> w :+ i \In asafe (Write x v).
Proof. by move=>D; rewrite -toPredE /= domPtUn inE /= eq_refl D. Qed.

Lemma asf_casP (A B : eqType) i (x : ptr) (ov : A) (nv : B) : 
        def i -> 
        (i \In asafe (CAS x ov nv) <-> exists j, exists v : A, i = x :-> v :+ j).
Proof.
move=>D; split.
- case=>/= H <-; case/heap_etaP: H D=>-> D; exists (free x i);
  by rewrite lookPtUn //=; exists (Dyn.val (look x i)).
case=>j[v] E; rewrite {i}E in D *.
by rewrite -toPredE /= domPtUn inE /= eq_refl lookPtUn // D.
Qed.

Lemma asf_casPQ (A B : eqType) A' i (x : ptr) (ov : A) (nv : B) (v : A') : 
        def (x :-> v :+ i) -> 
        (x :-> v :+ i \In asafe (CAS x ov nv) <-> Equality.sort A = A').
Proof.
move=>D; rewrite (asf_readP A x D); split.
- by case=>j [w]; case/(cancelT D)=>->.
by rbe2 v D; exists i, v.
Qed.

Lemma asf_casX (A B : eqType) i (x : ptr) (ov : A) (nv : B) (v : A) :
        def (x :-> v :+ i) -> x :-> v :+ i \In asafe (CAS x ov nv).
Proof. 
by move=>D; rewrite -toPredE /= domPtUn inE /= eq_refl D lookPtUn.
Qed.

Lemma asf_mono A i j (a : action A) :
        def (i :+ j) -> i \In asafe a -> i :+ j \In asafe a.
Proof.
case/actP:a.
- by move=>-> pf D; rewrite eqc.
- by move=>-> B v pf D; rewrite eqc. 
- move=>-> x pf D; rewrite eqc=>H.
  by rewrite -toPredE /= domUn !inE /= D H.
- move=>x D; case=>H <-.
  by rewrite -toPredE /= lookUnl // domUn !inE /= D H.
- move=>-> B x v pf D; rewrite eqc=>H.
  by rewrite -toPredE /= domUn !inE /= D H.
move=>-> B C x ov nv pf D; rewrite eqc=>[[H /= E]];
by rewrite -toPredE /= lookUnl // domUn !inE /= D -E H.
Qed.

Hint Resolve asf_idleP asf_allocP asf_deallocX asf_readX asf_writeX asf_casX.

(*******************************)
(* Stepping relative to a heap *)
(*******************************)

Definition astep A i (a : action A) : A -> heap -> Prop :=
  match a with
  | IdleAct pf => fun y m => m = i
  | AllocAct pf v => 
      fun y m => let: z := coerce (T:=id) y pf in
        [/\ m = z :-> Dyn.val v :+ i, z \notin dom i & z != null]
  | DeallocAct pf x => fun y m => m = free x i
  | ReadAct x => fun y m => m = i /\ dyn y = look x i
  | WriteAct pf x v => fun y m => m = upd i x v
  | CASAct A1 A2 pf x v1 v2 => 
      fun y m => let: b := coerce (T:=id) y pf in
      exists v : A1, 
        [/\ look x i = dyn v, b = (v == v1) &
            m = if b then upd i x (dyn v2) else i]
  (* | CASAct B1 B2 pf x ov nv => fun y m => *)
  (*     let: z := coerce (T:=id) y pf in *)
  (*     let: hd := look x i in *)
  (*       exists pf' : Dyn.typ hd = Equality.sort B1, *)
  (*         let: hv := coerce (T:=id) (Dyn.val hd) pf' in *)
  (*           if hv == ov *)
  (*           then m = upd i x (dyn nv) /\ z = true *)
  (*           else m = i /\ z = false *)
  end.

Lemma ast_idleP i m : astep i Idle tt m -> m = i.
Proof. by []. Qed.

Lemma ast_idleX i : astep i Idle tt i. 
Proof. by []. Qed.

Hint Resolve ast_idleX.

Lemma ast_allocP A i (v : A) (x : ptr) m : 
        astep i (Alloc v) x m <-> 
        [/\ m = x :-> v :+ i, x \notin dom i & x != null].
Proof. by []. Qed.

Lemma ast_allocX A i (v : A) (x : ptr) : 
        def (x :-> v :+ i) -> astep i (Alloc v) x (x :-> v :+ i).
Proof. by move=>D /=; rewrite (defPt_null D) (defPt_dom D). Qed.

Lemma ast_deallocP i (x : ptr) m : 
        astep i (Dealloc x) tt m <-> m = free x i.
Proof. by []. Qed.

Lemma ast_deallocX A i (v : A) (x : ptr) : 
        def (x :-> v :+ i) -> astep (x :-> v :+ i) (Dealloc x) tt i.
Proof. by move=>D /=; rewrite freePtUn. Qed.

Lemma ast_readP A i (v : A) (x : ptr) m : 
        astep i (@Read A x) v m -> m = i /\ dyn v = look x i.
Proof. by []. Qed.

Lemma ast_readPQ A i (v w : A) (x : ptr) m : 
        def (x :-> w :+ i) -> 
        astep (x :-> w :+ i) (@Read A x) v m -> m = x :-> v :+ i /\ v = w.
Proof.
move=>D /= [->]; rewrite lookPtUn // => E.
by rewrite (dyn_inj E).
Qed.

Lemma ast_readX A i (v : A) (x : ptr) : 
        def (x :-> v :+ i) -> 
        astep (x :-> v :+ i) (@Read A x) v (x :-> v :+ i).
Proof. by move=>D /=; rewrite lookPtUn. Qed.

Lemma ast_writeP A i (v : A) (x : ptr) m : 
        astep i (Write x v) tt m <-> m = upd i x (dyn v). 
Proof. by []. Qed. 

Lemma ast_writeX A B i (v : A) (w : B) (x : ptr) : 
        astep (x :-> w :+ i) (Write x v) tt (x :-> v :+ i).
Proof. by move=>/=; rewrite updPtUn. Qed.

Hint Resolve ast_writeX.

Lemma ast_casP (B1 B2 : eqType) x (v1 : B1) (v2 : B2) i m b :
        astep i (CAS x v1 v2) b m <->
        exists v : B1, 
          [/\ look x i = dyn v, b = (v == v1) &
              m = if b then upd i x (dyn v2) else i].
          (* (let: hd := look x i in *)
          (*  exists pf : Dyn.typ hd = Equality.sort B1, *)
          (*    let: hv := coerce (T:=id) (Dyn.val hd) pf in *)
          (*    if hv == v1 *)
          (*    then m = upd i x (dyn v2) /\ b = true *)
          (*    else m = i /\ b = false). *)
Proof. by []. Qed.

Lemma ast_casX (B1 B2 : eqType) x (v v1 : B1) (v2 : B2) i :
        def (x :-> v :+ i) ->
        astep (x :-> v :+ i) (CAS x v1 v2) (v == v1) 
              (if v == v1 then x :-> v2 :+ i else x :-> v :+ i).
Proof.
move=>D; apply/ast_casP; exists v; split=>//;
by [rewrite lookPtUn | case: eqP=>// ->; rewrite updPtUn].
Qed.
(* Lemma ast_casX (B1 B2 : eqType) (v0 v1 : B1) (v2 : B2) (x : ptr) i :  *)
(*         def (x :-> v0 :+ i) -> *)
(*         astep (x :-> v0 :+ i) (CAS x v1 v2) (v0 == v1) (if v0 == v1 then x :-> v2 :+ i else x :-> v0 :+ i). *)
(* Proof. *)
(* case:B1 v0 v1=>B1 [eq1 H1] T1 v0 v1; case:B2 v2=>[eq2 H2] T2  v2 D /=; *)
(* by rewrite updPtUn lookPtUn //=; exists (erefl B1); case:ifP=>->. *)
(* Qed. *)

(* definedness preservation lemma *)
Lemma astD A i m (v : A) (a : action A) :
        def i -> i \In asafe a -> astep i a v m -> def m. 
Proof.
case/actP:a.
- by rfe1 v=>{A} pf D _; rewrite eqc; move/ast_idleP=>->.
- rfe1 v=>{A} B w pf D _; rewrite eqc;
  by case/ast_allocP=>->; rewrite defPtUn D=>->->.
- rfe1 v=>{A} x pf D; rewrite eqc; case/(asf_deallocP _ D)=>j [A][w];
  by rfe1 D; move/ast_deallocP=>->; rewrite defF.
- move=>x D; case/(asf_readP _ _ D)=>j [w]; rfe1 D;
  by case/(ast_readPQ D)=>->->.
- rfe1 v=>{A} B x w pf D; rewrite eqc; case/(asf_writeP _ _ D)=>j [C][z];
  by rfe1 D; move/(ast_writeP)=>->; rewrite defU (defPt_null D).
rfe1 v=>B1 B2 x v1 v2 pf D; rewrite eqc; case/asf_casP=>// j [v0]; rfe1 D;
case/ast_casP=>w2 [_ _]; case: ifP=>_ -> //.
by rewrite updPtUn !defPtUn in D *.
(* by case/ast_casP=>E; case:ifP=>E' [->] _ //; rewrite defU (defPt_null D). *)
Qed.

(* progress lemma *)
Lemma astX A i (a : action A) : 
        def i -> i \In asafe a -> {m : heap & {v : A | astep i a v m}}.
Proof.
case/actP:a.
- by move=>-> pf D; rewrite eqc; exists i, tt.
- move=>-> B v pf D; rewrite eqc.
  exists (fresh i :-> v :+ i), (fresh i). 
  by rewrite /= -{3}[fresh i]ptr0 dom_fresh.
- move=>-> x pf; rewrite eqc; case=>D H.
  by exists (free x i), tt.
- move=>x D; case=>H <-.
  by exists i, (Dyn.val (look x i)); rewrite /= -dyn_eta.
- move=>-> B x v pf D _; rewrite eqc.
  by exists (upd i x (dyn v)), tt.
move=>-> A1 A2 /= x v1 v2 pf D; rewrite {pf}eqc.
case; case/heap_etaP; rfe1 D=>/=; rewrite lookPtUn //=.
by case: (look x i) D=>B v /= D; rfe2 v D; eexists _, (v == v1), v.
(* move=>-> B1 B2 /= x v1 v2 pf D; *)
(* rewrite eqc=>{pf}; case; case/heap_etaP; rfe1 D=>/=; rewrite lookPtUn //=. *)
(* case: B1 v1=>B1 /= M1 T1 v1; rbe2 v1 M1; case:M1=>op E. *)
(* (* XXX is there a nicer way of doing this: *) *)
(* have L : forall (b : bool) P, (b -> P) -> (~~ b -> P) -> P. *)
(* - by clear; case=>[P|P _]; move/(_ (erefl _)). *)
(* apply: (L (op (Dyn.val (look x i)) v1)); *)
(* move/(E (Dyn.val (look x i)) v1)=>H; *)
(* [exists (upd (x :-> Dyn.val (look x i) :+ free x i) x (dyn v2)), true *)
(* |exists (x :-> Dyn.val (look x i) :+ free x i), false]; *)
(* exists (erefl _); [|rewrite -if_neg]=>//=; *)
(* by [rewrite -H /eqP /ifP eq_refl| *)
(* case:ifP; move/eqP]. *)
Qed.

(* frame lemma *)
(* i'm including the def m clause in the conclusions because it's *)
(* often used *)
Lemma ast_frame A i j m (a : action A) (v : A) :
        def (i :+ j) -> i \In asafe a -> astep (i :+ j) a v m -> 
        exists m', [/\ m = m' :+ j, def m & astep i a v m'].
Proof.
move=>D H1 H2.
suff [m' [E] L]: exists m', m = m' :+ j /\ astep i a v m'.
- rewrite E in H2 *; exists m'; split=>//.
  by apply: astD H2=>//; apply: asf_mono.
case/actP: a v D H1 H2.
- by move=>-> pf [] D _; rewrite eqc; move/ast_idleP=>->; exists i.
- move=>-> B v pf x D _; rewrite eqc; case/ast_allocP=>-> H1 H2.
  exists (x :-> v :+ i); rewrite unA ast_allocP H2.
  by move: H1; rewrite domUn !inE D /= negb_or; case/andP=>->.
- move=>-> x pf v D; rewrite eqc.
  case/(asf_deallocP _ (defUnl D))=>k [B][w] E; rewrite {}E -unA in D *.
  case/ast_deallocP=>->; exists k; rewrite freePtUn //. 
  by rewrite unA in D; split=>//; apply: (ast_deallocX (defUnl D)).
- move=>x v D; case/(asf_readP _ _ (defUnl D))=>k [w] E. 
  rewrite {}E -unA in D *; case/(ast_readPQ D)=>->->.
  exists (x :->w :+ k); rewrite unA in D *; split=>//.
  by apply: ast_readX (defUnl D).
- move=>-> B x v pf w D; rewrite eqc.
  case/(asf_writeP _ _ (defUnl D))=>k [C][z] E; rewrite {}E in D *.
  rewrite -unA in D *; case/ast_writeP=>->.
  exists (x :-> v :+ k); rewrite updPtUn !unA in D *.
  by split=>//; apply: ast_writeX.
(* XXX clean this up *)
move=>-> A1 A2 x v1 v2 pf b D; rewrite {pf}eqc.
case/(asf_casP _ _ _ (defUnl D))=>k [v] E; rewrite {}E -unA in D *.
case/ast_casP=>w1 []; rewrite lookPtUn //; move/dyn_inj=>{w1}<- -> ->.
rewrite unA in D *.
eexists _; split; last by exists v; split=>//=; rewrite (lookPtUn (defUnl D)).
by case: eqP=>//; rewrite updPtUn -unA updPtUn unA.
(* move=>-> [B1] [E1 H1] T1 [B2] [E2 H2] T2 x v1 v2 pf v D; move/defUnl:(D)=>D'. *)
(* rewrite eqc=>{pf}; case/asf_casP=>//i' [v1']; rfe2 D D'=>[[]] /=; *)
(* rewrite -unA in D; rewrite -unA !lookPtUn //==>pf; case: ifP=>H [->->]. *)
(* - exists (upd (x :-> v1' :+ i') x (dyn v2)); split. *)
(*   - by do 2!rewrite updUnl (domPtX (defUnl D)); rewrite unA. *)
(*   by exists pf; rewrite H. *)
(* by exists (x :-> v1' :+ i'); split; [rewrite unA | exists pf; rewrite H]. *)
Qed.

(*****************)
(* Tree stepping *)
(*****************)

Inductive branch := LL | RR. (* b *)
Definition pick := seq branch.
Definition path := seq pick.

Section BranchEqType.

Definition branch_eq (u v : branch) : bool :=
  match u, v with
    | LL, LL => true
    | RR, RR => true
    | _, _ => false
  end.  

  (* oapp (fun x => oapp (eq_op x) false v) (~~ v) u. *)

Lemma branch_eqP : Equality.axiom branch_eq.
Proof. by do 2!case; constructor. Qed.

Canonical Structure branch_eqMixin := EqMixin branch_eqP.
Canonical Structure branch_eqType :=
  Eval hnf in EqType branch branch_eqMixin.

End BranchEqType.

(* Picking an action from the process following bs : pick *)
(* This could be made a function, but the function has to be partial *)
(* so we immediately define a relation instead *)

(* actions can be annotated with ghost functions *)
(* so we introduce special notation for that *)

Definition actiong (U : pcm) B := ((action B) * annot U)%type.

Fixpoint action_at U A B (p : tree U A) bs : Pred (actiong U B) :=
  match bs, p with
    [::], Cons C a g p' => 
      fun b => [/\ b.2 = g, B = C & jmeq action b.1 a]
  | [::], Par _ _ (Ret x) (Ret y) p' =>
      fun b => [/\ b.2 = id_annot, B = unit & jmeq action b.1 Idle]
  | LL :: cs, Par B1 B2 p' q r => action_at p' cs
  | RR :: cs, Par _ _ p q' r => action_at q' cs
  | _, _ => fun b => False
  end.

Lemma act_retP U A B (v : A) (ag : actiong U B) bs : 
       ag  \In action_at (Ret v) bs <-> False.
Proof. by split=>//; case: bs=>[|[]] //. Qed.

Lemma act_dummyP U A C (ag : actiong U C) bs : 
        ag \In action_at (@Dummy U A) bs <-> False.
Proof. by split=>//; case: bs=>[|[]]//. Qed.

Lemma act_consP U A A' B a (p : A -> tree U B) (a' : action A') g g' bs :
         (a', g') \In action_at (Cons a g p) bs <-> 
         [/\ bs = [::], g = g', A = A' & jmeq action a a'].
Proof.
case: bs=>[|b bs]; last by split; [case: b | case].
split=>[[/= -> E]|[_ -> E]]; first by rewrite {A'}E in a' *; move/jmE=>->.
by rewrite -{A'}E in a' *; move/jmE=>->.
Qed.

(* view occasionally useful for forward reasoning *)
Lemma act_consPQ U A B a (p : A -> tree U B) g : 
        (a, g) \In action_at (Cons a g p) [::].
Proof. by []. Qed.

Lemma act_retparP U B B1 B2 C p1 p2 (p' : B1 * B2 -> tree U C) g (a : action B) :
        (a, g) \In action_at (Par p1 p2 p') [::] <->
        exists v1, exists v2, 
          [/\ g = id_annot, B = unit, jmeq action a Idle, 
              p1 = Ret v1 & p2 = Ret v2].
Proof.
split; last first.
- by case=>v1 [v2][->][E]; rewrite {B}E in a *; move/jmE=>->->->.
case: p1=>// v1; case: p2=>// v2 [/= -> E].
by rewrite {B}E in a *; move/jmE=>/= ->; exists v1, v2.
Qed.

Lemma act_parP U B B1 B2 C p1 p2 (p' : B1 * B2 -> tree U C) 
               (a : action B) g bs :
        (a, g) \In action_at (Par p1 p2 p') bs <->
        (exists v1, exists v2, 
           [/\ bs = [::], g = id_annot, B = unit, jmeq action a Idle, 
               p1 = Ret v1 & p2 = Ret v2]) \/
        ((bs = LL :: behead bs /\ (a, g) \In action_at p1 (behead bs)) \/
         (bs = RR :: behead bs /\ (a, g) \In action_at p2 (behead bs))).
Proof.
split=>[|]; last by case=>[[v1][v2][->-> _ _ ->-> //]|[][->]].
case: bs; last by case=>bs H; right; [left|right].
by case/act_retparP=>v1 [v2][->][E1 E2 ->->]; left; exists v1, v2.
Qed.

Lemma act_lpar_dummyP U A1 A2 (q' : tree U A2) B (ag : actiong U B) bs :
        ag \In action_at (Par (@Dummy U A1) q' Ret) bs ->
        exists bs', bs = RR :: bs' /\ ag \In action_at q' bs'.
Proof.
case: ag=>a g; case/act_parP; first by case=>v1 [v2][].
case; first by case=>->; move/act_dummyP.
by case=>->; exists (behead bs).
Qed.

Lemma act_nilB U A (p : tree U A) B (a : action B) g :
        (a, g) \In action_at p [::] <->
        (exists p', p = Cons a g p') \/
        exists B1, exists B2, exists v1 : B1, exists v2 : B2, exists p',
          [/\ g = id_annot, B = unit, jmeq action a Idle & 
              p = Par (Ret v1) (Ret v2) p'].
Proof.
split; last first.
- case=>[[p'][->] //|[B1][B2][v1][v2][p'][->] E].
  by rewrite {B}E in a *; move/jmE=>->->.
case: p=>//.
- move=>B' a' g' p'; case/act_consP=>_ -> E.
  by rewrite {B'}E in a' p' *; move/jmE=>->; left; exists p'.
move=>B1 B2 p1 p2 p; case/act_retparP=>v1 [v2][->] E.
rewrite {B}E in a *; move/jmE=>->->->.
by right; exists B1, B2, v1, v2, p.
Qed.

Lemma act_consB U A (p : tree U A) B (a : action B) g b bs :
        (a, g) \In action_at p (b :: bs) <->
        exists B1, exists B2, 
        exists p1 : tree U B1, exists p2 : tree U B2, exists p',
          p = Par p1 p2 p' /\
         ((b = LL /\ (a, g) \In action_at p1 bs) \/ 
          (b = RR /\ (a, g) \In action_at p2 bs)).
Proof.
split; last by case=>B1 [B2][p1][p2][p'] [->]; case; case=>->.
case: p; case: b=>//; move=>B1 B2 p1 p2 p;
exists B1, B2, p1, p2, p; split=>//; 
by [left | right].
Qed.

(* action_at is a functional relation *)
Lemma actE U A B1 B2 (p : tree U A) 
           (a1 : action B1) (a2 : action B2) g1 g2 bs : 
        (a1, g1) \In action_at p bs -> (a2, g2) \In action_at p bs ->
        [/\ g1 = g2, B1 = B2 & jmeq action a1 a2].
Proof.
elim: bs A p=>[|b bs IH] A p; last first.
- by case: p; case: b=>//= C1 C2 p1 p2 r; apply: IH.
case: p=>//; [move=>B a g t [/= <-]|move=>C1 C2 [] // v1 [] // v2 r [/= ->]]; 
by rfejm1 a1=>[[/= <-]]; rfejm1 a2.
Qed.

(* action with non-empty type always makes a step *)
Lemma act_st U A B (p : tree U A) (a : action B) (v : B) g bs : 
        (a, g) \In action_at p bs -> 
        {q : tree U A | step p [tr a, v] q}.
Proof.
elim: bs A p=>[|b bs IH] A p; last first.
- case: b; case: p=>// B1 B2 p1 p2 r; case/IH=>q S.
  - by exists (Par q p2 r); apply: st_lpar S.
  by exists (Par p1 q r); apply: st_rpar S.
case: p=>//.
- by move=>B1 a1 g1 t [/= <-]; rfejm2 a v; exists (t v).
move=>B1 B2 [] // v1 [] // v2 r [/= _].
by rfejm2 a v; exists (r (v1, v2)); case: v.
Qed.

(* and vice-versa *)
Lemma st_act U A B (p q : tree U A) a (v : B) : 
        step p [tr a, v] q -> 
        exists g, exists bs, (a, g) \In action_at p bs.
Proof.
elim: p q a v=>{A}[A|A a1|A B1 a1 g1 p _|A B1 B2 p1 IH1 p2 IH2 r _] q a v.
- by case/st_dummyP.
- by case/st_retP.
- by case/st_consP=>/=; rbejm2 a v=>E; exists g1, [::].
case/st_parP=>[|[] {IH1 IH2}].
- case; case=>p [_]; [case/IH1 | case/IH2]=>g [bs] T; exists g;
  by [exists (LL :: bs) | exists (RR :: bs)].
by rewrite trD; case; rfejm2 a v=>_ [x][y][->][->] _; exists id_annot, [::].
Qed.

Prenex Implicits st_act.

(*
(* one can define redex and safe pick as below *)
(* but there is really no need to introduce new definitions *)

(* redex is an action which can step (i.e., it's type is not empty) *)
Definition redex A (p : tree A) bs :=
  exists B, exists a : action B, exists v : B,
    a \In action_at p bs. 

(* safe_pick is an action whose pre is satisfied *)
Definition safe_pick A (p : tree A) bs : Pred heap := 
  fun i => 
  exists B, exists a : action B, 
  a \In action_at p bs /\ i \In asafe a.
*)

(**************************)
(* Stepping with the heap *)
(**************************)

Fixpoint hstep U A i (p : tree U A) bs m q : Prop := 
  match bs, p with
    [::], Cons B a g r => 
      exists v, astep i a v m /\ q = r v
  | [::], Par _ _ (Ret v1) (Ret v2) r => 
      m = i /\ q = r (v1, v2)
  | LL :: cs, Par _ _  p1 p2 r =>
      exists p1', q = Par p1' p2 r /\ hstep i p1 cs m p1' 
  | RR :: cs, Par _ _ p1 p2 r =>
      exists p2', q = Par p1 p2' r /\ hstep i p2 cs m p2'
  | _, _ => False
  end.

(* a view for hstep *)
Lemma hstP U A i m bs (p q : tree U A) :
        hstep i p bs m q <->
        bs = [::] /\
          ((exists B, exists a, exists g, exists p', exists v : B,
              [/\ p = Cons a g p', q = p' v & astep i a v m]) \/
           (exists B1, exists B2, exists v1 : B1, exists v2 : B2, exists p',
              [/\ m = i, p = Par (Ret v1) (Ret v2) p' & q = p' (v1,v2) ])) \/
        exists B1, exists B2, 
        exists p1 : tree U B1, exists p2 : tree U B2, exists p',
          p = Par p1 p2 p' /\
          ((exists p1',
              [/\ bs = LL :: behead bs, q = Par p1' p2 p' & 
                  hstep i p1 (behead bs) m p1']) \/ 
           (exists p2',
              [/\ bs = RR :: behead bs, q = Par p1 p2' p' & 
                  hstep i p2 (behead bs) m p2'])).
Proof.
case:bs=>[|b bs]; split=>/=.
- case: p=>//.
  - move=>B a g t [v][T] ->; left; split=>//; left.
    by exists B, a, g, t, v.
  move=>B1 B2 [] // v1 [] // v2 p' [->->]; left; split=>//; right.
  by exists B1, B2, v1, v2, p'.
- case; last by case=>B1 [B2][p1][p2][p'][->][][] p1' [].
  case=>_; case; first by case=>B[a][g][p'][v] [->->T]; exists v.
  by case=>B1[B2][a][p'][v][->->->].
- by case: b; case: p=>//= B1 B2 p1 p2 p' [pi'][->] T; right;
  exists B1, B2, p1, p2, p'; 
  split=>//; [left | right]; exists pi'.
case=>[[//]|[B1][B2][p1][p2][p'][->]][];
by case=>pi' [[->->]] T; exists pi'.
Qed.

Lemma hst_dummyP U A i m q bs : hstep i (@Dummy U A) bs m q <-> False.
Proof. by case:bs=>[|b bs] //=; case: b. Qed.

Lemma hst_retP U A i m (x : A) q bs : 
        hstep i (Ret (U:=U) x) bs m q <-> False.
Proof. by case: bs=>[|b bs] //=; case: b. Qed.

Lemma hst_consP U A B g i m a (p : A -> tree U B) q bs : 
        hstep i (Cons a g p) bs m q <->
        exists v, [/\ q = p v, bs = [::] & astep i a v m].
Proof.
case:bs=>[|b bs]; last by split; [case: b=>// | case=>v []].
by split; case=>v; [case=>H -> | case=>-> _ H]; exists v.
Qed.

Lemma hst_consPQ U A B g i m (a : action A) (p : A -> tree U B) v : 
        astep i a v m -> hstep i (Cons a g p) [::] m (p v).
Proof. by move=>H; apply/hst_consP; exists v. Qed.

Lemma hst_parP U A1 A2 B i p q m (r : A1 * A2 -> tree U B) s bs :
        hstep i (Par p q r) bs m s <-> 
        ((exists p1,
            [/\ bs = LL :: behead bs, s = Par p1 q r &
                hstep i p (behead bs) m p1]) \/
         (exists q1,
            [/\ bs = RR :: behead bs, s = Par p q1 r &
                hstep i q (behead bs) m q1])) \/
          exists x, exists y, 
            [/\ bs = [::], m = i, s = r (x, y),
                p = Ret x & q = Ret y].
Proof.
split; last first.
- case; last by case=>x [y][->]->->->->.
  by case; case=>u [->] -> /= H; exists u.
case: bs=>[|b bs]; last first.
- by case: b; case=>u [->] H; left; [left | right]; exists u. 
case: p=>// x; case: q =>//= y [-> ->].
by right; exists x, y.
Qed.

Lemma hst_parPQ U A1 A2 B i m (p1 p2 : tree U A1) (q1 q2 : tree U A2) 
                              (r1 r2 : _ -> tree U B) bs :
        hstep i (Par p1 q1 r1) bs m (Par p2 q2 r2) <->
        (r2 = r1 /\ 
           ([/\ bs = LL :: behead bs, q2 = q1 &
                hstep i p1 (behead bs) m p2] \/
            [/\ bs = RR :: behead bs, p2 = p1 &
                hstep i q1 (behead bs) m q2])) \/
        exists x, exists y,
          [/\ bs = [::], m = i, r1 (x, y) = Par p2 q2 r2, 
              p1 = Ret x & q1 = Ret y].
Proof.
rewrite hst_parP; split.
- case; last by case=>x [y][->->] E ->->; right; exists x, y.
  by case; case=>p [->]; case/parE=>_ _; do 2![move/jmE=>->]; 
     move/jm2E=>-> /= H; left; split=>//; [left | right].
case; last  by case=>x [y][->->] E ->->; right; exists x, y.
case=>-> []; case=>->-> /= H; left; 
by [left; exists p2 | right; exists q2].
Qed.

Lemma hst_retparP U A1 A2 B i m p q (r : A1 * A2 -> tree U B) r' :
        hstep i (Par p q r) [::] m r' <->
        exists x, exists y,
          [/\ m = i, r' = r (x, y),
              p = Ret x & q = Ret y].
Proof.
split.
- case/hst_parP; first by case=>[] [pi] [].
  by case=>x[y] [_->->->->]; exists x, y.
by case=>x[y] [->->->->]; apply/hst_parP; right; exists x, y.
Qed.

Lemma hst_lparP U A B C i m (p1 p2 : tree U A) (q : tree U B) 
                            (r : _ -> tree U C) bs :
        hstep i p1 bs m p2 -> 
        hstep i (Par p1 q r) (LL :: bs) m (Par p2 q r).
Proof. by move=>*; apply/hst_parPQ; left; split=>//; left. Qed.

Lemma hst_rparP U A B C i m (p : tree U A) (q1 q2 : tree U B) 
                            (r : _ -> tree U C) bs :
        hstep i q1 bs m q2 -> 
        hstep i (Par p q1 r) (RR :: bs) m (Par p q2 r).
Proof. by move=>*; apply/hst_parPQ; left; split=>//; right. Qed.

Lemma hst_parretP U A1 A2 i m (p1 p2 : tree U A1) (q1 q2 : tree U A2) bs :
        hstep i (Par p1 q1 Ret) bs m (Par p2 q2 Ret) <->
        ([/\ bs = LL :: behead bs, q2 = q1 &
             hstep i p1 (behead bs) m p2] \/
         [/\ bs = RR :: behead bs, p2 = p1 &
             hstep i q1 (behead bs) m q2]).
Proof.
split.
- by case/hst_parPQ=>[[]|[v1][v2][]].
by case; case=>->->T; [apply/hst_lparP|apply/hst_rparP].
Qed.

Lemma hst_lpar_dummyP U i m bs A B p (q : tree U (A * B)) :
        hstep i (Par Dummy p Ret) bs m q ->
        exists p', exists bs', 
          q = Par Dummy p' Ret /\ hstep i p bs' m p'.
Proof.
case/hst_parP=>[[]|[x1][x2][]]//.
- by case=>r [_ _]; move/hst_dummyP.
by case=>p' [E->T]; exists p', (behead bs).
Qed.

Lemma hst_lpar_dummy_retP U i m bs A (v : A) B (p q : tree U B) r :
        hstep i (Par (@Dummy U A) p Ret) bs m (Par (Ret v) q r) <-> False.
Proof.
by split=>//; case/hst_lpar_dummyP=>p'[bs'] []; case/parE=>_ _; move/jmE.
Qed.

Lemma hst_lparPQ U A1 A2 B i m p q (r : A1 * A2 -> tree U B) r' bs :
        hstep i (Par p q r) (LL :: bs) m r' <->
        exists p', r' = Par p' q r /\ hstep i p bs m p'.
Proof. by split=>//. Qed.

Lemma hst_rparPQ U A1 A2 B i m p q (r : A1 * A2 -> tree U B) r' bs :
        hstep i (Par p q r) (RR :: bs) m r' <->
        exists q', r' = Par p q' r /\ hstep i q bs m q'.
Proof. by split=>//. Qed.

Lemma hst_lpar_retP U A B i m (v v' : A) (q q' : tree U B) bs :
        hstep i (Par (Ret v) q Ret) bs m (Par (Ret v') q' Ret)
        <-> [/\ v = v', bs = RR :: (behead bs) & hstep i q (behead bs) m q'].
Proof.
split; last by case=>->->; move/(hst_rparP (Ret v') (Ret)).
case/hst_parPQ; last by case=>v1 [v2][].
case=>_ []; case=>->; last by case=>->.
by move=>->; move/hst_retP.
Qed.

Lemma hst_rpar_retP U A B i m (v v' : A) (q q' : tree U B) bs :
        hstep i (Par q (Ret v) Ret) bs m (Par q' (Ret v') Ret)
        <-> [/\ v = v', bs = LL :: (behead bs) & hstep i q (behead bs) m q'].
Proof.
split; last by case=>->->; move/(hst_lparP (Ret v') (Ret)).
case/hst_parPQ; last by case=>v1 [v2][].
case=>_ []; case=>->; first by case=>->.
by move=>->; move/hst_retP.
Qed.

Lemma hst_rlparP U A1 A2 B i m (p1 p2 : tree U A1) (q1 q2 : tree U A2) 
                               (r1 r2 : _ -> tree U B) b bs :
        hstep i (Par p1 q1 r1) (b :: bs) m (Par p2 q2 r2) <->
        (r2 = r1 /\ 
          ([/\ b = LL, q2 = q1 & hstep i p1 bs m p2] \/
           [/\ b = RR, p2 = p1 & hstep i q1 bs m q2])).
Proof.
split=>//; last by case=>-> []; case=>->->; [move/hst_lparP | move/hst_rparP].
move/hst_parPQ; case; last by case=>x[y][].
by case=>-> []; case=>[][->->T]; split=>//; [left|right].
Qed.

Lemma hst_nilB U A i m (p q : tree U A) :
        hstep i p [::] m q <->
        ((exists B, exists a, exists g, exists p', exists v : B,
            [/\ p = Cons a g p', q = p' v & astep i a v m]) \/
         (exists B1, exists B2, exists v1 : B1, exists v2 : B2, exists p',
            [/\ m = i, p = Par (Ret v1) (Ret v2) p' & q = p' (v1,v2) ])).
Proof.
split.
- case/hstP; first by case=>_ []; [left | right].
  by case=>B1 [B2][p1][p2][p'] [_ []]; case=>pi [].
case; first by case=>B [a][g][r][v] [->->T]; apply: hst_consPQ T.
by case=>B1 [B2][v1][v2][r] [->->->].
Qed.

Lemma hst_consB U A i m b bs (p q : tree U A) :
        hstep i p (b :: bs) m q <->
        exists B1, exists B2, 
        exists p1 : tree U B1, exists p2 : tree U B2, exists p',
          p = Par p1 p2 p' /\
        ((exists p1',
            [/\ b = LL, q = Par p1' p2 p' & hstep i p1 bs m p1']) \/ 
         (exists p2',
            [/\ b = RR, q = Par p1 p2' p' & hstep i p2 bs m p2'])).
Proof.
split; last first.
- case=>B1 [B2][v1][v2][r] [->[]]; case=>pi' [->->];
  by [move/hst_lparP | move/hst_rparP].
case/hstP; first by case.
case=>B1 [B2][p1][p2][r] [->[]]; case=>pi' [[->]->T];
exists B1, B2, p1, p2, r; split=>//; [left|right]; 
by exists pi'.
Qed.

(* connection between stepping in the heap and stepping in isolation *)

Lemma hst_st U A i j (p q : tree U A) bs :
        hstep i p bs j q -> 
        exists B, exists a : action B, exists g, exists v : B, 
          [/\ (a, g) \In action_at p bs, 
              astep i a v j & step p [tr a, v] q].
Proof.
elim: bs A p q=>[|b bs IH] A p q; last first.
- case: b; case: p=>//= B1 B2 p1 p2 r [p'][->];
  case/IH=>B [a][g][v][H1] H2 H3; exists B, a, g, v; split=>//;
  by [apply: st_lpar H3 | apply: st_rpar H3].
case: p=>//; first by move=>B a g t [v][H1] ->; exists B, a, g, v.
move=>B1 B2 [] // v1 [] // v2 r [->] ->. 
by exists unit, Idle, id_annot, tt.
Qed.

(* The other direction doesn't hold as-is, for there's no guarantee that *)
(* the step p [tr a, v] q reduces the a action at exactly bs. Indeed,    *)
(* another a action may be taken; hence we must existentially abstract   *)
(* over bs.                                                              *)

Lemma st_hst U A B i j (a : action B) (v : B) (p q : tree U A) :
        astep i a v j -> step p [tr a, v] q -> 
        exists g, exists bs, 
          (a, g) \In action_at p bs /\ hstep i p bs j q.
Proof.
elim:p q=>{A} [A|A b|A B1 a1 g t _|A B1 B2 p1 IH1 p2 IH2 r _] q H.
- by case/st_dummyP.
- by case/st_retP.
- case/st_consP; rfejm2 a1 t; move/(_ (erefl _))=><-; exists g, [::];
  by split=>//; apply/hst_consP; exists v.
case/st_parP=>[|{IH1 IH2}[]]; last first.
- rewrite trD=>E; case:E H; rfejm2 a v=>E H [x][y][->][->] ->.
  by exists id_annot, [::].
case; case=>p [->]; [case/(IH1 _ H) | case/(IH2 _ H)]; 
move=>g [bs][T1 T2]; [exists g, (LL :: bs) | exists g, (RR :: bs)];
by split=>//; exists p.
Qed.

Lemma hst_actP U bs A (p : tree U A) C (a : action C) g v i m :
        (a, g) \In action_at p bs -> astep i a v m -> 
        exists q, hstep i p bs m q.
Proof.
elim: bs A p C a v i m=>[|bs ps IH] A p C a v i m.
- case/act_nilB=>[[q] ->|[B1][B2][v1][v2][q][_]];
  by [exists (q v), v | rfejm2 a v=>-> T; exists (q (v1, v2))].
case/act_consB=>B1 [B2][p1][p2][p'] [->].
case=>[][->] H; case/(IH _ _ _ _ _ _ _ H)=>q T;
by [exists (Par q p2 p'); apply: hst_lparP |
    exists (Par p1 q p'); apply: hst_rparP].
Qed.

(*******************************)
(* transitive closure of hstep *)
(*******************************)

Fixpoint hstept U A i (p : tree U A) ts m q := 
  if ts is bs::ts' then 
    exists m', exists q', hstep i p bs m' q' /\ hstept m' q' ts' m q
  else m = i /\ q = p.

Lemma hstt0_refl U A i (p : tree U A) : hstept i p [::] i p.
Proof. by []. Qed.

Hint Resolve hstt0_refl.

Lemma hstt0 U A i m (p q : tree U A) : 
        hstept i p [::] m q <-> i = m /\ p = q.
Proof. by split=>//; by case=>->->. Qed.

Lemma hstt1 U A bs i m (p q : tree U A) : 
        hstep i p bs m q <-> hstept i p [:: bs] m q.
Proof. 
split; first by exists m, q.
by case=>m' [q'][H][->] ->.
Qed.

Lemma hstt_catE U A i m (p q : tree U A) ts1 ts2 : 
        hstept i p (ts1 ++ ts2) m q <-> 
        exists m1, exists q1, 
          hstept i p ts1 m1 q1 /\ hstept m1 q1 ts2 m q.
Proof.
elim: ts1 i p=>[|t ts1 IH] i p /=.
- by split; [eauto | case=>m' [q'][[<- <-]]].
split; first by case=>m1 [q1][H1]; case/IH=>m2 [q2][H2]; eauto 10.
case=>m1 [q1][[m2]][q2][H1] H2 H3.
by exists m2, q2; rewrite IH; eauto.
Qed.

Lemma hstt_catE1 U A i m (p q : tree U A) bs ts : 
        hstept i p (bs :: ts) m q <-> 
        exists m1, exists q1, 
          hstep i p bs m1 q1 /\ hstept m1 q1 ts m q.
Proof. by move=>*; rewrite -cat1s. Qed.

Lemma hstt_catI U A i m h (p q r : tree U A) ts1 ts2 : 
        hstept i p ts1 h q -> hstept h q ts2 m r -> 
        hstept i p (ts1 ++ ts2) m r.
Proof. by move=>*; apply/hstt_catE; eauto. Qed.
    
Lemma hstt_catI1 U A i m h (p q r : tree U A) bs ts : 
        hstep i p bs h q -> hstept h q ts m r -> 
        hstept i p (bs :: ts) m r.
Proof. by move=>*; apply/hstt_catE1; eauto. Qed.
    
(* views for hstept *)

Lemma hstt_dummyP U A ts i m (p : tree U A) : 
        hstept i Dummy ts m p <-> [/\ ts = [::], m = i & p = Dummy].
Proof.
elim:ts i m p=>[|bs ts IH] i m p /=; first by split; case. 
by split; case=>// h [r][]; case: bs=>[//|b bs]; case: b.
Qed.

Lemma hstt_retP U A ts i m v (p : tree U A) : 
        hstept i (Ret v) ts m p <-> [/\ ts = [::], m = i & p = Ret v].
Proof.
elim:ts i m v p=>[|bs ts IH] /=; first by split; case. 
by split; case=>// h [r][]; case: bs=>[//|b bs]; case: b.
Qed.

Lemma hstt_consP U A ts B (a : action B) g i m p (q : tree U A) : 
        hstept i (Cons a g p) ts m q <-> 
        if ts is bs::ts' then 
          exists v, exists h, 
            [/\ bs = [::], astep i a v h & hstept h (p v) ts' m q]
        else m = i /\ q = Cons a g p.
Proof.
case: ts B a g i m p q=>[|bs ts] B a g i m p q //=.
split; last by case=>v [h][->]; eauto 10.
case=>m1 [q1][]; case/(hst_consP g)=>v [->->] H1 H2.
by exists v, m1.
Qed.

Lemma hstt_lparP U A B C ts h1 h2 (p1 p2 : tree U A) (q : tree U B) 
                                  (r : _ -> tree U C) :
        hstept h1 p1 ts h2 p2 -> 
        hstept h1 (Par p1 q r) (map (cons LL) ts) h2 (Par p2 q r).
Proof.
elim:ts h1 h2 p1 p2 q r=>[|bs ts IH] h1 h2 p1 p2 q r /=.
- by case=>->[->].
case=>m1 [q1][H1] H2; exists m1, (Par q1 q r).
by split; [exists q1 | apply: IH].
Qed.

Lemma hstt_rparP U A B C ts h1 h2 (q1 q2 : tree U B) (p : tree U A)
                                  (r : _ -> tree U C) :
        hstept h1 q1 ts h2 q2 -> 
        hstept h1 (Par p q1 r) (map (cons RR) ts) h2 (Par p q2 r).
Proof.
elim:ts h1 h2 q1 q2 p r=>[|bs ts IH] h1 h2 p1 p2 q r /=.
- by case=>->[->].
case=>m1 [q1][H1] H2; exists m1, (Par q q1 r).
by split; [exists q1 | apply: IH].
Qed.

Lemma hstt_lpar_retP U A B i m (v v' : A) (q q' : tree U B) ps :
        hstept i (Par (Ret v) q Ret) ps m (Par (Ret v') q' Ret)
        <-> [/\ v = v',
                all (fun bs => bs == (RR :: behead bs)) ps
              & hstept i q (map behead ps) m q'].
Proof.
elim:ps i q=>[|bs ps IH] i q.
- split; last by case=>->_ [->->].
  by case/hstt0=>->; case/parE=>_ _; case/jmE=>->; move/jmE=>-> _.
split; last first.
- case=>Ev; case/andP; move/eqP=>->E; case/hstt_catE1=>m'[p] [T T'].
  apply/hstt_catE1; exists m', (Par (Ret v) p Ret); split=>//.
  - by rewrite Ev; apply/hst_rparP.
  by apply/IH.
case/hstt_catE1=>m'[p] []; case/hst_parP; last first.
- by case=>v1[v2] [->->->[->]->]; case/hstt_retP.
case; case=>pi [->->]; first by move/hst_retP.
by move=>T; case/IH=>->E; move/(hstt_catI1 T)=>{T}T; split=>//; apply/andP.
Qed.

Lemma hstt_lpar_dummyP U i m ps A B p (q : tree U (A * B)) :
        hstept i (Par Dummy p Ret) ps m q ->
        exists p', exists ps', q = Par Dummy p' Ret /\ hstept i p ps' m p'.
Proof.
elim: ps i p=>[|bs ps IH] i p.
- by move/hstt0=>[<-<-]; exists p, [::].
case/hstt_catE1=>m'[p1] []; case/hst_lpar_dummyP=>p1'[bs'] [->T];
by case/(IH m' p1')=>p'[ps'] [->]; move/(hstt_catI1 T); exists p', (bs'::ps').
Qed.

Lemma hstt_lpar_dummy_retP U i m ps A (v : A) B (p q : tree U B) r :
        hstept i (Par (@Dummy U A) p Ret) ps m (Par (Ret v) q r) <-> False.
Proof.
elim:ps i p=>[//|bs ps IH] i p; split=>//.
- by case/hstt0=>_; case/parE=>_ _; move/jmE.
by case/hstt_catE1=>m'[q12'] []; case/hst_lpar_dummyP=>q' [bs'] [->_]; move/IH.
Qed.

(***************************************)
(* relating hstept and other steppings *)
(***************************************)

Lemma hstt_ind U A i (p : tree U A) (P : heap -> tree U A -> Prop) : 
        P i p ->
        (forall i p bs m q, P i p -> hstep i p bs m q -> P m q) -> 
        forall ps m q, hstept i p ps m q -> P m q.
Proof.
move=>H IH1 ps m q.
elim: ps i p H=>[|bs ps IH2] i p H /=; first by case=>->->.
by case=>m' [q'] [] Hs; apply: IH2; apply: IH1 H Hs.
Qed.

(*
Definition X_preserves 
  T (Step : forall A : Type, heap -> tree A -> T -> heap -> tree A -> Prop)
  i m A (pE qE : tree A) (ts : T)
  (P : heap -> tree A -> Prop) :=
  P i pE -> Step _ i pE ts m qE -> P m qE.
Definition hst_preserves := X_preserves hstep.
Definition hstt_preserves := X_preserves hstept.

Lemma hstt_closure i m A (pE qE : tree A) ps (P : heap -> tree A -> Prop)
        (step : forall i m pE qE bs, hst_preserves i m pE qE bs P) :
        hstt_preserves i m pE qE ps P.
Proof.
elim:ps i pE=>[|bs ps IH] i pE H; first by case/hstt0=><-<-.
by case/hstt_catE1=>i'[pE'] []; move/(step _ _ _ _ _ H)=>H'; move/(IH _ _ H').
Qed.

Definition X_preservesE
  S (Step : forall A : Type, heap -> tree A -> S -> heap -> tree A -> Prop)
  B i m (u : B) A (pE qE : tree A) (ts : S)
  (P : heap -> tree A -> B -> Prop) :=
  P i pE u -> Step _ i pE ts m qE -> exists v, P m qE v.
Definition hst_preservesE := X_preservesE hstep.
Definition hstt_preservesE := X_preservesE hstept.

Lemma hstt_closureE B i m (u : B) A (pE qE : tree A) ps (P : heap -> tree A -> B -> Prop)
        (step : forall i m u pE qE bs, hst_preservesE i m u pE qE bs P) :
        hstt_preservesE i m u pE qE ps P.
Proof.
have step' : forall i' m' (p' q' : tree A) bs,
  hst_preserves i' m' p' q' bs (fun i p => exists v, P i p v).
- move=>i' m' p' q' bs [u'] S T; apply:step S T.
move=>S T.
have S' : exists u, P i pE u; first by exists u.
by apply: hstt_closure step' S' T.
Qed.

Definition X_par_preserves
  T (Step : forall A : Type, heap -> tree A -> T -> heap -> tree A -> Prop)
  i A (p1 : tree A) E (pE : tree E)
  (P : heap -> tree E -> Prop) :=
  (* XXX could pass v1 to P *)
  forall ts m v1 qE, Step _ i (Par p1 pE Ret) ts m (Par (Ret v1) qE Ret) -> P m qE.
Definition hst_par_preserves := X_par_preserves hstep.
Definition hstt_par_preserves := X_par_preserves hstept.
*)

Lemma stutt_hst U A i (p q : tree U A) :
        stutt p q -> exists ps, hstept i p ps i q.
Proof.
case=>n; elim: n p=>[|n IH] p /=; first by  move=>->; exists [::].
case=>r []; case/(st_hst (ast_idleX i))=>g [bs][_].
move/hstt1=>H1; case/IH=>ps H2; exists ([:: bs] ++ ps).
by apply/hstt_catE; exists i, r.
Qed.

Lemma wst_hst U A B i j (a : action B) (v : B) (p q : tree U A) :
        astep i a v j -> wstep p [tr a, v] q -> 
        exists ps, hstept i p ps j q.
Proof.
move=>H [p' q']; case/(stutt_hst i)=>bs1 H1.
case/(st_hst H)=>g [bs2][_]; move/hstt1=>H2; case/(stutt_hst j)=>bs3 H3.
exists (bs1 ++ [:: bs2] ++ bs3); apply/hstt_catE; exists i, p'.
by split=>//; apply/hstt_catE; exists j, q'.
Qed.

Lemma wwst_hst U A B i j (a : action B) (v : B) (p q : tree U A) :
        astep i a v j -> wwstep p [tr a, v] q -> 
        exists ps, hstept i p ps j q.
Proof.
move=>H []; case=>[|[]]; first by apply: wst_hst H.
by rewrite trD=>E; case:E H; rfejm2 a v=>_ ->; move/stutt_hst.
Qed.

(**************************)
(* Reasoning about safety *)
(**************************)

(* definedness preservation for hstep *)
Lemma hstDA U A B (p q : tree U A) (a : action B) g bs i m : 
        def i -> (a, g) \In action_at p bs -> i \In asafe a -> 
        hstep i p bs m q -> def m.
Proof.
move=>D H1 S; case/hst_st=>B1 [b1][g1][v][H2] T _.
by case: {H1 H2}(actE H1 H2) T=>_; rbejm2 b1 v; move/(astD D S).
Qed.

(* progress *)
Lemma hstX U A B (p : tree U A) (a : action B) g bs i :
        def i -> (a, g) \In action_at p bs -> i \In asafe a -> 
        {m : heap & {q : tree U A | hstep i p bs m q}}.
Proof.
elim: bs A p=>[|b bs IH] A p D; last first.
- case: p; case: b=>// B1 B2 p1 p2 r /= H1; case/(IH _ _ D H1)=>m [q] H2; 
  exists m; [exists (Par q p2 r) | exists (Par p1 q r)]; 
  by exists q.
case: p=>//; last first.
- by move=>B1 B2 [] // v1 [] // v2 r; exists i, (r (v1, v2)).
move=>B1 a1 g1 t [/= _]; rfejm1 a; case/(astX D)=>m [v] H2;
by exists m; exists (t v), v.
Qed.

Lemma hst_frame U A B i j m (p q : tree U A) (a : action B) g bs :
        def (i :+ j) -> (a, g) \In action_at p bs -> i \In asafe a -> 
        hstep (i :+ j) p bs m q ->
        exists m', [/\ m = m' :+ j, def m & hstep i p bs m' q].
Proof.
elim: bs A p q=>[|b bs IH] A; last first.
- case: b; case=>//= B1 B2 p1 p2 t q D S H [r][->];
  by case/(IH _ _ _ D S H)=>m' [->] T; exists m'; split=>//; exists r.
case=>//; last first.
- by move=>B1 B2 [] // v1 [] // v2 t q D _ S [->] ->; exists i. 
move=>B1 b1 g1 t q D [/= _]; rfejm1 a=>S [v] [].
case/(ast_frame D S)=>m' [->] D' H ->.
by exists m'; split=>//; exists v.
Qed.

(* safety iterated over whole paths *)
(* the quantification over v:B guarantees B is inhabited for redex *)
Fixpoint path_safe U A (p : tree U A) (ps : path) : heap -> Prop :=
  if ps is bs :: ps' then 
     fun i => forall B (a : action B) g (v : B),
                (a, g) \In action_at p bs -> i \In asafe a /\ 
                forall m q, hstep i p bs m q -> path_safe q ps' m
  else fun i => True.

(* global safety -- for all paths *)
Definition safe U A (p : tree U A) := 
  [Pred i | forall ps, path_safe p ps i].

(* preservation of safety *)
Lemma hstS U A i m (p q : tree U A) bs :
        i \In safe p -> hstep i p bs m q -> m \In safe q. 
Proof.
move=>S H; case/hst_st: (H)=>B [a][g][v][T _ _] ps.
by case/(S (bs :: ps) B a g v): T=>_; apply.
Qed.

Implicit Arguments hstS [U A i m p q]. 

(* preservation of safety for hstept *)
Lemma hsttS ps U A i m (p q : tree U A) :
        i \In safe p -> hstept i p ps m q -> m \In safe q. 
Proof.
elim:ps A i m p q=>[|bs ps IH] A i m p q S /=; first by case=>->->.
by case=>m' [q'][]; move/(hstS _ S); apply: IH.
Qed.

Implicit Arguments hsttS [U A i m p q]. 

Lemma hstD U A (p q : tree U A) bs i m : 
        def i -> i \In safe p -> 
        hstep i p bs m q -> def m.
Proof.
move=>D S; case/hst_st=>B[a][g][v] [H T _];
by case:(S [::bs] B a g v H)=>{S} S; move:(astD D S T).
Qed.

(* definedness preservation for hstept *)
Lemma hsttD ps U A (p q : tree U A) i m :
        def i -> i \In safe p -> 
        hstept i p ps m q -> def m.
Proof.
elim:ps A p q i m=>[|bs ps IH] A p q i m D S; first by case/hstt0=><-.
case/hstt_catE1=>m'[q'] [T]; move:(hstD D S T) (hstS _ S T)=>D' S';
by move/(IH _ _ _ _ _  D' S').
Qed.

(* equivalent formulation of sfP with hstept *)
Lemma sfP i U A (p : tree U A) :
        i \In safe p <->
        forall ps m q bs B (a : action B) g (v : B), hstept i p ps m q ->
          (a, g) \In action_at q bs -> m \In asafe a.
Proof.
(* it's annoying to have to generalize with hstept for the converse,
   but I'm not sure of another way *)
have sfP' : forall ps i m A (p q : tree U A),
        (forall ps m q bs B (a : action B) g (v : B), hstept i p ps m q ->
          (a, g) \In action_at q bs -> m \In asafe a) ->
        hstept i p ps m q ->
        m \In safe q.
- move=>{A p i} ps i m A p q H T ps'.
  elim:ps' ps i m A p q H T=>//.
  move=>bs' ps' IH ps i m A p q H T B a g v Ha.
  split=>//.
  - by apply: (H ps m q bs' B a g v T Ha).
  by move=>m' q'; move/hstt1; move/(hstt_catI T); apply: IH.
split.
- move=>S ps; elim:ps p i S=>// [|bs' ps' IH] p i S m q bs B a g v.
  - by case=>->->; case/(S [:: bs] _ a g v).
  by case=>[m'][q'] [T T']; move/(IH _ _ (hstS _ S T) _ _ _ _ _ g v T').
by move=>H; apply: (sfP' _ _ _ _ _ _ H (hstt0_refl _ _)).
Qed.

(* safety monotonicity *)
Lemma sf_mono U A (p : tree U A) i j :
        def (i :+ j) -> i \In safe p -> i :+ j \In safe p.
Proof.
move=>D S ps. 
elim: ps A p i j D S=>[|bs ps IH] a p i j D S //=.
move=>B b g v T; case/(S [:: bs] B b g v): (T)=>H _.
split=>[|m q]; first by apply: asf_mono D H.
by case/(hst_frame D T H)=>m' [->] D'; move/(hstS _ S); apply: IH.
Qed.

(* frame *)
Lemma hst_frameS U A i j m bs (p q : tree U A) :
        def (i :+ j) -> i \In safe p -> hstep (i :+ j) p bs m q ->
        exists m', 
          [/\ m = m' :+ j, def m, m' \In safe q & hstep i p bs m' q].
Proof.
move=>D S1 H1.
case: (hst_st H1)=>B [a][g][v][T _ _].
case: (S1 [:: bs] B a g v T)=>S2 _.
case/(hst_frame D T S2): H1=>m' [->] D' H2.
by exists m'; split=>//; apply: hstS S1 H2.
Qed.

(* some helper lemmas to prove common safety subgoals *)

Lemma sf_dummyP U A i : i \In safe (@Dummy U A).
Proof. by move=>[//|][//|]; case. Qed.

Lemma sf_retP U A i (x : A) : i \In safe (Ret (U:=U) x).
Proof. by move=>[//|][//|]; case. Qed.

Hint Resolve sf_dummyP sf_retP.

Lemma sf_consP U A B i (a : action A) g (p : A -> tree U B) :
        i \In safe (Cons a g p) <->
        (forall w : A, i \In asafe a /\ 
           forall m v, astep i a v m -> m \In safe (p v)).
Proof.
split; last first.
- move=>H [//|bs ps]; case: bs=>[|[]] //= B1 b1 g1 v [/= _]; rfejm2 b1 v;
  by case: (H v)=>H1 H2; split=>// m q [w][H3] ->; apply: H2.
move=>S w; split=>[|m v].
- by case: (S [:: [::]] A a g w (act_consPQ _ _ g)).
move=>H ps; case: (S ([::]::ps) A a g w (act_consPQ _ _ g))=>_.
by apply; exists v.
Qed.

Lemma sf_retconsP (U : pcm) A i (a : action A) (g : annot U) :
        i \In asafe a -> i \In safe (Cons a g Ret).
Proof. by move=>S; apply/sf_consP. Qed.

Lemma sf_retlparP U A B i (x : A) (p : tree U B) :
        i \In safe p -> i \In safe (Par (Ret x) p Ret).
Proof.
move=>S ps; elim: ps p i S=>[//|bs ps IH] p i S.
case: bs=>[|b bs].
- case: p S=>//= b _ B1 a g v [/= _]; rfejm2 a v.
  by split=>//m q [->->]; apply: sf_retP.
case: b=>[|/= B1 a g v T]; first by case: bs=>//; case.
split=>[|m q]; first by case: (S [:: bs] B1 a g v T).
by case=>p' [->] H; apply: IH; apply: hstS S H.
Qed.

Lemma sf_retrparP U A B i (x : A) (p : tree U B) :
        i \In safe p -> i \In safe (Par p (Ret x) Ret).
Proof.
move=>S ps; elim: ps p i S=>[//|bs ps IH] p i S.
case: bs=>[|b bs].
- case: p S=>//= b _ B1 a g v [/= _]; rfejm2 a v. 
  by split=>// m q [->->]; apply: sf_retP.
case: b=>[/= B1 a g v T|]; last by case: bs=>//; case.
split=>[|m q]; first by case: (S [:: bs] B1 a g v T).
by case=>p' [->] H; apply: IH; apply: hstS S H.
Qed.

Lemma sf_lparP U A1 A2 B p1 p2 (r : A1 * A2 -> tree U B) i :
        i \In safe (Par p1 p2 r) -> i \In safe p1.
Proof.
move=>S ps; elim: ps A1 A2 B p1 p2 r i S=>[|bs ps IH];
move=>A1 A2 B p1 p2 r i S //= B1 a g v.
case: bs=>[|b bs] T; last first.
- case: b T; [pose bs' := LL::LL::bs | pose bs' := LL::RR::bs];
  case/(S [:: bs'] B1 a g v)=>S1 _; split=>[//|m q] H;
  apply: (IH _ _ _ q p2 r); apply: (hstS bs') S _;
  by exists q.
split=>[|m q]; first by case: (S [:: [:: LL]] B1 a g v T).
case: p1 S T=>//.
- move=>C c g' t S T [w][H] ->; apply: (IH _ _ _ (t w) p2 r).
  by apply: (hstS [:: LL]) S _=>/=; eauto.
move=>C1 C2 [] // v1 [] //= v2 t S _ [-> ->]. 
apply: (IH _ _ _ (t (v1, v2)) p2 r).
by apply: (hstS [:: LL]) S _ =>/=; eauto.
Qed.

Lemma sf_rparP U A1 A2 B p1 p2 (r : A1 * A2 -> tree U B) i :
        i \In safe (Par p1 p2 r) -> i \In safe p2.
Proof.
move=>S ps; elim: ps A1 A2 B p1 p2 r i S=>[|bs ps IH];
move=>A1 A2 B p1 p2 r i S //= B1 a g v.
case: bs=>[|b bs] T; last first.
- case: b T; [pose bs' := RR::LL::bs | pose bs' := RR::RR::bs];
  case/(S [:: bs'] B1 a g v)=>S1 _; split=>[//|m q] H;
  apply: (IH _ _ _ p1 q r); apply: (hstS bs') S _;
  by exists q.
split=>[|m q]; first by case: (S [:: [:: RR]] B1 a g v T).
case: p2 S T=>//.
- move=>C c g' t S T [w][H] ->; apply: (IH _ _ _ p1 (t w) r).
  by apply: (hstS [:: RR]) S _=>/=; eauto.
move=>C1 C2 [] // v1 [] //= v2 t S _ [-> ->]. 
apply: (IH _ _ _ p1 (t (v1, v2)) r).
by apply: (hstS [:: RR]) S _ =>/=; eauto.
Qed.

Lemma sf_tmapP U A B i (f : A -> B) (p : tree U A) :
        i \In safe p -> i \In safe (tmap f p).
Proof.
move=>S ps; elim: ps A p i f S=>[|bs ps IH] A p i f S //=.
case: bs=>[|b bs] B1 b1 g1 w1.
- case: p S=>//.
  - move=>C a g t; move/sf_consP=>S [/= _]; rfejm2 b1 w1; case: (S w1)=>H1 H2;
    by split=>[//|m q][v][H3] ->; apply: IH (H2 _ _ H3).
  move=>C1 C2 [] // v1 [] // v2 t S [/= _]; rfejm2 b1 w1;
  by split=>[//|m q][-> ->]; apply: IH; apply: (hstS [::]) S _.
case: b p S; case=>//= C1 C2 p1 p2 t S T; split.
- by case: (S [:: LL::bs] B1 b1 g1 w1 T).
- move=> m q [p1'][->] H; apply: (IH _ (Par p1' p2 t)).
  by apply: (hstS (LL::bs)) S _; exists p1'.
- by case: (S [:: RR::bs] B1 b1 g1 w1 T).
move=>m q [p2'][->] H; apply: (IH _ (Par p1 p2' t)).
by apply: (hstS (RR::bs)) S _; exists p2'.
Qed.

Lemma st_asf U A B (p q : tree U A) a (v : B) i :
        i \In safe p -> step p [tr a, v] q -> i \In asafe a.
Proof. by move=>S; case/st_act=>g [bs]; case/(S [:: bs] B a g v). Qed.

Lemma wst_asf U A B (p q : tree U A) a (v : B) i :
        i \In safe p -> wstep p [tr a, v] q -> i \In asafe a.
Proof.
move=>S [p1 q1].
by case/(stutt_hst i)=>ps; move/(hsttS _ S)=>{S} S; move/(st_asf S).
Qed.

Lemma wwst_asf U A B (p q : tree U A) a (v : B) i :
        i \In safe p -> wwstep p [tr a, v] q -> i \In asafe a.
Proof.
move=>S []; case=>[|[]]; last by case: a.
by apply: wst_asf S.
Qed.

(***************************)
(* Bisimulation and safety *)
(***************************)

(* first wbisim and actions *)
Lemma wbsf_act U A B (p1 p2 : tree U A) (a : action B) g (v : B) bs i :
        i \In safe p1 -> p1 ==~~ p2 -> 
        (a, g) \In action_at p2 bs -> i \In asafe a.
Proof.
move=>S H; case/(act_st v)=>q2.
by case/(stwP (wb_sym H))=>q1 [_]; apply: wwst_asf S.
Qed.

(* sbisim and actions is a consequence; probably not needed *)
Lemma sbsf_act U A B (p1 p2 : tree U A) (a : action B) g (v : B) bs i :
        i \In safe p1 -> p1 ==~ p2 -> 
        (a, g) \In action_at p2 bs -> i \In asafe a.
Proof. by move=>H; move/sbwb; apply: wbsf_act. Qed.

(* wbisim preserves safety *)
Lemma wbS U A (p q : tree U A) i : 
        i \In safe p -> p ==~~ q -> i \In safe q.
Proof.
move=>S H ps; elim: ps A p q i S H=>[|bs ps IH] A p q i S H //= B a g v T;
split=>[|m q1]; first by apply: (wbsf_act v S H T).
case/hst_st=>B1 [a1][g1][v1][]; case/(actE T)=>_; rbejm2 a1 v1=>{T} T;
case/(stwP (wb_sym H))=>p1 [H1] H2; apply: IH (wb_sym H1);
by case/(wwst_hst T): H2=>bs'; apply: hsttS S.
Qed.

(* and so does strong bisim *)
Lemma sbS U A (p q : tree U A) i : i \In safe p -> p ==~ q -> i \In safe q.
Proof. by move=>H; move/sbwb; apply: wbS H. Qed.

(*
(* XXX not sure if the following wb/sb lemmas will be necessary *)
Lemma sb_hst i m A (p p' q : tree A) bs :
  p ==~ p' ->
  hstep i p bs m q ->
  exists q', exists bs', q ==~ q' /\ hstep i p' bs' m q'.
Proof.
admit.
Qed.

Lemma sb_hstt i m A (p p' q : tree A) ps :
  p ==~ p' ->
  hstep i p ps m q ->
  exists q', exists ps', q ==~ q' /\ hstep i p' ps' m q'.
Proof.
admit.
Qed.

Lemma wb_hst i m A (p p' q : tree A) bs :
  p ==~~ p' ->
  hstep i p bs m q ->
  exists q', exists bs', q ==~~ q' /\ hstept i p' bs' m q'.
Proof.
admit.
Qed.

Lemma wb_hstt i m A (p p' q : tree A) ps :
  p ==~~ p' ->
  hstept i p ps m q ->
  exists q', exists ps', q ==~~ q' /\ hstept i p' ps' m q'.
Proof.
admit.
Qed.
*)

(***************************)
(* Concatenation vs. hstep *)
(***************************)

Lemma hst_tcat U A i j (p q : tree U A) B (r : A -> tree U B) bs :
        hstep i p bs j q -> hstep i (tcat p r) bs j (tcat q r).
Proof.
case:bs A p q r=>[|b bs] A p q r; last first.
- case/hst_consB=>B1[B2][p1][p2][p'] [->] []; case=>pi' [->->T];
  apply/hst_consB; exists B1, B2, p1, p2, (fun v => tcat (p' v) r);
  split=>//; [left|right]; 
  by exists pi'.
case/hst_nilB.
- case=>C [a][g][p'][v] [->->T]; apply/hst_nilB; left;
  by exists C, a, g, (fun v => tcat (p' v) r), v.
case=>B1 [B2][v1][v2][p'] [->->->]; apply/hst_nilB; right;
by exists B1, B2, v1, v2, (fun v => tcat (p' v) r).
Qed.

Lemma hstt_append U A B A' (p1 : tree U A) p2 (p3 : tree U A') 
                           (q1 q2 q3 : tree U B) ps1 ps2 v h1 h2 h3 :
        hstept h1 (Par p1 q1 Ret) ps1 h2 (Par (Ret v) q2 Ret) ->
        hstept h2 (Par (p2 v) q2 Ret) ps2 h3 (Par p3 q3 Ret) ->
        hstept h1 (Par (tcat p1 p2) q1 Ret) (ps1 ++ ps2) h3 (Par p3 q3 Ret).
Proof.
elim: ps1 h1 p1 q1=>[|bs ps IH] h1 p1 q1.
- by case=>-> []; rbp; rfp.
case/hstt_catE1=>h'[q'] []; case/hst_parP; last first.
- by case=>x[y] [->->->->->]; case/hstt_retP.
case; case=>r [->->Tr] T'; move/(IH _ _ _ T')=>T;
apply/hstt_catE1; exists h';
[exists (Par (tcat r p2) q1 Ret) | exists (Par (tcat p1 p2) r Ret)]; 
split=>//; by [apply: hst_lparP; apply: hst_tcat | apply: hst_rparP].
Qed.

Lemma tcatp_actP U A1 A2 (p1 : tree U A1) 
                         (p12 : tree U A2) x2 B (a : action B) g bs :
        p12 \In tcatp p1 x2 -> ~~ Retb p1 ->
        ((a, g) \In action_at p12 bs <-> (a, g) \In action_at p1 bs).
Proof.
by case/tcatpP=>
  [[->->]
  |[v1][->H]
  |[B'][a'][g'][p'] [->] [p''] [->]
  |[B1][B2][p1'][p2'][p'] [->] [p''] [->]].
Qed.

(* this could be more general to deal with ~~(active p1) *)
Lemma tcatp_hstP U A2 A1 i m (p1 : tree U A1) (p12 p12' : tree U A2) x2 bs :
        p12 \In tcatp p1 x2 -> hstep i p12 bs m p12' -> active p1 ->
        exists p1', p12' \In tcatp p1' x2 /\ hstep i p1 bs m p1'.
Proof.
case/tcatpP=>[|[v_]|[B'][a'][g'][p']
              |[B1'][B2'][p1'][p2'][p']] [->] // [p''] [->E].
- case/hst_consP=>v [->->T] _;
  by exists (p' v); split=>//; apply/hst_consP; exists v.
case/hst_parP; last by case=>v1[v2][->->->->->] _; exists (p' (v1, v2)).
case; case=>pi [->->T] _;
[exists (Par pi p2' p') | exists (Par p1' pi p')]; split;
by [apply: hst_lparP | apply/tcatp_parP; exists p''|
    apply: hst_rparP | apply/tcatp_parP; exists p''].
Qed.

Lemma tcatp_sfP U A1 i (p1 : tree U A1) A2 s2 (p12 : tree U A2) :
        i \In safe p1 ->
        (forall ps m v1 p2, p2 \In s2 v1 -> 
           hstept i p1 ps m (Ret v1) -> m \In safe p2) ->
        p12 \In tcatp p1 s2 ->
        i \In safe p12.
Proof.
move=>H1 H2 H3 ps.
elim: ps i A1 p1 A2 s2 p12 H1 H2 H3=>[//|bs ps IH] i A1 p1 A2 s2 p12 H1 H2 H3.
move=>B a g v H.
case T: (Retb p1).
- case: p1 T H1 H2 H3=>// w _ H1 H2; move/(H2 [::] i).
  by move/(_ (conj (erefl _) (erefl _)) (bs::ps) _ a g v H).
move: H; rewrite (tcatp_actP a g bs H3 (negbT T))=>H.
split; first by move: (H1 [:: bs]); case/(_ _ _ g v H).
case A: (active p1); last first.
- case: p1 T A H1 H2 H3 H=>// _ _ H1 _.
  by move/tcatp_XdummyP=>->; move/act_dummyP.
move=>m q H4; move: (tcatp_hstP H3 H4 A).
case=>p1' [H3'] T'.
apply: IH H3'; first by apply: hstS T'.
move=>ps1 m1 v1 p2 H5 Ts.
by apply: (H2 (bs::ps1)) H5 _; exists m, p1'.
Qed.

Lemma tcatp_hsttP U A1 A2 ps i m (p1 : tree U A1) (p12 p12' : tree U A2) x2:
        p12 \In tcatp p1 x2 -> hstept i p12 ps m p12' ->
    (* XXX could be strengthened with (~~ Retb p1') in first conjunct *)
    (exists p1', p12' \In tcatp p1' x2 /\ hstept i p1 ps m p1') \/
    exists ps1, exists ps2, exists m', exists v1, exists p2',
      [/\ ps = ps1 ++ ps2,
          p2' \In (x2 v1),
          hstept i p1 ps1 m' (Ret v1),
          hstept i p12 ps1 m' p2'
        & hstept m' p2' ps2 m p12'].
Proof.
elim:ps i m A2 A1 p1 p12 p12' x2=>[|bs' ps' IH] i m A2 A1 p1 p12 p12' x2 H.
- by case=>[->->]; left; exists p1.
case/tcatpP:H (H).
- by case=>->-> _; case/hstt_dummyP.
- case=>v1 [->H12] _ T12; right;
  by exists [::], (bs'::ps'), i, v1, p12.
- case=>B[a][g][p1''] [->][p12''] [->E] H12. 
  (* this is identical to Par case below *)
  case/hstt_catE1=>[m'][q'] [T12 T12'].
  case:(tcatp_hstP H12 T12 (erefl _))=>p1' [H12'] T1.
  case:(IH _ _ _ _ _ _ _ _ H12' T12').
  - by case=>p1_ [H12_]; move/(hstt_catI1 T1)=>T1x; left; exists p1_.
  case=>ps1[ps2][m''][v1][p2'] [-> H2]; move/(hstt_catI1 T1)=>T1x;
  move/(hstt_catI1 T12)=>T12x T_; right;
  by exists (bs'::ps1), ps2, m'', v1, p2'.
case=>B1[B2][t1'][t2'][p1''] [-> [p12''] [-> E]] H12;
(* this is identical to Cons case above *)
case/hstt_catE1=>[m'][q'] [T12 T12'];
case:(tcatp_hstP H12 T12 (erefl _))=>p1' [H12'] T1;
case:(IH _ _ _ _ _ _ _ _ H12' T12').
- by case=>p1_ [H12_]; move/(hstt_catI1 T1)=>T1x; left; exists p1_.
case=>ps1[ps2][m''][v1][p2'] [-> H2]; move/(hstt_catI1 T1)=>T1x;
move/(hstt_catI1 T12)=>T12x T_; right;
by exists (bs'::ps1), ps2, m'', v1, p2'.
Qed.

Lemma tcatp_act_lparP U A1 A2 B C (p1 : tree U A1) (p12 : tree U A2) 
                      x2 (a : action B) g bs (q : tree U C) :
        p12 \In tcatp p1 x2 -> ~~ Retb p1 ->
        ((a, g) \In action_at (Par p12 q Ret) bs <-> 
         (a, g) \In action_at (Par p1 q Ret) bs).
Proof.
move=>H R1; case: bs; last by case=>ps //; apply: tcatp_actP H R1.
split=>//; move=>Ha; case/act_retparP: Ha H R1=>v2 [v'][/= ->];
by rfejm1 a=>->-> //; case/tcatp_retP=>v1 [->].
Qed.

Lemma tcatp_hstt_lparP U A1 A2 C ps i m (p1 : tree U A1) (p12 : tree U A2)
                       (pE : tree U C) (r : tree U (A2 * C)) x2 :
        p12 \In tcatp p1 x2 -> hstept i (Par p12 pE Ret) ps m r ->
        (exists q1, exists q12, exists qE,
          [/\ r = Par q12 qE Ret, ~~ Retb q1,
              q12 \In tcatp q1 x2
            & hstept i (Par p1 pE Ret) ps m (Par q1 qE Ret)]) \/
        exists ps1, exists ps2, exists m', exists v1, exists p2, exists qE,
          [/\ ps = ps1 ++ ps2, p2 \In (x2 v1),
              hstept i (Par p1 pE Ret) ps1 m' (Par (Ret v1) qE Ret),
              hstept i (Par p12 pE Ret) ps1 m' (Par p2 qE Ret)
            & hstept m' (Par p2 qE Ret) ps2 m r].
Proof.
elim:ps i m A2 A1 p1 p12 C pE r x2=>[|bs' ps' IH] i m A2 A1 p1 p12 C pE r x2.
- case:(RetbP p1).
  - case=>v1 ->; move=>R1; case/hstt0=>-><-; right;
    by exists [::], [::], m, v1, p12, pE.
  move/RetbP=>Rx; move=>R1; case/hstt0=>-><-; left;
  by exists p1, p12, pE; split=>//.
- case: (RetbP p1).
  - case=>v1 -> R1 T2; right;
  by exists [::], (bs'::ps'), i, v1, p12, pE.
case: bs'=>[|[] bs'] Rx R1; case/hstt_catE1=>m'[r'] [].
- move=>T;
  case/hst_retparP:T R1 Rx=>v2[v] [->->->->]; case/tcatp_retP=>v1 [->_] [];
  by exists v1.
- case/hst_lparPQ=>p12' [->T12].
  (* sloppy subproof *)
  have H1 : active p1.
  - case: (hst_st T12)=>B [a][g][v] []; move/RetbP:Rx=>Rx;
    move/(tcatp_actP _ _ _ R1 Rx)=>Ha1; case/(hst_actP Ha1)=>p1';
    by case/hst_st=>B' [a'][g'][v'] [_ _]; move/st_actP.
  case: (tcatp_hstP R1 T12 H1)=>p1' [R1' T1'].
  case/(IH _ _ _ _ _ _ _ _ _ _ R1').
  - case=>p1''[p12''][q'] [->Rx' R1''];
    move/(hstt_catI1 (hst_lparP pE Ret T1'))=>{T1'}T1'.
    by left; exists p1'', p12'', q'.
  case=>ps1[ps2][m''][v1][p2'][q''] [->R2];
  move/(hstt_catI1 (hst_lparP pE Ret T1'))=>{T1'}T1';
  move/(hstt_catI1 (hst_lparP pE Ret T12))=>T12' T2'; right;
  by exists ((LL::bs')::ps1), ps2, m'', v1, p2', q''.
case/hst_rparPQ=>q0 [->Tq0]; case/(IH _ _ _ _ _ _ _ _ _ _ R1).
- case=>p1'[p12'][q'] [->Rx' R1']; move/(hstt_catI1 (hst_rparP _ Ret Tq0))=>T';
  by left; exists p1', p12', q'.
case=>ps1[ps2][m''][v1][p2'][q''] [->R1'];
move/(hstt_catI1 (hst_rparP _ Ret Tq0))=>T1';
move/(hstt_catI1 (hst_rparP _ Ret Tq0))=>T12' T2'; right;
by exists ((RR::bs')::ps1), ps2, m'', v1, p2', q''.
Qed.

(********************)
(* Prefix vs. hstep *)
(********************)

Prenex Implicits tpfx_retXP.

Lemma tpfx_action_at U A B (a : action B) g (p p' : tree U A) bs : 
         p << p' -> (a, g) \In action_at p bs -> (a, g) \In action_at p' bs.
Proof.
elim: p' p B a bs=>{A}[A|A v|A C c f t IH|A B1 B2 t1 IH1 t2 IH2 t _] p B a bs.
- by move/tpfx_XdummyP=>->.
- by move/tpfx_XretP; case=>->.
- move/tpfx_XconsP; move=>[->|[r'][->] H]; first by move/act_dummyP.
  by move/act_consP=>[->->] E; rewrite {C}E in c t IH r' H *; move/jmE=>->.
move/tpfx_XparP=>[->|[p1'][p2'][r'][->] H1 H2 H]; first by move/act_dummyP.
case/act_parP; last by case; case=>->; [apply: IH1 | apply: IH2].
case=>v1 [v2][->->] E; rewrite {B}E in a *; move/jmE=>-> E1 E2.
rewrite {p1'}E1 {p2'}E2 in H1 H2.
by move/tpfx_retXP: H1=>->; move/tpfx_retXP: H2=>->.
Qed.

Lemma tpfx_hstP U A bs i m (p p' q : tree U A) :
        p << p' -> hstep i p bs m q ->
        exists q', q << q' /\ hstep i p' bs m q'.
Proof.
elim: bs i m A p p' q=>[|b bs IH] i m A p p' q P T; move:T P.
- move/hst_nilB; case.
  - case=>B[a][g][r][v] [->->T];
    case/tpfx_consXP=>r' [->]; move/(_ v)=>P;
    by exists (r' v); split=>//; exists v.
  case=>B1[B2][v1][v2][r] [->->->];
  case/tpfx_parXP=>p1'[p2'][r'] [->];
  by do 2!move/tpfx_retXP=>->; move/(_ (v1, v2))=>P; exists (r' (v1, v2)).
case/hst_consB=>B1[B2][p1][p2][r] [->]; case; case=>pi [->->T];
case/tpfx_parXP=>p1'[p2'][r'] [-> P1 P2 P];
[case:(IH _ _ _ _ _ _ P1 T)=>pi' [T' Pi']; exists (Par pi' p2' r')
|case:(IH _ _ _ _ _ _ P2 T)=>pi' [T' Pi']; exists (Par p1' pi' r')];
split=>//; try apply/tpfx_parP=>//;
[by apply: hst_lparP|by apply: hst_rparP].
Qed.

Lemma tpfx_sfP U A i (p p' : tree U A) :
        p << p' -> i \In safe p' -> i \In safe p.
Proof.
move=>H S ps; elim: ps i p p' H S=>[//|bs ps IH] i p p' H S /=.
move=>B a g v; move/(tpfx_action_at H)=>H1. 
split=>[|m q]; first by case: (S [:: bs] _ a g v H1).
by case/(tpfx_hstP H)=>q' [H'] T'; apply: IH H' _; apply: hstS S T'.
Qed.

Lemma tpfx_hst_lparP U A bs i m (p p' q : tree U A) C (p2 p2' : tree U C) :
        p << p' -> hstep i (Par p p2 Ret) bs m (Par q p2' Ret) ->
        exists q', q << q' /\ hstep i (Par p' p2 Ret) bs m (Par q' p2' Ret).
Proof.
elim: bs i m A p p' q C p2 p2'=>[|b bs IH] i m A p p' q C p2 p2' P T;
move:T P; first by case/hst_retparP=>v1[v2] [].
case/hst_rlparP=>_ []; case=>->->T P.
- case:(tpfx_hstP P T)=>q' [P' T'];
  by exists q'; split=>//; apply: hst_lparP.
by exists p'; split=>//; apply: hst_rparP.
Qed.

Lemma tpfx_hsttP U A ps i m (p p' q : tree U A) :
        p << p' -> hstept i p ps m q -> 
        exists q', q << q' /\ hstept i p' ps m q'.
Proof.
elim: ps i m A p p' q=>[|bs ps IH] i m A p p' q P.
- by case=>->->; exists p'.
case/hstt_catE1=>j[r] []; case/(tpfx_hstP P)=>r' [Pr Tr];
by case/(IH _ _ _ _ _ _ Pr)=>q' [Pq]; move/(hstt_catI1 Tr)=>T'; exists q'.
Qed.

Lemma tpfx_hstt_lparP U A ps i m (p p' q : tree U A) C (p2 p2' : tree U C) :
        p << p' -> hstept i (Par p p2 Ret) ps m (Par q p2' Ret) ->
        exists q', q << q' /\ hstept i (Par p' p2 Ret) ps m (Par q' p2' Ret).
Proof.
elim: ps i m A p p' q C p2 p2'=>[|bs ps IH] i m A p p' q C p2 p2' P.
- by case=>-> []; do 2!move/inj_pair2=>->; exists p'.
case/hstt_catE1=>j[r] []; case/hst_parP.
- case; case=>pi [->->];
  [case/(tpfx_hstP P)=>{P} pi' [P T] | move=>T];
  case/(IH _ _ _ _ _ _ _ _ _ P)=>q' [P'];
  [pose T' := hst_lparP p2 Ret T | pose T' := hst_rparP p' Ret T];
  by move/(hstt_catI1 T')=>T''; exists q'.
by move=>T; case:T P=>v1[v2] [->->->->->]; move/tpfx_retXP=>->; case/hstt_retP.
Qed.

