Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq rels ordtype.
Require Import prelude domains heaps pcm resources8 proc9 step9 hstep9.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Section LockSafe.
Variables R : resource.

Notation U := (univ_of R).
Notation x := (lock_of R).
Notation shareable := (shareable x).
Notation procHeap := (procHeap U).
Notation mod_inv := (@mod_inv_of R).
Notation mod_inv_prec := (@mod_inv_prec R).
Notation Inv := (Inv (R:=R)).
(* Notation prec := (@prec_of R). *)
(* Notation func := (@func_of R). *)
Notation state := (state R).

Definition env_steps (s1 s2 : state) : Prop :=
  [/\ local s1 = local s2, 
      owns (mx (local s1)) -> gh (env s1) = gh (env s2) 
    & s2 \In coherent].

Lemma envs_refl s : s \In coherent -> env_steps s s.
Proof. by case. Qed.

Lemma envs_trans s1 s2 s3 :
        env_steps s1 s2 -> env_steps s2 s3 -> env_steps s1 s3.
Proof.
case: s1 s2 s3=>[sh1 lp1 ep1][sh2 lp2 ep2][sh3 lp3 ep3].
move=>[/= <- I1 C1][/= <- I2 C2].
by split=>//= O; rewrite I1 // I2.
Qed.

Lemma envs_frame sh1 sh2 lp ep1 ep2 h : 
        env_steps [sh1, lp \+ h, ep1] [sh2, lp \+ h, ep2] ->
        env_steps [sh1, lp, ep1 \+ h] [sh2, lp, ep2 \+ h].
Proof.
case=>/= _ O C; split=>//=; last by rewrite joinC; apply/coh_shuffle.
move=>O'; rewrite O // own_joinE ?O' //. 
by move/cohV: C; case/andP; move/validL.
Qed.

Lemma envs_sub (sh1 sh2 : shareable) lp ep1 ep2 s : 
        (owns (mx lp) -> gh ep1 = gh ep2) ->
        env_steps [sh1, lp, ep1] s ->
        env_steps [sh2, lp, ep2] s.
Proof. 
case: s=>sh lp' ep J1; case=>/= <-{lp'} J2 C.
by split=>//= O; rewrite -J1 // -J2.
Qed.

Lemma envsR (sh : shareable) gE1 gE2 hE gL hL inv : 
        [/\ def (sh :+ hL :+ hE), lockd sh, valid (gL \+ gE2),
            subheap inv hE & Inv (gL \+ gE2) inv] ->
        env_steps 
          [sh, [mgh false, gL, hL], [mgh true, gE1, hE]]
          [receive sh inv, [mgh false, gL, hL], 
                           [mgh false, gE2, hE :- inv]].
Proof. 
case=>D L V Sh I; split=>//; apply: cohRep.
by do !split=>//=; rewrite L.
Qed.

Lemma envsG (sh : shareable) gE1 gE2 hE gL hL : 
        [/\ def (sh :+ hL :+ hE) & valid (gL \+ gE2)] ->
        env_steps 
          [sh, [mgh false, gL, hL], [mgh false, gE1, hE]]
          [grant sh, [mgh false, gL, hL],
                     [mgh true, gE2, hE :+ payload sh]].
Proof. by case=>D V; split=>//; apply: cohGep. Qed.

(* XXX move to resources?.v *)

Definition case_step (f : annot U) gE (sh1 : shareable) lp1 sh2 lp2 :=
  match lockd sh1, lockd sh2 with
    true, false => 
      owns (mx lp1) /\ 
      exists inv, exists gL',
        [/\ sh2 = receive sh1 inv, 
            lp2 = [mgh false, gL', hp lp1 :- inv],
            assert f (gh lp1) = Some gL',
            (* gh lp1 \In assert f,  *)
            valid (gL' \+ gE), subheap inv lp1 & 
            Inv (gL' \+ gE)  inv]
  | false, true => 
      [/\  sh2 = grant sh1, 
           lp2 = [mgh true, gh lp1, hp lp1 :+ payload sh1],
           valid (gh lp1 \+ gE) & Inv (gh lp1 \+ gE) (payload sh1)]
  | _, _ => sh2 = sh1 /\ mg lp2 = mg lp1
  end.

CoInductive case_spec f gE (sh1 : shareable) 
                           (lp1 : procHeap) sh2 lp2 : Prop :=
| no_transfer of sh2 = sh1 /\ mg lp2 = mg lp1
| epsh_transfer inv gL' of
    sh2 = receive sh1 inv & 
    lp2 = [mgh false, gL', hp lp1 :- inv] &
    owns (mx lp1) & assert f (gh lp1) = Some gL' (* gh lp1 \In assert f *) & 
    valid (gL' \+ gE) & lockd sh1 & 
    subheap inv lp1 & Inv (gL' \+ gE) inv
| shep_transfer of
    lockd sh1 = false & sh2 = grant sh1 & 
    lp2 = [mgh true, gh lp1, hp lp1 :+ payload sh1] &
    valid (gh lp1 \+ gE) & Inv (gh lp1 \+ gE) (payload sh1).

Lemma csP f gE (sh1 sh2 : shareable) lp1 lp2 :
        case_step f gE sh1 lp1 sh2 lp2 <->
        case_spec f gE sh1 lp1 sh2 lp2.
Proof.
split; rewrite /case_step /=; last first.
- case; last 1 first.
  - by move=>->->-> V inv /=; rewrite !lockdG.
  - by case=>->->; case: ifP=>->.
  move=>inv gL' ->-> H An V -> Sh I; rewrite lockdR.
  by split=>//; exists inv, gL'.
case L1: (lockd sh1); case L2: (lockd sh2); last 1 first. 
- by apply: no_transfer.
- by apply: no_transfer.
- by case=>H [inv][gL'][->] -> An V; apply: epsh_transfer.
by case=>->-> V I; apply: shep_transfer.
Qed.

Lemma csP1 f gE sh lp1 lp2 : 
        mg lp1 = mg lp2 -> case_step f gE sh lp1 sh lp2.
Proof. by move=>H; apply/csP; apply: no_transfer. Qed.

(*
Lemma func gL gL1 gL2 gE inv1 inv2 lp :
        valid (gL \+ gE) ->
        subheap inv1 lp -> Inv (gL1 \+ gE) inv1 ->
        subheap inv2 lp -> Inv (gL2 \+ gE) inv2 ->
        gL1 = gL2 /\ inv1 = inv2.
Proof.
move=>V Sh1 [MI1] [rp1] [RI1] N1 CI1 Sh2 [MI2] [rp2] [RI2] N2 CI2.
rewrite -(mod_inv_prec Sh1 MI1 Sh2 MI2) in RI2 *.
rewrite -(repn_func (proj1 (subheap_def Sh1)) RI1 RI2) in N2 CI2.
(* have V : valid (gL \+ gE) by admit. (* XXX beef up resource defn *) *)
by rewrite -(next_func V N1 N2).
Qed.
*)

Lemma cs_func f gE (sh sh1 sh2 : shareable) lp lp1 lp2 : 
        valid (gh lp \+ gE) ->
        lockd sh1 = lockd sh2 ->
        case_step f gE sh lp sh1 lp1 -> 
        case_step f gE sh lp sh2 lp2 -> 
        sh1 = sh2 /\ mg lp1 = mg lp2.
Proof.
rewrite /case_step => V <-.
case L: (lockd sh); case L1: (lockd sh1); last 1 first.
- by case=>->-> [->] ->.
- by case=>->-> [->] ->.
- case=>O [inv1][gL1][->] -> -> V1 Sh1 I1 [_][inv2][gL2][->] -> [<-] V2 Sh2 I2.
  by rewrite (mod_inv_prec Sh1 I1 Sh2 I2).
by case=>->-> _ _ [->] -> _ _.
Qed.

Lemma csK f sh sh1 sh2 ep lp lp1 lp2 : 
        valid (gh lp \+ gh ep) ->
        [sh1, lp1, ep] \In coherent ->
        case_step f (gh ep) sh lp sh1 lp1 ->
        case_step f (gh ep) sh lp sh2 lp2 ->
        sh1 :+ lp1 :+ ep = sh2 :+ lp2 :+ ep ->
        sh1 = sh2 /\ lp1 = lp2.
Proof.
move=>V C Cs1 Cs2 E; move: (coh_lockd C E)=>L.
case: (cs_func V L Cs1 Cs2)=>E1 [E2 E3]; rewrite -E1 in E *.
move/cohD: C E=>/= D; rewrite -!(unAC _ ep) in D *; move/(unhKl D)=>E4.
by split=>//; apply/mghE.
Qed.

Lemma cs_coh f (sh1 sh2 : shareable) ep lp1 lp2 : 
        [sh1, lp1, ep] \In coherent ->
        case_step f (gh ep) sh1 lp1 sh2 lp2 ->
        def (sh2 :+ lp2 :+ ep) ->
        [sh2, lp2, ep] \In coherent.
Proof.
move=>C; move/cohD: (C)=>/= D; case/csP.
- by case=>-> E D'; apply: coh_sub C.
- move=>inv gL ->-> H An V L Sh I D'; apply: cohRlp.
  by move: (coh_ownsL C); rewrite H.
move=>L ->-> V I D'; apply: cohGlp.
by move: (coh_lkd_owns C); rewrite L /=; case/andP=>_ ->.
Qed.

Lemma cs_envs f (sh1 sh2 : shareable) lp ep1 ep2 :
        [sh1, lp, ep1] \In coherent ->
        case_step f (gh lp) sh1 ep1 sh2 ep2 ->
        def (sh2 :+ lp :+ ep2) ->
        env_steps [sh1, lp, ep1] [sh2, lp, ep2].
Proof.
move=>C1 Cs D.
have C2 : [sh2, lp, ep2] \In coherent.
- move/cohC: C1=>C1; rewrite unAC in D. 
  by apply/cohC; apply: cs_coh C1 Cs D.
move: Cs; rewrite /case_step /env_steps /=.
case L1: (lockd sh1); case L2: (lockd sh2)=>/=; last 1 first.
- by case=>E [E1 ->].
- by case=>E [E1 ->].
- case=>H [inv][gL][E1] E2 V Sh I; split=>//.
  by move: (coh_ownsE C1); rewrite H /=; move/eqP=>->.
case=>E1 E2 V I; split=>//.
move: (coh_lkd_owns C1); rewrite L1. 
by case/andP; case/eqP=>->.
Qed.

Definition annot_at A (p : tree U A) bs : Pred (annot U) :=
  [fun f => exists B, exists b : action B, (b, f) \In action_at p bs].

Lemma annotK A (p : tree U A) bs f1 f2 : 
        f1 \In annot_at p bs -> f2 \In annot_at p bs -> f1 = f2.
Proof. by case=>A1 [a1] Ac1 [A2][a2] Ac2; case: (actE Ac1 Ac2). Qed.

Lemma tcatp_annotP A1 A2 (p1 : tree U A1) 
                         (p12 : tree U A2) x2 f bs :
        p12 \In tcatp p1 x2 -> active p1 ->
        (f \In annot_at p12 bs <-> f \In annot_at p1 bs).
Proof.
move=>T A; have L: ~~ Retb p1 by case: p1 A {T}.
by split; case=>B [b] Ac; exists B, b; apply/(tcatp_actP b f bs T L).
Qed.

Lemma tpfx_annotP A f (p1 p2 : tree U A) bs : 
         p1 << p2 -> f \In annot_at p1 bs -> f \In annot_at p2 bs.
Proof. by move=>H [B][b] Ac; exists B, b; apply: tpfx_action_at Ac. Qed.

Definition local_step A (p : tree U A) bs s h2 (P : state -> Prop) :=
  let: [sh1, lp1, ep] := s in 
  exists sh2 : shareable, exists lp2 : procHeap, exists f,
    [/\ h2 = sh2 :+ lp2 :+ ep, f \In annot_at p bs,
        case_step f (gh ep) sh1 lp1 sh2 lp2 
      & P [sh2, lp2, ep]].

(* Definition of modal operators for expressing prototol safety.  The
environment process pE in the initial heap i respects the locking
protocol over x iff the following holds, wrt the postcondition True. *)

Fixpoint always_path A s1 p ps (P : state -> tree U A -> Prop) :=
  if ps is bs :: qs then 
    s1 \In coherent /\
    forall s2, env_steps s1 s2 -> 
      [/\ shared s2 :+ local s2 \In safe p, P s2 p &
          forall h3 q, hstep (heap_of s2) p bs h3 q -> 
          local_step p bs s2 h3 (fun s3 => always_path s3 q qs P)]
  else 
    s1 \In coherent /\ 
    forall s2, env_steps s1 s2 -> 
      shared s2 :+ local s2 \In safe p /\ P s2 p.

(* unfolding of the fixed point in always_path *)
(* and pulling out the common parts of the branches *)

Lemma alwpE A s1 (p : tree U A) ps P :
        always_path s1 p ps P <->
        s1 \In coherent /\
        forall s2, env_steps s1 s2 ->
          [/\ shared s2 :+ local s2 \In safe p, P s2 p &
          if ps is bs :: qs then 
             forall h3 q, 
               hstep (heap_of s2) p bs h3 q -> 
               local_step p bs s2 h3 (fun s3 => always_path s3 q qs P)
          else True].
Proof. 
split=>[|[C]]; case: ps=>[|bs ps] //=.
- by case=>C H; split=>// s2; case/H.
by move=>H; split=>// pt2; case/H.
Qed.

Definition always A s (p : tree U A) P := 
  forall ps, always_path s p ps P.

(* some structural properties of always_path and always *)

Lemma alwp_coh A s (p : tree U A) ps P : 
        always_path s p ps P -> s \In coherent.
Proof. by case/alwpE. Qed.

Lemma alw_coh A s (p : tree U A) P : 
        always s p P -> s \In coherent.
Proof. by move/(_ [::]); move/alwp_coh. Qed.

Lemma alwpD A s (p : tree U A) ps P : 
        always_path s p ps P -> def (heap_of s).
Proof. by move/alwp_coh/cohD. Qed.

Lemma alwD A s (p : tree U A) P : always s p P -> def (heap_of s).
Proof. by move/(_ [::]); move/alwpD. Qed.

Lemma alwp_dom A s (p : tree U A) ps P : 
        always_path s p ps P -> 
        [/\ def (heap_of s), x \in dom (shared s) & s \In coherent].
Proof. by move/alwp_coh=>C; case: (C)=>/= ->; rewrite shr_ldom. Qed.

Lemma alw_dom A s (p : tree U A) P : 
        always s p P -> 
        [/\ def (heap_of s), x \in dom (shared s) & s \In coherent].
Proof. by move/(_ [::]); move/alwp_dom. Qed.

Lemma alwp_safe A s (p : tree U A) ps P : 
        always_path s p ps P -> shared s :+ local s \In safe p.
Proof. by case/alwpE=>C T; case/T: (envs_refl C). Qed.

Lemma alw_safe A s (p : tree U A) P :
        always s p P -> shared s :+ local s \In safe p.
Proof. by move/(_ [::])/alwp_safe. Qed.

Lemma alwp_refl A s (p : tree U A) ps P : 
        always_path s p ps P -> P s p.
Proof. by case/alwpE=>C T; case/T: (envs_refl C). Qed.

Lemma alw_refl A s (p : tree U A) P :
        always s p P -> P s p.
Proof. by move/(_ [::])/alwp_refl. Qed.

Lemma alwp_envs A s1 s2 (p : tree U A) ps P : 
        always_path s1 p ps P ->
        env_steps s1 s2 ->
        always_path s2 p ps P.
Proof.
case/alwpE=>C T L; apply/alwpE; split; first by case: L.
by move=>s3; move/(envs_trans L)/T.
Qed.
 
Lemma alw_envs A s1 s2 (p : tree U A) P :
        always s1 p P ->
        env_steps s1 s2 ->
        always s2 p P.
Proof. by move=>S L ps; apply: alwp_envs (S ps) L. Qed.

Lemma alw_sub A sh1 sh2 ep1 ep2 lp (p : tree U A) P : 
        [sh2, lp, ep2] \In coherent -> 
        (owns (mx lp) -> gh ep1 = gh ep2) ->
        always [sh1, lp, ep1] p P ->
        always [sh2, lp, ep2] p P.
Proof.
move=>C J L ps; move/(_ ps): L; case/alwpE=>C' T'; apply/alwpE.
by split=>// s2 T; apply: T'; apply: envs_sub T; move/J.
Qed.

(* specialized versions of alw_sub when sh *)
(* and environment ghosts are fixed *)

Lemma alw_sub_def A (sh : shareable) (ep1 ep2 lp : procHeap) 
                  (p : tree U A) P :
        [/\ def (sh :+ lp :+ ep2) & mg ep1 = mg ep2] ->
        always [sh, lp, ep1] p P ->
        always [sh, lp, ep2] p P.
Proof. 
case=>D [E1 E2] L; apply: alw_sub (L)=>//.
by case/alw_coh: L; rewrite E1 E2.
Qed.

Lemma alw_sub_subheap A sh (ep1 ep2 lp : procHeap) (p : tree U A) P :
        [/\ subheap ep2 ep1 & mg ep1 = mg ep2] ->
        always [sh, lp, ep1] p P ->
        always [sh, lp, ep2] p P.
Proof.
case=>Sh E L; apply: alw_sub_def (L); move/alwD: L=>/= D.
by split=>//; case/subheapX: Sh D=>-> _; defcheck.
Qed.

Lemma alw_hstep A sh lp ep (p q : tree U A) m bs P :
        always [sh, lp, ep] p P ->
        hstep (sh :+ lp :+ ep) p bs m q ->
        local_step p bs [sh, lp, ep] m (fun s2 => always s2 q P).
Proof.
move=>Ls Ts; move: (Ls [:: bs]) (Ts)=>/= [C].
case/(_ _ (envs_refl C))=>_ _; case/(_ _ _ Ts)=>{Ts} [sh1][ep1][f1][->].
move=>An1 Cs1 [C1] _ Ts.
exists sh1, ep1, f1; split=>// ps; move: (Ls (bs :: ps))=>/= [_].
case/(_ _ (envs_refl C))=>_ _; case/(_ _ _ Ts)=>sh2 [ep2][f2][E] An2 Cs2 Ls'.
move/alwp_coh: (Ls')=>C2; case/(annotK An1): An2 Cs2=><-{f2} Cs2. 
by case/(csK (cohVgh C) C1 Cs1 Cs2): E=>->->.
Qed.

(*
Lemma alw_hstep_cs A (sh1 sh2 : shareable) (lp1 lp2 : procHeap) ep 
                   (p1 p2 : tree U A) bs P : 
        always [sh1, lp1, ep] p1 P ->
        hstep (sh1 :+ lp1 :+ ep) p1 bs (sh2 :+ lp2 :+ ep) p2 ->
        
        case_step f (gh ep) sh1 lp1 sh2 lp2 ->
        always [sh2, lp2, ep] p2 P.
Proof.
move=>Ls1; case/(alw_hstep Ls1)=>sh3 [lp3][B][a][g]. [E] Cs3 Ls2 Cs2.
by move/sym: E; case/(csK (cohVgh (alw_coh Ls1)) (alw_coh Ls2) Cs3 Cs2)=><-<-.
Qed.
*)

Lemma alw_hstep_alw A sh lp ep (p q : tree U A) m bs P :
        always [sh, lp, ep] p P ->
        hstep (sh :+ lp :+ ep) p bs m q ->
        exists s, m = heap_of s /\ always s q P.
Proof. 
by move=>L; case/(alw_hstep L)=>sh1 [lp1][f][->]; exists [sh1, lp1, ep].
Qed.

Lemma alw_hstept_alw A sh lp ep (p q : tree U A) m ps P :
        always [sh, lp, ep] p P ->
        hstept (sh :+ lp :+ ep) p ps m q ->
        exists s, m = heap_of s /\ always s q P. 
Proof.
move=>Ls. 
set P' := fun m q => exists s, m = heap_of s /\ always s q P.
apply: (hstt_ind (P := P')); first by exists [sh, lp, ep].
by move=>i r bs j t [[rs' hL' hE']][->]; apply: alw_hstep_alw.
Qed.

Lemma alw_frame A (sh : shareable) (lp ep h : procHeap) (p : tree U A) 
                   (Q : state -> tree U A -> Prop) : 
        always [sh, lp, ep \+ h] p Q ->
        always [sh, lp \+ h, ep] p 
          (fun s p => exists lp', local s = lp' \+ h /\
                        Q [shared s, lp', env s \+ h] p).
Proof.
move=>Ls ps.
elim: ps sh lp ep h p Ls=>[|bs ps IH] sh lp ep h p Ls /=.
- split; first by apply/coh_shuffle; rewrite joinC; apply: alw_coh Ls.
  case=>sh2 lp2 ep2 [/= <-] J C2. 
  have E: env_steps [sh, lp, ep \+ h] [sh2, lp, ep2 \+ h].
  - by apply: envs_frame.
  have {Ls} Ls: always [sh2, lp, ep2 \+ h] p Q.
  - by apply: alw_envs Ls E.
  have D : def (sh2 :+ lp :+ ep2 :+ h).
  - by move/alwD: Ls; defcheck. 
  split; last by exists lp; split=>//; apply: alw_refl.
  by rewrite unA; apply: sf_mono (alw_safe Ls); move: D; defcheck. 
split; first by apply/coh_shuffle; rewrite joinC; apply: alw_coh Ls.
case=>sh2 lp2 ep2 E; case: E (E)=>/= <-{lp2} J C2; move/envs_frame=>E.
have {Ls} Ls : always [sh2, lp, ep2 \+ h] p Q by apply: alw_envs Ls E.
have D : def (sh2 :+ lp :+ ep2 :+ h) by move/alwD: Ls; defcheck. 
split.
- by rewrite unA; apply: sf_mono (alw_safe Ls); move: D; defcheck. 
- by exists lp; split=>//; apply: alw_refl.
move=>h3 q T.
have Sf : sh2 :+ lp :+ ep2 \In safe p.
- by apply: sf_mono (alw_safe Ls); move: D; defcheck. 
rewrite unA unAC in T.
case/(hst_frameS D Sf): T (T)=>m3 [->{h3}] D3 Sf3 _.
rewrite -unA.
case/(alw_hstep Ls)=>sh3 [lp3][f][E''] An3 Cs3 Ls3.
exists sh3, (lp3 \+ h), f. 
split=>//; last 1 first.
- by apply: IH=>//; rewrite -mgh_eta.
- by rewrite unA unAC -unA. 
case/csP: Cs3=>/=.
- by case=>-> [E1 E2]; apply/csP1; rewrite /= E1 E2.
- move=>inv gL ->-> /= O An V L2 Sh I.
  rewrite /case_step L2 lockdR /= own_joinE; last first. 
  - by move/cohV: C2; move/validL; case/andP. 
  rewrite O; split=>//; exists inv, (gL \+ gh h). 
  rewrite (mgh_eta h) mgh_join /=.  
  move/coh_shuffle/coh_ownsL: (C2). 
  rewrite O owno_join /=; case/andP=>O1 _; rewrite (eqP O1); split=>//.
  - apply/mghE=>/=; split=>//; rewrite subUnC //.
    by case/subheapX: Sh D=>-> _; defcheck.
  - apply: assertP=>//.
    by move/cohV: C2; case/andP=>_; move/validL. 
  - by rewrite -joinA (joinC (gh h)). 
  - by apply: subheap_trans Sh _; apply: subheapUnl; move: D; defcheck.
  by rewrite -joinA (joinC (gh h)). 
move=>L ->-> /= V Sh; rewrite /case_step L lockdG unAC. 
rewrite (mgh_eta h) mgh_join /=. 
move/coh_lkd_owns: (C2); rewrite L owno_join; do 2![case/andP]=>_ O _.
by rewrite (eqP O); split=>//; rewrite -joinA (joinC (gh h)). 
Qed.

Lemma alwp_dummyP A s ps (P : _ -> _ -> Prop) :
        s \In coherent -> 
        (forall s2, env_steps s s2 -> P s2 (@Dummy U A)) -> 
        always_path s Dummy ps P.
Proof.
by case: ps=>[|bs ps] C H; split=>// s2; move/H=>//; case: bs=>[|[]].
Qed.

Lemma alw_dummyP A s (P : _ -> _ -> Prop) :
        s \In coherent -> 
        (forall s2, env_steps s s2 -> P s2 (@Dummy U A)) -> 
        always s Dummy P.
Proof. by move=>C H ps; apply: alwp_dummyP. Qed.

Lemma alwp_retP A s (v : A) ps (P : _ -> _ -> Prop) : 
        s \In coherent -> 
        (forall s2, env_steps s s2 -> P s2 (Ret v)) -> 
        always_path s (Ret v) ps P.
Proof. 
by case: ps=>[|bs ps] C H; split=>// s2; move/H=>//; case: bs=>[|[]]. 
Qed.

Lemma alw_retP A s (v : A) (P : _ -> _ -> Prop) : 
        s \In coherent -> 
        (forall s2, env_steps s s2 -> P s2 (Ret v)) -> 
        always s (Ret v) P.
Proof. by move=>C H ps; apply: alwp_retP. Qed.

(* special case of always_path and always, when P = True *)

Notation lock_safe_path s p ps := (always_path s p ps (fun _ _ => True)).
Notation lock_safe s p := (always s p (fun _ _ => True)).

Lemma lsfp_dummyP A s ps : 
        s \In coherent -> lock_safe_path s (@Dummy U A) ps.
Proof. by move=>C; apply: alwp_dummyP. Qed.

Lemma lsf_dummyP A s : 
        s \In coherent -> lock_safe s (@Dummy U A).
Proof. by move/lsfp_dummyP; apply. Qed.

Lemma lsfp_retP A s (v : A) ps : 
        s \In coherent -> lock_safe_path s (Ret v) ps.
Proof. by move=>C; apply: alwp_retP. Qed.

Lemma lsf_retP A s (v : A) : 
        s \In coherent -> lock_safe s (Ret v).
Proof. by move=>C ps; apply: lsfp_retP. Qed.

(* always is a transitive operator *)

Lemma alwT A s1 (p1 : tree U A) P :
        always s1 p1 (fun s2 p2 => always s2 p2 P) <-> always s1 p1 P.
Proof.
split=>[|H1 ps]; first by move/alw_refl.
elim: ps p1 s1 H1 => [| bs ps IH] p1 s1 H1 /=.
- split=>[|s2]; first by apply: alw_coh H1.
  move/(alw_envs H1)=>H2; split=>//.
  by move/alw_safe: H2.
split=>[|s2]; first by apply: alw_coh H1.
move/(alw_envs H1)=>H2; split=>//.
- by move/alw_safe: H2.
case: s2 H2=>sh2 lp2 ep2 H2 h3 q /=.
case/(alw_hstep H2)=>sh3 [ep3][f][->] An Cs H3.
by exists sh3, ep3, f; split=>//; apply: IH H3.
Qed.

(* The above lemma is more concisely stated as : 

   always (always P) =p always P

where =p is iff-and-only-iff on relations over (partition * tree A).

But for this, we will need to rearrange the types, which I don't want
to do now.  *)

(* always is a natural operator *)
(* what's the official name for this in modal logic? K? *)

Lemma alwp_imp A s (p : tree _ A) ps (P1 P2 : _ -> _ -> Prop) : 
         (forall s p, s \In coherent -> 
                      shared s :+ local s \In safe p ->
                      P1 s p -> P2 s p) -> 
         always_path s p ps P1 -> always_path s p ps P2.
Proof.
elim: ps s p=>[|bs ps IH] s p /= I.
- case=>C L; split=>// s2 Envs; case/L: (Envs)=>S Q; split=>//.
  by apply: I S Q; case: Envs.
case=>C L; split=>// s2 Envs; case/L: (Envs)=>S Q T; split=>//.
- by apply: I S Q; case: Envs.
move=>h3 q; move/T; case: s2 {Envs S Q T}=>sh2 lp2 ep2 /=.
case=>sh3 [ep3][f3][->] An Cs H.
by exists sh3, ep3, f3; split=>//; apply: IH.
Qed.

Lemma alw_imp A s (p : tree _ A) (P1 P2 : _ -> _ -> Prop) : 
        (forall s p, s \In coherent -> 
                     shared s :+ local s \In safe p ->
                     P1 s p -> P2 s p) -> 
        always s p P1 -> always s p P2.
Proof. by move=>I H ps; apply: alwp_imp I _. Qed.

(* always commutes with universals *)

Lemma alwpA A B s (p : tree _ A) ps (P : B -> _ -> _ -> Prop) : 
        lock_safe_path s p ps ->
        (always_path s p ps (fun s' p' => forall x, P x s' p') <->
         forall x, always_path s p ps (fun s' p' => P x s' p')).
Proof.
move=>Ls; split. 
- elim: ps s p {Ls}=>[|bs ps IH] s p /=.
  - by case=>C Envs x; split=>// s2; case/Envs. 
  case=>C Envs x; split=>//.
  case=>sh lp ep; case/Envs=>/= S H L; split=>//.
  move=>h3 q; case/L=>sh2 [lp2][f2][->] An Cs Ls'. 
  by exists sh2, lp2, f2; split=>//; apply: IH Ls' _.
elim: ps s p Ls=>[|bs ps IH] s p /=.
- case=>C Envs Ha; split=>// s2 Env2; case/Envs:(Env2)=>S _.
  by split=>// x; case: (Ha x)=> _; case/(_ _ Env2).
case=>C Envs Ha; split=>//; case=>sh lp ep Env2.
case/Envs: (Env2)=>/= S _ L; split=>//.
- by move=>x; case: (Ha x)=>_; move/(_ _ Env2); case.
move=>h3 q Ts; case/L: Ts (Ts)=>sh2 [lp2][f2][->] An Cs Ls' Ts.
exists sh2, lp2, f2; split=>//; apply: IH (Ls') _.
move=>x; case: (Ha x)=>_; case/(_ _ Env2)=>_ _.
case/(_ _ _ Ts)=>sh3 [lp3][f3][E] An3 Cs3.
rewrite -(annotK An An3) in Cs3.
case: Env2=>_ _; move/cohVgh=>V.
by case: (csK V (alwp_coh Ls') Cs Cs3 E)=>->->.
Qed.

Lemma alwA A B s (p : tree _ A) (P : B -> _ -> _ -> Prop) : 
        lock_safe s p ->
        (always s p (fun s' p' => forall x, P x s' p') <->
         forall x, always s p (fun s' p' => P x s' p')).
Proof.
move=>Ls; split.
- by move=>H x ps; move: x; apply/(alwpA _ (Ls ps)).
by move=>H ps; apply/(alwpA _ (Ls ps))=>x; apply: H.
Qed. 

Implicit Arguments alwpA [A B s p ps P].
Implicit Arguments alwA [A B s p P].

(**************************************)
(* Always with sequential composition *)
(**************************************)

Lemma alw_bnd A B (p1 : tree _ A) (p12 : tree _ B) pp2 s1 P :
         p12 \In tcatp p1 pp2 -> 
         always s1 p1 (fun s2 p2 => 
           forall p12, p12 \In tcatp p2 pp2 -> 
                       always s2 p12 P) ->
         always s1 p12 P.
Proof.
case: s1=>sh1 lp1 ep1 T Ha ps.
elim: ps p12 sh1 lp1 ep1 p1 T Ha=>[|bs ps IH] p12 sh1 lp1 ep1 p1 T Ha /=.
- split=>[|s2]; first by apply: alw_coh Ha.
  move/(alw_envs Ha)/alw_refl/(_ _ T)=>H.
  by split; [move/alw_safe: H | move/alw_refl: H].
split=>[|[sh2 lp2 ep2] /= Envs]; first by apply: alw_coh Ha.
split=>[||h p12' Hs].
- by move/(alw_envs Ha): Envs; move/alw_refl/(_ _ T)/alw_safe.
- by move/(alw_envs Ha): Envs; move/alw_refl/(_ _ T)/alw_refl.
case Ac: (active p1); last first.
- case: p1 Ac T Hs Ha=>//.
  - by move=>_; move/tcatp_XdummyP=>->; move/hst_dummyP.
  move=>a _; move/tcatp_XretP=>T Hs.
  by case/alw_refl/(_ _ T (bs::ps))/proj2/(_ _ Envs)=>_ _; apply.
move/(alw_envs Ha): Envs=>{Ha} Ha.
case: (tcatp_hstP T Hs Ac)=>p1' [T'].
case/(alw_hstep Ha)=>sh3 [ep3][f3][->].
rewrite -(tcatp_annotP f3 bs T Ac)=>An Cs Ha'.
by exists sh3, ep3, f3; split=>//; apply: IH T' Ha'.
Qed.

Lemma alw_bnd_safe A B (p1 : tree _ A) (p12 : tree _ B) pp2 s1 :
        p12 \In tcatp p1 pp2 -> 
        always s1 p1 (fun s2 p2 => 
          forall p v, p2 = Ret v -> p \In pp2 v -> 
            shared s2 :+ local s2 \In safe p) ->
        shared s1 :+ local s1 \In safe p12.
Proof.
case: s1=>sh1 lp1 ep1 T Ha ps /=.
(* weaken ep1 to empty, but keep the ghosts *)
have {Ha} /= Ha: 
  always [sh1, lp1, ProcHeap (mx ep1) (gh ep1) empty] p1
    (fun s2 p2 => forall p v, p2 = Ret v -> p \In pp2 v ->
                    shared s2 :+ local s2 \In safe p).
- apply: alw_sub_def (Ha)=>//=; split=>//.
  by move: (alwD Ha)=>/=; defcheck.
elim: ps p12 sh1 lp1 p1 T Ha=>[|bs ps IH] p12 sh1 lp1 p1 T Ha //.
case K: (active p1); last first.
- case: p1 T K Ha=>//.
  - by move/tcatp_XdummyP=>->; case: bs=>//; case.
  move=>a; move/tcatp_XretP=>T _.
  by move/alw_refl/(_ _ _ (erefl _) T)/(_ (bs::ps)).
have W: ~~ Retb p1 by case: p1 K T Ha.
move=>C a f v Ac; split.
- move/(tcatp_actP a f _ T W): Ac=>Ac.
  by move/alw_safe: Ha; case/(_ (bs::ps))/(_ C a f v Ac).
move=>m p12' Hs {C a f v Ac}.
case: (tcatp_hstP T Hs K)=>{Hs} p1' [T'].
rewrite -[hp lp1]unh0 unA.
case/(alw_hstep Ha)=>sh2 [lp2][f3][->] An Cs Ha'. 
by rewrite unh0; apply: IH T' Ha'.
Qed.

(* A slightly different formulation of alw_tcatp *)
(* which seems more appropriate for lsf_bind *)
(* I could not, however, prove it out of alw_tcatp *)

Lemma alw_bnd' A B (p1 : tree _ A) (p12 : tree _ B) pp2 s1 
                 (P : state -> B -> Prop) :
        p12 \In tcatp p1 pp2 -> 
        always s1 p1 (fun s2 p2 => 
          forall p v, p2 = Ret v -> p \In pp2 v -> 
            always s2 p (fun s p => forall v, p = Ret v -> P s v)) ->
        always s1 p12 (fun s p => forall v, p = Ret v -> P s v).
Proof.
case: s1=>sh1 lp1 ep1 T Ha ps.
elim: ps p12 sh1 lp1 ep1 p1 T Ha =>[|bs ps IH] p12 sh1 lp1 ep1 p1 T Ha /=.
- split; first by move: (alw_coh Ha).
  move=>s2 Envs; move/(alw_envs Ha): (Envs)=>Ha2. 
  split.
   - apply: (alw_bnd_safe T); apply: alw_envs Envs.
     by apply: alw_imp Ha=>t p _ _ L q v E; move/(L _ _ E); move/alw_safe.
  case K: (Retb p1); last first.
  - move=>v E; rewrite E in T. 
    by case/tcatp_retP: T K=>w [->].
  case: p1 K T Ha2 Ha=>//=.
  move=>a _; move/tcatp_XretP=>T.
  by move/alw_refl/(_ p12 a (erefl _) T)/alw_refl.
split; first by apply: alw_coh Ha.
case=>sh2 lp2 ep2 Lcs; split.
- apply: (alw_bnd_safe T); apply: alw_envs Lcs.
  by apply: alw_imp Ha=>t p _ _ L q v E; move/(L _ _ E); move/alw_safe.
- move=>v E.
  rewrite E in T.
  case/tcatp_retP: T Ha=>w [->{p1}] T Ha.
  move/(alw_envs Ha): Lcs.
  by move/alw_refl/(_ _ w (erefl _) T)/alw_refl; apply.
move=>h3 p12' Hs.
case K: (active p1); last first.
- case: p1 K T Hs Ha=>//.
  - by move=>_; move/tcatp_XdummyP=>->; move/hst_dummyP.
  move=>a _; move/tcatp_XretP=>T Hs.
  move/alw_refl/(_ p12 a (erefl _) T)/(_ (bs :: ps))/proj2.
  by case/(_ _ Lcs)=>_ _; apply.
case: (tcatp_hstP T Hs K)=>p1' [T'] Hs'.
move: (alw_envs Ha Lcs)=>{Ha} Ha.
case/(alw_hstep Ha): Hs'=>sh3 [ep3][f3][->].
rewrite -(tcatp_annotP f3 bs T K)=>An Cs Ha'.
by exists sh3, ep3, f3; split=>//; apply: IH T' Ha'.
Qed.

Lemma lsf_bnd' A B (p1 : tree _ A) (p12 : tree _ B) s1 pp2 :
        p12 \In tcatp p1 pp2 ->
        always s1 p1 
          (fun s2 p2 =>
            forall p v, p2 = Ret v -> p \In pp2 v -> lock_safe s2 p) ->
        lock_safe s1 p12.
Proof.
move=>T Ha.
suff H: always s1 p12 (fun s p => forall v : B, p = Ret v -> True).
- by apply: alw_imp H.
apply: alw_bnd' T _; apply: alw_imp Ha=>s p _ _ I q v H.
by move/(I _ _ H); apply: alw_imp.
Qed.

(* this suggest capturing the pattern by means of a new definition *)

Definition after A s (p : tree _ A) (P : state -> A -> Prop) := 
  always s p (fun s2 p2 => forall v, p2 = Ret v -> P s2 v).

Lemma aft_bnd A B (p1 : tree _ A) (p12 : tree _ B) pp2 s1 P :
        p12 \In tcatp p1 pp2 -> 
        after s1 p1 (fun s v => 
          forall p, p \In pp2 v -> after s p P) ->
        after s1 p12 P.
Proof. 
move=>T H; apply: alw_bnd' T _.
by apply: alw_imp H=>s p _ _ I q v E; move/(I v E).
Qed.

Lemma aft_lsf A s (p : tree _ A) :
        lock_safe s p <-> after s p (fun s v => True).
Proof. by split; apply: alw_imp. Qed.

Lemma aft_imp A s (p : tree _ A) (P1 P2 : _ -> _ -> Prop) : 
        (forall s v, s \In coherent -> 
                     P1 s v -> P2 s v) -> 
        after s p P1 -> after s p P2.
Proof. by move=>I; apply: alw_imp=>s1 p1 C _ H v; move/H; apply: I. Qed.

Lemma aftA A B s (p : tree _ A) (P : B -> _ -> _ -> Prop) : 
        lock_safe s p ->
        (after s p (fun s' v' => forall x, P x s' v') <->
         forall x, after s p (fun s' v' => P x s' v')).
Proof.
move=>Ls; rewrite -(alwA Ls).
split; apply: alw_imp=>t q _ _ I.
- by move=>x v; move/I.
by move=>v; move/I.
Qed.

Implicit Arguments aftA [A B s p P].

Lemma lsf_bnd A B (p1 : tree _ A) (p12 : tree _ B) s1 pp2 :
        p12 \In tcatp p1 pp2 ->
        after s1 p1 (fun s v => 
          forall p, p \In pp2 v -> lock_safe s p) ->
        lock_safe s1 p12.
Proof.
move=>T Ha; apply/aft_lsf; apply: aft_bnd T _.
by apply: aft_imp Ha=>s v C I p K; apply/aft_lsf; apply: I K.
Qed.

(*****************************)
(* Always with tree prefixes *)
(*****************************)

(*
Lemma alw_tpfx A (s : state) (p1 p2 : tree U A) (P : state -> Prop) : 
        p1 << p2 -> 
        always s p2 (fun s _ => P s) -> 
        always s p1 (fun s _ => P s).
Proof.
move=>H Ls ps.
elim: ps p1 p2 s H Ls=>[|bs ps IH] p1 p2 s H Ls /=.
- split=>[|s2 Envs]; first by apply: alw_coh Ls.
  move/(alw_envs Ls): Envs=>{Ls} Ls.
  split; last by move/alw_refl: Ls.
  by apply: tpfx_sfP H _; move/alw_safe: Ls.
split=>[|s2 Envs]; first by apply: alw_coh Ls.
move/(alw_envs Ls): Envs=>{Ls} Ls.
split=>[||h3 q Ts].
- by apply: tpfx_sfP H _; move/alw_safe: Ls.
- by move/alw_refl: Ls.
case: s2 Ls Ts=>sh2 lp2 ep2 /= Ls.
case/(tpfx_hstP H)=>q' [H'].
case/(alw_hstep Ls)=>sh3 [lp3][f3][->] An Cs Alw.
exists sh3, lp3, f3; split=>//; last by apply: IH H' Alw.
apply: tpfx_annotP An.
(* hmm, can't just use tpfx_annotP here; have to make use *)
(* the fact that p1 steps at bs, but it doesn't seem i have that *)
(* lemma around. so comment out *)
Qed.

Lemma aft_tpfx A (s : state) (p1 p2 : tree A) P : 
        p1 << p2 -> after s p2 P -> after s p1 P.
Proof.
move=>H Ls ps.
elim: ps p1 p2 s H Ls=>[|bs ps IH] p1 p2 s H Ls /=.
- split=>[|s2 Envs]; first by apply: alw_coh Ls.
  move/(alw_envs Ls): Envs=>{Ls} Ls.
  split=>[|v E1]; first by apply: tpfx_sfP H _; move/alw_safe: Ls.
  rewrite E1 in H; move/tpfx_retXP: H=>E2; rewrite E2 in Ls.
  by move/alw_refl: Ls; apply.
split=>[|s2 Envs]; first by apply: alw_coh Ls.
move/(alw_envs Ls): Envs=>{Ls} Ls.
split=>[|v E1|h3 q Ts].
- by apply: tpfx_sfP H _; move/alw_safe: Ls.
  rewrite E1 in H; move/tpfx_retXP: H=>E2; rewrite E2 in Ls.
  by move/alw_refl: Ls; apply.
case: s2 Ls Ts=>sh2 lp2 ep2 /= Ls.
case/(tpfx_hstP H)=>q' [H'].
case/(alw_hstep Ls)=>sh3 [lp3][->] Cs Alw.
by exists sh3, lp3; split=>//; apply: IH H' Alw.
Qed.
*)

Lemma alw_cons A B (s1 : state) a g (p : A -> tree _ B) P : 
         s1 \In coherent ->
         (forall s2, env_steps s1 s2 ->
            [/\ shared s2 :+ local s2 \In safe (Cons a g p), 
                P s2 (Cons a g p) &
                forall v m, astep (heap_of s2) a v m ->
                local_step (Cons a g p) [::] s2 m 
                (fun s3 => always s3 (p v) P)]) ->
         always s1 (Cons a g p) P.
Proof.
move=>C H ps; apply/alwpE; split=>// s2; case/H=>H1 H2 H3; split=>//.
case: ps=>[//| bs ps].
move=>h3 q; case: bs=>[|b bs] /=; last by case: b.
case=>v []; move/H3=>H4 ->.
case: s2 H4 {H1 H2 H3}=>sh lp ep [sh2][lp2][f][->] An Cs H2.
by exists sh2, lp2, f.
Qed.

(***********************************)
(* always and parallel composition *)
(***********************************)

Lemma alw_asafe A B sh lp ep (p : tree _ A) (a : action B) f (v : B) bs P : 
        always [sh, lp, ep] p P -> 
        (a, f) \In action_at p bs ->
        sh :+ lp \In asafe a.
Proof. by move/alw_safe=>/= S H; case: (S [:: bs] B a f v H). Qed.

Lemma alw_safe_par A1 A2 sh ep lp1 lp2 
                   (p1 : tree _ A1) (p2 : tree _ A2) P1 P2 :
        always [sh, lp1, lp2 \+ ep] p1 P1 ->
        always [sh, lp2, lp1 \+ ep] p2 P2 -> 
        sh :+ lp1 :+ lp2 \In safe (Par p1 p2 Ret).
Proof.
move=>Ls1 Ls2 ps; move/alwD: (Ls2)=>/=.
(* weaken ep to empty *)
rewrite unA; move/defUnl; move=>D.
set ep' := ProcHeap (mx ep) (gh ep) empty.
have {Ls1} Ls1 : always [sh, lp1, lp2 \+ ep'] p1 P1.
- by apply: alw_sub_def Ls1=>/=; split=>//; move: D; defcheck.
have {Ls2} Ls2 : always [sh, lp2, lp1 \+ ep'] p2 P2.
- by apply: alw_sub_def Ls2=>/=; split=>//; move: D; defcheck.
rewrite {}/ep' in Ls1 Ls2.
case: ep lp1 lp2 {D} Ls1 Ls2=>b g _ [b1 g1 h1][b2 g2 h2] /=.
rewrite !mgh_join !unitR.
elim: ps b b1 b2 sh g g1 g2 h1 h2 p1 p2 => [//|bs ps IH] 
         b b1 b2 sh g g1 g2 h1 h2 p1 p2 Ls1 Ls2. 
move=>A a f v H; move/alwD: (Ls2)=>/= D.
case/act_parP: H.
- case=>v1 [v2][->] {f}_ E; rewrite {A}E in a v *.
  by move/jmE=>->->->; split=>//= m q [->] ->; apply: sf_retP.
case; case=>-> /= Ac.
- split=>[|m q [p1'][->]].
  - by move/(alw_asafe v Ls1): Ac; apply: asf_mono; rewrite unAC.
  case/(alw_hstep Ls1)=>sh' [[b1' g1' h1']][f1'][->{m}] An Cs Ls1';
  case/csP: Cs An Ls1' Ls1 Ls2=>/=.
  - case=>->[->->] An Ls1' Ls1 Ls2; move/alwD: (Ls1')=>/= D'. 
    apply: IH Ls1' _; apply: alw_envs (Ls2) _; split=>//.
    by apply: coh_sub (alw_coh Ls2)=>/=; rewrite -unAC.
  - move=>inv gL -> [->->->] H Asrt V L Sh I An Ls1' Ls1 Ls2.
    rewrite {H b1}(eqP H) /= in Ls1' Ls1 Ls2 *.
    move: (coh_ownsL (alw_coh Ls1)); rewrite owno_join; case/andP=>E2 E.
    rewrite {b E}(eqP E) {b E2}(eqP E2) /= in Ls1' Ls1 Ls2 *.
    apply: IH Ls1' _; apply: alw_envs (Ls2) _; apply: envsR.  
    by rewrite joinCA.
  move=>L -> [-> -> ->] V I An Ls1' Ls1 Ls2.
  move: (coh_lkd_owns (alw_coh Ls1))=>/=. 
  rewrite (negbT L) /= owno_join; case/andP=>E1; case/andP=>E2 E.  
  rewrite {b E}(eqP E) {b1 E1}(eqP E1) {b2 E2}(eqP E2) in Ls1' Ls1 Ls2 *.
  apply: IH Ls1' _; apply: alw_envs (Ls2) _; apply: envsG.
  by rewrite joinCA.
split=>[|m q [p2'][->]].
- by rewrite unAC; move/(alw_asafe v Ls2): Ac; apply: asf_mono.
rewrite unAC; case/(alw_hstep Ls2)=>sh' [[b2' g2' h2']][f2'][->{m}] An Cs Ls2'.
case/csP: Cs An Ls2' Ls1 Ls2=>/=.
- case=>->[->->] An Ls2' Ls1 Ls2; move/alwD: (Ls2')=>/= D'. 
  rewrite unAC; apply: IH Ls2'; apply: alw_envs (Ls1) _; split=>//.
  by apply: coh_sub (alw_coh Ls1)=>/=; rewrite -unAC.
- move=>inv gL -> [->->] -> H Asrt V L Sh I An Ls2' Ls1 Ls2.
  rewrite {H b2}(eqP H) /= in Ls2' Ls1 Ls2 *.
  move: (coh_ownsL (alw_coh Ls2)); rewrite owno_join; case/andP=>E1 E.
  rewrite {b E}(eqP E) {b1 E1}(eqP E1) /= in Ls2' Ls1 Ls2 *.
  rewrite unAC; apply: IH Ls2'; apply: alw_envs (Ls1) _; apply: envsR. 
  by rewrite unAC joinCA. 
move=>L -> [-> -> ->] V I An Ls2' Ls1 Ls2.
move: (coh_lkd_owns (alw_coh Ls2))=>/=. 
rewrite (negbT L) /= owno_join; case/andP=>E2; case/andP=>E1 E.  
rewrite {b E}(eqP E) {b1 E1}(eqP E1) {b2 E2}(eqP E2) in Ls2' Ls1 Ls2 *.
rewrite unAC; apply: IH Ls2'; apply: alw_envs (Ls1) _; apply: envsG.
by rewrite unAC joinCA. 
Qed.

Lemma alw_envs_lpar A sh1 sh2 ep1 ep2 lp lp' (p : tree _ A) P :
        env_steps [sh1, lp \+ lp', ep1] 
                  [sh2, lp \+ lp', ep2] ->
        always [sh1, lp, lp' \+ ep1] p P ->
        always [sh2, lp, lp' \+ ep2] p P.
Proof.
move=>Envs Ls; apply: alw_envs Ls _; move: Envs.
rewrite /env_steps !InE /= /valid /= !unA.
case=>_ J [D]; case/andP=>V1 V2.
rewrite (own_joinE (validL V1)) in J * => H I. 
rewrite D !joinA V1 V2 own_joinE; last first.
- by rewrite -joinA in V1; apply: validR V1.
rewrite orbA; split=>//.
by case: (owns (mx lp)) J=>//= ->.
Qed.

Lemma alw_envs_rpar A sh1 sh2 ep1 ep2 lp lp' (p : tree _ A) P :
        env_steps [sh1, lp' \+ lp, ep1] 
                  [sh2, lp' \+ lp, ep2] ->
        always [sh1, lp, lp' \+ ep1] p P ->
        always [sh2, lp, lp' \+ ep2] p P.
Proof. by rewrite (joinC lp'); apply: alw_envs_lpar. Qed.
 
Lemma alw_envs_safe_par A1 A2 sh sh' ep ep' lp1 lp2
                        (p1 : tree _ A1) (p2 : tree _ A2) P1 P2 : 
        always [sh, lp1, lp2 \+ ep] p1 P1 ->
        always [sh, lp2, lp1 \+ ep] p2 P2 ->
        env_steps [sh, lp1 \+ lp2, ep] [sh', lp1 \+ lp2, ep'] ->
        sh' :+ lp1 :+ lp2 \In safe (Par p1 p2 Ret).
Proof.
move=>Ls1 Ls2 Envs.
move/(alw_envs_lpar Envs): Ls1=>Ls1; move/(alw_envs_rpar Envs): Ls2=>Ls2.
apply: alw_safe_par Ls1 Ls2.
Qed.

Lemma id_annotP A1 A2 (v1 : A1) (v2 : A2) : 
        id_annot \In annot_at (Par (Ret v1) (Ret v2) Ret) [::].
Proof. by exists unit, Idle. Qed.

Hint Resolve id_annotP.

Lemma alw_par_main A1 A2 sh ep lp1 lp2 (p1 : tree _ A1) (p2 : tree _ A2) P1 P2 :
        always [sh, lp1, lp2 \+ ep] p1 P1 ->
        always [sh, lp2, lp1 \+ ep] p2 P2 -> 
        always [sh, lp1 \+ lp2, ep] (Par p1 p2 Ret)
          (fun s' p12' => 
             exists sh', exists lp1', exists lp2', exists ep',
               [/\ s' = [sh', lp1' \+ lp2', ep'],
                   forall p1' p2', p12' = Par p1' p2' Ret ->
                     P1 [sh', lp1', lp2' \+ ep'] p1' /\
                     P2 [sh', lp2', lp1' \+ ep'] p2' &
                   forall v1' v2', p12' = Ret (v1', v2') ->
                     P1 [sh', lp1', lp2' \+ ep'] (Ret v1') /\
                     P2 [sh', lp2', lp1' \+ ep'] (Ret v2')]).
Proof.
move=>Ls1 Ls2 ps.
elim: ps sh ep lp1 lp2 p1 p2 Ls1 Ls2 => [|bs ps IH]
         sh ep lp1 lp2 p1 p2 Ls1 Ls2 /=.
- split; first by apply/coh_shuffle; apply: alw_coh Ls1.
  case=>sh' lp' ep' Envs; case: Envs (Envs)=>/= <-{lp'} _ _ Envs.
  rewrite unA; split; first by apply: alw_envs_safe_par Ls1 Ls2 Envs.
  exists sh', lp1, lp2, ep'; split=>//.
  move=>p1' p2'; case/parEX=><-<- _; split. 
  - by move: (alw_envs_lpar Envs Ls1); move/alw_refl.
  by move: (alw_envs_rpar Envs Ls2); move/alw_refl.
split; first by apply/coh_shuffle; apply: alw_coh Ls1. 
case=>sh' lp' ep' /= Envs; case: Envs (Envs)=>/= <-{lp'} _ C Envs.
split; first by rewrite unA; apply: alw_envs_safe_par Ls1 Ls2 Envs.
- exists sh', lp1, lp2, ep'; split=>//.
- move=>p1' p2'; case/parEX=><-<- _; split.
  - by move: (alw_envs_lpar Envs Ls1); move/alw_refl.
  by move: (alw_envs_rpar Envs Ls2); move/alw_refl.
move=>h3 q H; case/hst_parP: H; last first.
- case=>v1 [v2][->->->{q} E1 E2]; rewrite {p1}E1 {p2}E2 in Ls1 Ls2 *.
  exists sh', (lp1 \+ lp2), id_annot; split=>//. 
  - by rewrite /case_step /=; case: (lockd sh').
  apply: alw_retP=>//; case=>sh'' lp'' ep'' Envs'.
  case: Envs' (Envs')=>/= <- _ S Envs'.
  exists sh'', lp1, lp2, ep''; split=>//.
  move=>_ _ [<-<-]; split.
  - move: (alw_envs_lpar Envs Ls1); move/(alw_envs_lpar Envs').
    by move/alw_refl; apply.
  move: (alw_envs_rpar Envs Ls2); move/(alw_envs_rpar Envs').
  by move/alw_refl; apply.
move/(alw_envs_lpar Envs): Ls1=>Ls1; move/(alw_envs_rpar Envs): Ls2=>Ls2.
case=>[[p1']|[p2']]; case=>->->{q} /=.
- rewrite unA -(unA _ _ ep').
  case/(alw_hstep Ls1)=>sh'' [lp1'][f1'][->{h3}] An Cs Ls1'.
  exists sh'', (lp1' \+ lp2), f1'; move/alwD: (Ls1'); rewrite /= !unA.
  case/csP: Cs An Ls1'.
  - case=>-> E An Ls1' D; split=>//; first by apply/csP1; rewrite !mg_join E.
    apply: IH Ls1' _; apply: alw_envs (Ls2) _.
    split=>//=; first by case: E=>_ ->.
    by apply: coh_sub (alw_coh Ls2); rewrite -unA (unCA lp2) !unA !mg_join E.
  - move=>inv gL ->-> E1 /= Asrt V L' Sh I /= An Ls1' D; move/cohD: (C)=>/= D'. 
    move: (coh_ownsL (alw_coh Ls1)); rewrite E1 /= owno_join; case/andP=>E2 E'. 
    rewrite /case_step /= L' lockdR own_joinE; last by case/cohV/validL/andP: C. 
    rewrite E1; do !split=>//.
    - exists inv, (gL \+ gh lp2).
      rewrite (mgh_eta lp2) !mgh_join (eqP E2) /= (* -joinAC *).
      rewrite -!joinA; split=>//=.
      (* rewrite -!joinA (joinCA (gh lp2)) 2!joinA; split=>//. *)
      - rewrite heap_join -subUnC //; last first.
        by move/defUnl: D'; case/subheapX: Sh=>-> _; defcheck.
      - by apply: assertP=>//; apply: (validL (proj2 (andP (cohV C)))).
      by apply: subheap_trans Sh _; apply: subheapUnl; move/defUnl: D'; defcheck.
    apply: IH Ls1' _; apply: alw_envs (Ls2) _; split=>//=.
    - by rewrite (eqP E2).
    rewrite InE /= (eqP E') (eqP E2) /valid /= lockdR -unA (unCA lp2) !unA.
    rewrite joinCA payloadR //.
    by case/subheapX: Sh D'=>-> _; defcheck.
Focus 1.
  move=>L' ->-> /= V I An Ls1' D; move: (coh_lkd_owns C).
  rewrite L' owno_join; case/andP; case/andP=>E1 E2 E'.
  rewrite /case_step L' lockdG.
  rewrite (mgh_eta lp2) (eqP E2) in Ls1' *.
  rewrite {1}mgh_join heap_join (unAC _ (payload sh')) -joinA; split=>//=.
  apply: IH Ls1' _; apply: alw_envs (Ls2) _.
  rewrite (mgh_eta lp2) (mgh_eta ep') (mgh_eta lp1).
  rewrite (eqP E2) (eqP E') (eqP E1) /= !mgh_join joinAC.
  by apply: envsG; move: (cohD C); rewrite /= -!unA -(unCA lp2) joinCA.
rewrite heap_join -(unC lp2) unA -unA.
case/(alw_hstep Ls2)=>sh''[lp2''][f2'][->{h3}] An Cs Ls2'.
exists sh'', (lp1 \+ lp2''), f2'; move/alwD: (Ls2').
rewrite /= -!unA (unCA lp1) !unA; case/csP: Cs An Ls2'.
- case=>-> E An Ls2' D; split=>//; first by apply/csP1; rewrite !mg_join E.
  apply: IH Ls2'; apply: alw_envs (Ls1) _; split=>//=.
  - by case: E=> _ ->.
  by apply: coh_sub (alw_coh Ls1); rewrite -unA (unCA lp1) !unA !mg_join E.
- move=>inv gL ->-> E2 /= Asrt V L' Sh I /= An Ls2' D; move/cohD: (C)=>/= D'. 
  move: (coh_ownsL (alw_coh Ls2)); rewrite E2 /= owno_join; case/andP=>E1 E'. 
  rewrite /case_step /= L' lockdR own_joinE; last by case/cohV/validL/andP: C.
  rewrite E2 orbT; do !split=>//.
Focus 1.
  - exists inv, (gh lp1 \+ gL).
    rewrite (mgh_eta lp1) !mgh_join (eqP E1) /=; split=>//.
Focus 1.
    - rewrite heap_join valid_joinE//= unC -subUnC unC=>//.
      by move/defUnl: D'; case/subheapX: Sh =>-> _; defcheck.
    - rewrite 2!(joinC (gh lp1)); apply: assertP=>//.
      by rewrite joinC; apply: (validL (proj2 (andP (cohV C)))).
    - by rewrite -joinA joinCA.
    - apply: subheap_trans Sh _; apply: subheapUnr.
      by move/defUnl: D'; defcheck.
    by rewrite -joinA joinCA.
  apply: IH Ls2'; apply: alw_envs (Ls1) _; split=>//.
  - by rewrite (eqP E1).
  rewrite InE /= (eqP E') (eqP E1) /valid /= lockdR.
  rewrite -unA (unCA lp1) !unA joinCA payloadR //.
  by case/subheapX: Sh D'=>-> _; defcheck.
move=>L' ->-> /= V I An Ls2' D; move: (coh_lkd_owns C).
rewrite L' owno_join; case/andP; case/andP=>E1 E2 E'.
rewrite /case_step L' lockdG.
rewrite (mgh_eta lp1) (eqP E1) /= in Ls2' *.
rewrite {1}mgh_join (* !joinA *) -joinA joinCA !heap_join !unA; split=>//.
apply: IH Ls2'; apply: alw_envs (Ls1) _.
rewrite (mgh_eta lp1) (mgh_eta ep') (mgh_eta lp2).
rewrite (eqP E1) (eqP E') (eqP E2) /= !mgh_join joinAC.
by apply: envsG; move: (cohD C); rewrite /= !unA joinCA.
Qed.

(* splitting the two cases of alw_par_main: *)
(* when p1 || p2 did not terminate *)

Lemma alw_par A1 A2 sh ep lp1 lp2 (p1 : tree _ A1) (p2 : tree _ A2) P1 P2 :
        always [sh, lp1, lp2 \+ ep] p1 P1 ->
        always [sh, lp2, lp1 \+ ep] p2 P2 -> 
        always [sh, lp1 \+ lp2, ep] (Par p1 p2 Ret)
          (fun s' p12' => 
             exists sh', exists lp1', exists lp2', exists ep',
               s' = [sh', lp1' \+ lp2', ep'] /\
               forall p1' p2', p12' = Par p1' p2' Ret ->
                 P1 [sh', lp1', lp2' \+ ep'] p1' /\
                 P2 [sh', lp2', lp1' \+ ep'] p2').
Proof.
move=>Ls1; move/(alw_par_main Ls1); apply: alw_imp.
move=>s p _ _ [sh'][lp1'][lp2'][ep'][->] L1 L2.
by exists sh', lp1', lp2', ep'.
Qed.

(* and when p1 || p2 did terminate *)

Lemma aft_par A1 A2 sh ep lp1 lp2 (p1 : tree _ A1) (p2 : tree _ A2) P1 P2 :
        after [sh, lp1, lp2 \+ ep] p1 P1 ->
        after [sh, lp2, lp1 \+ ep] p2 P2 -> 
        after [sh, lp1 \+ lp2, ep] (Par p1 p2 Ret)
          (fun s' v' => 
             exists sh', exists lp1', exists lp2', exists ep',
               [/\ s' = [sh', lp1' \+ lp2', ep'], 
                   P1 [sh', lp1', lp2' \+ ep'] v'.1 &
                   P2 [sh', lp2', lp1' \+ ep'] v'.2]).
Proof.
move=>Ls1; move/(alw_par_main Ls1); apply: alw_imp.
move=>s p _ _ [sh'][lp1'][lp2'][ep'][->] L1 L2 [v1 v2].
case/L2=>H1 H2; exists sh', lp1', lp2', ep'.
by split=>//; [apply: H1 | apply: H2].
Qed.

Lemma lsf_par A1 A2 sh ep lp1 lp2 (p1 : tree _ A1) (p2 : tree _ A2) :
        lock_safe [sh, lp1, lp2 \+ ep] p1 ->
        lock_safe [sh, lp2, lp1 \+ ep] p2 -> 
        lock_safe [sh, lp1 \+ lp2, ep] (Par p1 p2 Ret).
Proof. by move=>Ls1; move/(alw_par Ls1); apply: alw_imp. Qed.

End LockSafe.

Notation lock_safe_path s p ps := (always_path s p ps (fun _ _ => True)).
Notation lock_safe s p := (always s p (fun _ _ => True)).

Implicit Arguments alwpA [R A B s p ps P].
Implicit Arguments alwA [R A B s p P].
Implicit Arguments aftA [R A B s p P].
