Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq rels ordtype.
Require Import Eqdep prelude domains heaps hprop.
Require Import pcm resources8 proc9 step9 hstep9 sbisim9 wbisim9.
Require Import locksafe9 prog11.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(* small footprints *)

(********************)
(* Separation monad *)
(********************)

Section Separation.
Variable R : resource.
Notation U := (univ_of R).
Notation tree A := (tree U A).
Notation lk := (lock_of R).
Notation procHeap := (procHeap U).
Notation Inv := (Inv (R:=R)).
Notation pre := (@pre R).
Notation post := (@post R).
Notation cont := (@cont R).
Notation spec := (spec (R:=R)).

Definition ptop (p : procHeap -> Prop) : procHeap -> Prop := 
  fun lp => exists lp', exists h, lp = lp' \+ [mg0 h] /\ p lp'.

Definition lolli (p : procHeap -> Prop) 
                 (q : procHeap -> procHeap -> Prop) i m := 
  forall i1 h, i = i1 \+ [mg0 h] -> def i -> p i1 -> 
    exists m1, m = m1 \+ [mg0 h] /\ def m /\ q i1 m1.

Notation "p '--o' q"   := (lolli p q) (at level 75). 

Lemma mghK (m1 m2 : procHeap) h h2 : 
        def (m1 :+ h) -> 
        m1 \+ [mg0 h] = m2 \+ [mg0 h2 :+ h] -> m1 = m2 \+ [mg0 h2].
Proof.
move=>D []; case; rewrite !unitR joinA => E1 E2.
by move/(unKhl D)=>E; apply/mghE; rewrite /= E1 E2 !unitR.
Qed.

Lemma antiframe p q i m h : 
        def (hp i :+ h) -> 
        (p --o q) (i \+ [mg0 h]) (m \+ [mg0 h]) -> (p --o q) i m.
Proof.
move=>D1 H i2 m2 E D2 H1; rewrite {i}E in H D1 D2 *.
move: (H i2 (m2 :+ h)). 
rewrite -{1}joinA {1}mgh_join /= !unitL. 
case/(_ (erefl _) D1 H1)=>h2 [E][D3 H2].
move: (mghK D3 E)=>{E} E; rewrite E in D3.
move: (H i2 (m2 :+ h)).  
rewrite -{1}joinA {1}mgh_join /= !unitL.
case/(_ (erefl _) D1 H1)=>m1 [T][D] Q.
by exists m1; rewrite -(mghK D T) (defUnl D).
Qed.

(* p --o q is local *)
Lemma locality p q (i1 m : procHeap) h : 
        def (hp i1 :+ h) -> ptop p i1 -> (p --o q) (i1 \+ [mg0 h]) m -> 
          exists m1, m = m1 \+ [mg0 h] /\ def m /\ (p --o q) i1 m1.
Proof.
move=>D [h1][h2][E] H1 H2; rewrite {i1}E in D H2 *.
move: (H2 h1 (h2 :+ h)); rewrite -{1}joinA {1}mgh_join /= !unitL.
case/(_ (erefl _) D H1)=>m1 [E][D2] Q.
rewrite {m}E in H2 D2 *; exists (m1 \+ [mg0 h2]).
rewrite -{1}joinA {1}mgh_join /= !unitL; do !split=>//.
by apply: antiframe D _; rewrite -(joinA m1) {1}mgh_join /= !unitL.
Qed.  

Lemma fr_pre p i h : ptop p i -> ptop p (i \+ [mg0 h]).
Proof. 
case=>i1 [i2][->] H; exists i1, (i2 :+ h).
by rewrite -{1}joinA {1}mgh_join /= !unitL.
Qed.

Definition fr A (s : spec A) : spec A := 
  (fun gE => ptop (s.1 gE), 
   fun x gE gE' => s.1 gE --o (s.2 x gE gE')).

Prenex Implicits fr.

Notation "[ s ]" := (fr s).

Definition progsep A (s : spec A) := prog A [s]. 

(*********************)
(* Basic combinators *)
(*********************)

(* recursion *)

Section SepFix.
Variables (A : Type) (B : A -> Type) (s : forall x, spec (B x)).
Notation tp := (forall x, progsep (s x)).

Definition ffix (f : tp -> tp) : tp := prog11.ffix f.

End SepFix.

(* monadic unit *)

Section SepReturn.
Variables (A : Type) (x : A).

Definition ret_s : spec A := 
  (fun gE lp => hp lp = empty, 
   fun v gE1 gE2 lp1 lp2 => 
     [/\ lp2 = lp1, v = x & 
         owns (mx lp2) -> gE2 = gE1]).

Program 
Definition ret : progsep ret_s :=
             prog11.do' (prog11.ret x) _.
Next Obligation.
move=>lp gE [h1][h2][->] E1 sh mE hE C. 
split=>// _ sh' lp' mE' gE' hE' C' /= [E] -> O.
rewrite {lp'}E /= in C' O *; move=>i h E2.
rewrite E2 unitR in O C C' *; move=>D E3.
exists i; do !split=>//.
by case: E2 O; rewrite !unitR=>->.
Qed.

End SepReturn.

(* monadic bind *)

Lemma cohDlp sh lp ep : [sh, lp, ep] \In @coherent R -> def lp.
Proof. by move/cohD; move/defUnl/defUnr. Qed.

Lemma cohF sh lp ep h : [sh, lp \+ [mg0 h], ep] \In @coherent R -> 
                        [sh, lp, ep] \In @coherent R.
Proof.
rewrite /coherent !InE /=; case=>D; rewrite !unitR=>V O J.
by split=>//; move: D; rewrite unA unAC; move/defUnl.
Qed.



Section SepBind.
Variables (A B : Type) (s1 : spec A) (s2 : A -> spec B).
Variables (e1 : progsep s1) (e2 : forall x, progsep (s2 x)).

Definition bind_s : spec B := prog11.bind_s [s1] (fr \o s2).

Program
Definition bind : progsep bind_s := 
  prog11.do' (prog11.bind e1 e2) _.
Next Obligation.
move=>_ gE [i][h][->][H] T sh mE hE C; split=>[|{H T}].
- split=>[|x sh' lp' ep' C']; first by apply: fr_pre.
  case/(locality (cohDlp C) H)=>m1 [E'][D'].
  move/T; move/(_ sh'); rewrite {}E' in C' D' *; move/(_ (cohF C')).
  by apply: fr_pre.
move=>x' sh' lp' mE' gE' hE' C' T' i' h' E' D' [H' S'].
rewrite {i}E' in C D' T' *.
case: T'=>x1 [sh1][lp1][ep1][C1].
case/(locality D' H')=>i2 /= [E1][D1] T1; rewrite {lp1}E1 in C1 D1 *.
move/S': (T1); move/(_ sh1 (cohF C1))=>H2.
case/(locality D1 H2)=>m1 [->][D2] T2.
exists m1; do !split=>//; exists x1, sh1, i2, ep1.
by split=>//; apply: cohF C1.
Qed.

End SepBind.



Section SepPar.
Variables (A B : Type) (s1 : spec A) (s2 : spec B).
Variables (e1 : progsep s1) (e2 : progsep s2).

Definition par_s : spec (A * B) := prog11.par_s [s1] [s2].

Program
Definition par : progsep par_s := 
  prog11.do' (prog11.par e1 e2) _.
Next Obligation.
move=>_ gE1 [_][h1][->][lpl1][lpr1][->] Sl1 Sr1 sh1 mE1 hE1 C1. 
split=>[|x2 sh2 i2 mE2 gE2 hE2 C2 {Sl1 Sr1 sh1 mE1 hE1 C1}].
- exists lpl1, (lpr1 \+ [mg0 h1]); rewrite /= unitR joinA.
  by split=>//; apply: fr_pre. 
move=>H i' h2 E D [lpl2][lpr2][E'] Sl2 Sr2. 
rewrite {lpl1 lpr1}E {i'}E' in D H *; rewrite -joinA in D.
move: (H lpl2 (lpr2 \+ [mg0 h2])); rewrite joinA /= unitR.
case/(_ (erefl _) Sl2 (fr_pre _ Sr2))=>lpl3 [i'][E'] _.
rewrite {i2}E' in C2 H *.
case/(locality (defUnr D) Sr2)=>lpr3 [E'][_] _.
rewrite {Sl2 Sr2 i'}E' /= in C2 H *.
exists (lpl3 \+ lpr3); split; first by rewrite joinA.
split; first by apply: cohDlp C2.
move=>lpl1 lpr1 E' Sl1 Sr1. 
rewrite joinA in D; rewrite {lpl2 lpr2}E' in D H; rewrite -joinA in D.
move: {H} (H lpl1 (lpr1 \+ [mg0 h2])); rewrite joinA /= unitR.
case/(_ (erefl _) Sl1 (fr_pre h2 Sr1))=>lpl4 [i'][E'] Ql1.
case/(locality (defUnr D) Sr1)=>lpr4 [E''][_] Qr1.
rewrite {i'}E'' /= unitR in E' Ql1 *. 
rewrite !joinA in E' C2; exists lpl4, lpr4; split=>//.
apply/mghE; case: E'; rewrite /= !unitR; move=>->->. 
by move/(unKhl (cohDlp C2)).
Qed.

End SepPar.



Definition verify' A lp1 gE1 (s : spec A) (r : cont A) := 
  forall sh1 mE1 hE1, [sh1, lp1, [Mgh mE1, gE1, hE1]] \In coherent -> 
  s.1 gE1 lp1 /\ 
  forall x sh2 lp2 mE2 gE2 hE2, 
    [sh2, lp2, [Mgh mE2, gE2, hE2]] \In coherent -> 
    s.2 x gE1 gE2 lp1 lp2 -> r x gE2 lp2.

Notation verify lp gE s r := (@verify' _ lp gE (fr s) r).



Section SepFrame.
Variables (A : Type) (s : spec A). 

Lemma frame' lp gE h (r : cont A) : 
        verify lp gE s (fun x' gE' lp' => 
          def (lp' \+ [mg0 h]) -> r x' gE' (lp' \+ [mg0 h])) ->
        verify (lp \+ [mg0 h]) gE s r.
Proof.
move=>H sh1 mE1 hE1 C1.
case: (H sh1 mE1 hE1 (cohF C1))=>P1 Q1. 
split=>[|x sh2 lp2 mE2 gE2 hE2 C2]; first by apply: fr_pre.
case/(locality (cohDlp C1) P1)=>lp' [E][D2].
rewrite {lp2}E /= in C2 D2 *.
by move/Q1; move/(_ _ _ _ (cohF C2) D2).
Qed.

(* unpackaging lp *)
Lemma frame mL gL gE i h (r : cont A) : 
        verify [Mgh mL, gL, i] gE s (fun x' gE' lp' => 
          def (lp' \+ [mg0 h]) -> 
          r x' gE' [Mgh mx lp', gh lp', hp lp' \+ h]) ->
        verify [Mgh mL, gL, i \+ h] gE s r.
Proof.
move=>H.
rewrite (_ : [Mgh mL, gL, i \+ h] = [Mgh mL, gL, i] \+ [mg0 h]); last first.
- by rewrite mgh_join !unitR.
apply: frame'; apply: vrf_mono H=>x' gL' [mE'] gE' hE' /= H. 
by move/H; rewrite mgh_join !unitR.
Qed.

Lemma frame0 lp gE r : verify' lp gE s r -> verify lp gE s r.
Proof.
move=>H sh1 mE1 hE1 C1.
case: (H sh1 mE1 hE1 C1)=>P1 Q1. 
split=>[|x sh2 lp2 mE2 gE2 hE2 C2].
- by exists lp, empty; rewrite unitR.
move/(_ lp empty); rewrite unitR.
case/(_ (erefl _) (cohDlp C1) P1)=>lp' [].
by rewrite unitR => <- [_]; apply: Q1 C2.
Qed.

Lemma frame1 lp gE (r : cont A) : 
        verify' [Mgh mx lp, gh lp, empty] gE s 
          (fun x' gE' lp' => def (lp' \+ [mg0 lp]) -> 
             r x' gE' (lp' \+ [mg0 lp])) ->
        verify lp gE s r.
Proof.
case: lp=>mL gL hL /=.
rewrite (_ : [Mgh mL, gL, hL] = [Mgh mL, gL, empty] \+ [mg0 hL]).
- by move=>H; apply: frame'; apply: frame0. 
by rewrite mgh_join !unitR unitL.
Qed.

End SepFrame.
 


Definition conseq A (s : spec A) (p : pre) (q : post A) := 
  forall lp gE, p gE lp -> 
    verify lp gE s (fun x' gE' lp' => q x' gE gE' lp lp').

Local Notation conseq1 := 
  (fun A (s1 s2 : spec A) => 
     conseq s1 (let 'pair x _ := s2 in x) 
               (let 'pair _ x := s2 in x)).
 
Lemma conseq_refl A (s : spec A) : conseq1 A s s. 
Proof. by case: s=>s1 s2 lp gE H; apply: frame0. Qed.

Hint Resolve conseq_refl.



Section SepConseq.
Variables (A : Type) (s1 s2 : spec A).
Variables (e : progsep s1) (pf : conseq1 A s1 s2). 

Program Definition do' : progsep s2 := prog11.do' e _.
Next Obligation.
move=>lp1 gE1 /= H sh1 mE1 hE1 C1.
split=>[|x sh2 lp2 mE2 gE2 hE2 C2].
- case: H C1=>i1 [i2][->{lp1}] P2 C1.
  case: (@pf i1 gE1 P2 sh1 mE1 hE1 (cohF C1))=>P1 _.
  by apply: fr_pre.
move=>Q1 i1 i2 E D1 P2; rewrite {lp1}E in C1 H Q1 D1 P2.
case: (@pf i1 gE1 P2 sh1 mE1 hE1 (cohF C1))=>P1 T.
case/(locality D1 P1): Q1=>lp1 [E][D2]; rewrite {lp2}E in C2 D2 *.
by move/T; move/(_ sh2 mE2 hE2 (cohF C2)); exists lp1.
Qed.

End SepConseq.

Notation "'Do' e" := (do' e _) (at level 80).


(*****************************)
(* Singleton heap operations *)
(*****************************)

Section SepRead.
Variables (A : Type) (x : ptr).

Definition read_s : spec A := 
  (fun gE lp => exists v : A, 
     lp = [Mgh mx lp, gh lp, x :-> v],
   fun y gE1 gE2 lp1 lp2 => 
     forall v : A, 
       lp1 = [Mgh mx lp1, gh lp1, x :-> v] ->
       [/\ lp2 = lp1, y = v &
           owns (mx lp2) -> gE2 = gE1]).

Program Definition read : progsep read_s := 
  prog11.do' (prog11.read A x) _.
Next Obligation.
move=>lp1 gE1 /= S1 sh1 mE1 hE1 C1.
split=>[|y sh2 lp2 mE2 gE2 hE2 C2]. 
- by case: S1 C1=>i [h][->{lp1}][v] -> /= C1; exists h, v.
move=>T lp' h E1 D [v] {C1 C2 S1}; case/mghE=>/= _ _ E2.
rewrite {lp1}E1 in T D *; move/(_ h v): T.
rewrite (mgh_eta lp') mgh_join /= !unitR {}E2 in D *.
case/(_ (erefl _))=>->-> /= O.
exists [Mgh mx lp', gh lp', x :-> v].
rewrite mgh_join !unitR; do 2!split=>//.
by move=>w []; move/(pts_inj (defUnl D))=><-.
Qed.

End SepRead.



Section SepWrite.
Variables (A : Type) (x : ptr) (v : A).

Definition write_s : spec unit := 
  (fun gE lp => exists B, exists w : B, 
     lp = [Mgh mx lp, gh lp, x :-> w],
   fun _ gE1 gE2 lp1 lp2 => 
     forall B (w : B), 
       lp1 = [Mgh mx lp1, gh lp1, x :-> w] -> 
       lp2 = [Mgh mx lp1, gh lp1, x :-> v] /\
       (owns (mx lp2) -> gE2 = gE1)).

Program Definition write : progsep write_s := 
  prog11.do' (prog11.write x v) _.
Next Obligation.
move=>lp1 gE1 /= S1 sh1 mE1 hE1 C1.
split=>[|[] sh2 lp2 mE2 gE2 hE2 C2].
- case: S1 C1=>i [h][->{lp1}][B][w] -> /= C1.
  by exists h, B, w.
move=>T lp' h E1 D [B] {C1 C2 S1} [w] E2.
rewrite {lp1}E1 in T D *; move/(_ h B w): T.
rewrite (mgh_eta lp') mgh_join /= !unitR {}E2 /= in D *. 
case/(_ (erefl _))=>-> /= O.
exists [Mgh mx lp', gh lp', x :-> v].
rewrite mgh_join !unitR; do 2!split=>//.
by rewrite !defPtUn in D *.
Qed.

End SepWrite.



Section SepAllocation.
Variables (A : Type) (v : A).

Definition alloc_s : spec ptr := 
  (fun gE lp => hp lp = empty,
   fun x gE1 gE2 lp1 lp2 => 
     lp2 = [Mgh mx lp1, gh lp1, x :-> v] /\
     (owns (mx lp2) -> gE2 = gE1)).

Program Definition alloc : progsep alloc_s := 
  prog11.do' (prog11.alloc v) _.
Next Obligation.
move=>lp1 gE1 /= S1 sh1 mE1 hE1 C1.
split=>//= x sh2 lp2 mE2 gE2 hE2 C2.
case=>{S1 C1} E O lp' h E1 _ /= E2.
rewrite {lp2}E /= {lp1}E1 (mgh_eta lp') /= !unitR {}E2 !unitL in C2 O *.
exists [Mgh mx lp', gh lp', x :-> v]; rewrite mgh_join !unitR.
by do !split=>//; move/defUnl/defUnr: (cohD C2).
Qed.

End SepAllocation.



Section SepDeallocation.
Variable x : ptr.

Definition dealloc_s : spec unit := 
  (fun gE lp => 
     exists A, exists v : A, 
       lp = [Mgh mx lp, gh lp, x :-> v],
   fun _ gE1 gE2 lp1 lp2 => 
     forall A (v : A), 
       lp1 = [Mgh mx lp1, gh lp1, x :-> v] -> 
       lp2 = [Mgh mx lp1, gh lp1, empty] /\
       (owns (mx lp2) -> gE2 = gE1)).

Program Definition dealloc : progsep dealloc_s := 
  prog11.do' (prog11.dealloc x) _.
Next Obligation.
move=>lp1 gE1 /= S1 sh1 mE1 hE1 C1.
split=>[|[] sh2 lp2 mE2 gE2 hE2 C2] /=.
- case: S1 C1=>i [h][->{lp1}][A][v] -> /= C1.
  by exists h, A, v.
move=>T lp' h E1 D [A][v] E2 {S1 C1}; move/(_ h A v): T.
rewrite {lp1}E1 {}E2 /= mgh_join !unitR in D *.
case/(_ (erefl _))=>-> /= O.
exists [Mgh mx lp', gh lp', empty].
by rewrite mgh_join !unitR unitL (defUnr D).
Qed.

End SepDeallocation. 

(* I make CAS only possible on the shared state. *)
(* There's little point in doing it elsewhere, since *)
(* the semantics enforces protocol safety only on the shared state *)

Section SepCAS.

Definition cas_s : spec bool := 
  (fun gE lp => ownsno (mx lp) /\ hp lp = empty,
   fun y gE1 gE2 lp1 lp2 => 
     if y then 
       exists inv, 
         lp2 = [mgh true, gh lp1, inv :+ lp1] /\ 
         Inv (gh lp2 \+ gE2) inv
     else lp2 = lp1).

Program Definition cas : progsep cas_s := 
  prog11.do' (prog11.cas) _.
Next Obligation.
move=>lp1 gE1 /= S1 sh1 mE1 hE1 C1.
split=>[|x sh2 lp2 mE2 gE2 hE2 C2] /=.
- case: S1 C1=>i [h][->{lp1}][O] _ _ /=.
  by rewrite unitR.
case: x.
- case=>inv [E1] J lp' h E2 D [O] E3 {S1 C1}.
  rewrite {lp2}E1 {lp1}E2 /= !unitR in C2 J D *.
  exists [mgh true, gh lp', inv :+ lp'].
  rewrite mgh_join !unitR. rewrite unA.
  split=>//; split; last by exists inv.
  by move/defUnl/defUnr: (cohD C2); rewrite unA.
move=>E1 lp' h E2 D [O] E3 {S1 C1}.
rewrite {lp2}E1 {lp1}E2 /= in C2 D *.
by exists lp'.
Qed.

End SepCAS.

Section SepUnlock.
(* G : U -> Pred U, i.e., a function of the old ghost *)
(* G : U -> U, i.e., a function of the old ghost *)
(* Variables (G : U -> U). *)
Variables (next : annot U).
Notation Inv_prec := (@mod_inv_prec R).

Definition unlock_s : spec unit :=
  (fun gE lp => 
     exists inv, exists gL,
       [/\ lp = [mgh true, gh lp, inv], 
           assert next (gh lp) = Some gL,
           valid (gL \+ gE) &
           Inv (gL \+ gE) inv],
   fun _ gE1 gE2 lp1 lp2 =>
     forall inv gL, 
       lp1 = [mgh true, gh lp1, inv] ->
       assert next (gh lp1) = Some gL ->
       valid (gL \+ gE1) ->
       Inv (gL \+ gE1) inv ->
       lp2 = [mgh false, gL, empty]).

Program Definition unlock : progsep unlock_s := 
  prog11.do' (prog11.unlock next) _.
Next Obligation.
move=>lp1 gE1 /= S1 sh1 mE1 hE1 C1.
split=>[| [] sh2 lp2 mE2 gE2 hE2 C2] /=.
- case: S1 C1=>i [h][->{lp1}][inv][gL][->] /= A1 V1 J1 C1.
  by exists h, inv, gL; rewrite mgh_join !unitR.
move=>T lp' h E1 D [inv][gL][E2] A V J {S1 C1}.
move/(_ h inv gL): T.
rewrite {lp1}E1 {}E2 /= mgh_join !unitR in D *.
move/(_ (erefl _) A V J)=>->.
exists [mgh false, gL, empty]. 
rewrite mgh_join !unitR unitL /= (defUnr D); do !split=>//.
move=>inv1 gL1 [<-] A1 V1 J1.
by rewrite A in A1; case: A1=>->.
Qed.

End SepUnlock.

(**********************************)
(* Conditionals for various types *)
(**********************************)

Section CondBool.
Variables (A : Type) (b : bool) (s1 s2 : spec A). 

Program
Definition If (e1 : progsep s1) (e2 : progsep s2) : 
              progsep (if b then s1 else s2) :=
  match b with true => Do e1 | false => Do e2 end.
End CondBool.

Section CondDecide.
Variable (A : Type) (p1 p2 : Prop) (b : {p1} + {p2}) 
         (s1 : p1 -> spec A) (s2 : p2 -> spec A).

Program
Definition Match_dec (e1 : forall x, progsep (s1 x))
                     (e2 : forall x, progsep (s2 x)) : 
             progsep (match b with left x => s1 x | right x => s2 x end) :=
  match b with left x => Do (e1 x) | right x => Do (e2 x) end.
End CondDecide. 

(* add more conditionals here *)

(********************************************)
(* Derived logical rules for easier proving *)
(********************************************)

Lemma val_do A mL gL i j gE (s : spec A) (r : cont A) :
        s.1 gE [Mgh mL, gL, i] -> 
        (forall x gE' mL' gL' i', 
           s.2 x gE gE' [Mgh mL, gL, i] [Mgh mL', gL', i'] -> 
           valid (gL' \+ gE') -> def (i' :+ j) ->
           r x gE' [Mgh mL', gL', i' :+ j]) ->
        verify [Mgh mL, gL, i :+ j] gE s r. 
Proof.
move=>H1 H2; apply: frame; apply: frame0=>sh mE1 hE1 C1.
split=>// x sh2 [mL2 gL2 hL2] mE2 gE2 hE2 /= C2 T D.
by apply: H2=>//=; case/andP: (cohV C2).
Qed.

Lemma bnd_do' A B (s : spec A) s2 lp gE h (r : cont B) : 
        s.1 gE lp -> 
        (forall x gE' lp', 
           s.2 x gE gE' lp lp' -> 
           verify (lp' \+ [mg0 h]) gE' (s2 x) r) -> 
        verify (lp \+ [mg0 h]) gE (bind_s s s2) r.
Proof.
move=>P T; apply: frame0=>sh mE hE C. 
split=>[| x sh2 lp2 mE2 gE2 hE2 C2].
- split; first by apply: fr_pre; exists lp, empty; rewrite unitR.
  move=>x2 sh2 lp2 ep2 C2.
  case/(_ lp h (erefl _) (cohDlp C) P)=>lp' [E][D].
  by rewrite {lp2}E in C2 D *; case/T/(_ _ _ _ C2).
case=>x1 [sh1][lp1][ep1][C1].
case/(_ lp h (erefl _) (cohDlp C) P)=>lp' [E][D].
rewrite {lp1}E in C1 D *.
by move/T/(_ _ _ _ C1)/proj2/(_ _ _ _ _ _ _ C2); apply.
Qed.

Lemma bnd_hoare' A B (s1 : spec A) s2 lp gE h (r1 : cont A) (r : cont B) : 
        verify lp gE s1 r1 -> 
        (forall x gE' lp', 
           r1 x gE' lp' -> 
           verify (lp' \+ [mg0 h]) gE' (s2 x) r) ->
        verify (lp \+ [mg0 h]) gE (bind_s s1 s2) r.
Proof.
move=>P T.
apply: frame0=>sh mE hE C; move/frame_coh: (C)=>C'.
have P1 : [s1].1 gE lp by case: (P _ _ _ C').
have P2 : forall x sh2 lp2 mE2 gE2 hE2,
    [sh2, lp2, [Mgh mE2, gE2, hE2]] \In coherent ->
    [s1].2 x gE gE2 (lp \+ [mg0 h]) lp2 -> 
    exists lp3, lp2 = lp3 \+ [mg0 h] /\ r1 x gE2 lp3.
- case: (P _ _ _ C')=>T1 T2 x sh2 lp2 mE2 gE2 hE2 C2 H2.
  move/locality: H2; move/(_ (cohDlp C)); move/(_ P1).
  case=>m1 [E] [Df1] H2; rewrite {lp2}E in C2 Df1 *.
  by exists m1; split=>//; move/T2: (frame_coh C2); apply. 
split.
- split; first by apply: fr_pre. 
  move=>x2 sh2 lp2 [mE2 gE2 hE2] C2 H2.
  move/P2: H2; case/(_ _ _ _ C2)=>lp3 [E]; move/T.
  by rewrite E in C2 *; case/(_ _ _ _ C2). 
move=>x sh2 lp2 mE2 gE2 hE2 C2.
case=>x1 [sh1][lp1][[mE1 gE1 hE1]] [C1] /= T1 T2.
move/locality: T1; move/(_ (cohDlp C)); move/(_ P1).
case=>m1 [E][Df1]; rewrite {lp1}E in C1 T2 Df1 *.
case: (P _ _ _ C')=>_ P'; move/P'. 
move/(_ _ _ _ (frame_coh C1)); move/T=>V1; case/V1: C1=>W2.
by move/(_ _ _ _ _ _ _ C2); apply.
Qed.

Lemma bnd_do A B (s : spec A) s2 mL gL gE i j (r : cont B) : 
        s.1 gE [Mgh mL, gL, i] -> 
        (forall x gE' mL' gL' i', 
           s.2 x gE gE' [Mgh mL, gL, i] [Mgh mL', gL', i'] -> 
           verify [Mgh mL', gL', i' :+ j] gE' (s2 x) r) -> 
        verify [Mgh mL, gL, i :+ j] gE (bind_s s s2) r.
Proof.
move=>P T. 
rewrite (_ : [Mgh mL, gL, i :+ j] = [Mgh mL, gL, i] \+ [mg0 j]); last first.
- by rewrite mgh_join !unitR.
apply: bnd_do'=>// x gE' [mL' gL' hL'].
by move/T; rewrite mgh_join !unitR.
Qed.


Lemma bnd_hoare A B (s1 : spec A) s2 mL gL gE i j (r1 : cont A) (r : cont B) : 
        verify [Mgh mL, gL, i] gE s1 r1 -> 
        (forall x gE' mL' gL' i', 
           r1 x gE' [Mgh mL', gL', i'] ->
           verify [Mgh mL', gL', i' :+ j] gE' (s2 x) r) ->
        verify [Mgh mL, gL, i :+ j] gE (bind_s s1 s2) r.
Proof.
move=>P T.
rewrite (_ : [Mgh mL, gL, i :+ j] = [Mgh mL, gL, i] \+ [mg0 j]); last first.
- by rewrite mgh_join !unitR.
apply: (bnd_hoare' (r1 := r1))=>// x gE' [mL' gL' hL'].
by move/T; rewrite mgh_join !unitR.
Qed.

(* in par we need to supply intermediate annotations r1 and r2 *)
(* hopefully, these will usually be the default ones, e.g. *)
(* from the postcondition of e1 and e2; so we won't have to actually *)
(* write them explicitly *)

Lemma par_do' A B (s1 : spec A) (s2 : spec B) lp1 lp2 gE
                 (r1 : cont A) (r2 : cont B) (r : cont (A * B)) :
        verify lp1 (gh lp2 \+ gE) s1 r1 ->
        verify lp2 (gh lp1 \+ gE) s2 r2 ->
        (forall x1 x2 lp1 lp2 gE, 
           valid (gh lp1 \+ gh lp2 \+ gE) ->
           r1 x1 (gh lp2 \+ gE) lp1 -> r2 x2 (gh lp1 \+ gE) lp2 ->
           r (x1, x2) gE (lp1 \+ lp2)) ->
        verify (lp1 \+ lp2) gE (par_s s1 s2) r.
Proof.
move=>H1 H2 H sh mE hE C. 
have C1 : [sh, lp1, lp2 \+ [Mgh mE, gE, hE]] \In coherent.
- by apply/coh_shuffle.
have C2 : [sh, lp2, lp1 \+ [Mgh mE, gE, hE]] \In coherent.
- by apply/coh_shuffle; rewrite joinC.
have S1 : [s1].1 (gh lp2 \+ gE) lp1 by case/H1: C1.
have S2 : [s2].1 (gh lp1 \+ gE) lp2 by case/H2: C2.
have S: (par_s s1 s2).1 gE (lp1 \+ lp2) by exists lp1, lp2.
split; first by exists (lp1 \+ lp2), empty; rewrite unitR.
case=>x1 x2 sh' lp' mE' gE' hE' C' T.
move/(_ (lp1 \+ lp2) empty): T C'; rewrite unitR.
case/(_ (erefl _) (cohDlp C) S)=>m1 [->{lp'}][_].
case/(_ _ _ (erefl _) S1 S2)=>lp1' [lp2'][E'] T1' T2' C'.
rewrite {m1}E' unitR in C' *.
apply: H; first by case: C'=>/= _; case/andP. 
- move/coh_shuffle: C'=>C'.
  by move/(_ _ _ _ C1)/proj2/(_ _ _ _ _ _ _ C'): H1; apply.
rewrite joinC in C'; move/coh_shuffle: C'=>C'.
by move/(_ _ _ _ C2)/proj2/(_ _ _ _ _ _ _ C'): H2; apply.
Qed.

Lemma par_do A B (s1 : spec A) (s2 : spec B) m1 m2 g1 g2 h1 h2 gE
                 (r1 : cont A) (r2 : cont B) (r : cont (A * B)) :
        verify [Mgh m1, g1, h1] (g2 \+ gE) s1 r1 ->
        verify [Mgh m2, g2, h2] (g1 \+ gE) s2 r2 ->
        (forall x1 x2 m1 m2 g1 g2 h1 h2 gE, 
           valid (g1 \+ g2 \+ gE) ->
           r1 x1 (g2 \+ gE) [Mgh m1, g1, h1] -> 
           r2 x2 (g1 \+ gE) [Mgh m2, g2, h2] ->
           r (x1, x2) gE [Mgh m1 \+ m2, g1 \+ g2, h1 \+ h2]) ->
        verify [Mgh m1 \+ m2, g1 \+ g2, h1 \+ h2] gE (par_s s1 s2) r.
Proof.
move=>H1 H2 H; rewrite -mgh_join.
apply: (par_do' (r1:=r1) (r2:=r2))=>//=.
by move=>x1 x2 [m1' g1' h1'][m2' g2' h2'] gE'; apply: H.
Qed.

Lemma val_ret A x mL gL hL gE (r : cont A) : 
       (forall gE', 
          (owns mL -> gE' = gE) ->
          valid (gL \+ gE') -> def hL ->
          r x gE' [Mgh mL, gL, hL]) -> 
       verify [Mgh mL, gL, hL] gE (ret_s x) r. 
Proof.
move=>T; rewrite -[hL]un0h; apply: val_do=>//=. 
by move=>y gE' mL' gL' i' [[->->->]] ->; rewrite un0h; apply:  T.
Qed.

Lemma bnd_ret A B (x : A) mL gL hL gE s (r : cont B) : 
        (forall gE', 
           (owns mL -> gE' = gE) ->
           verify [Mgh mL, gL, hL] gE' (s x) r) -> 
        verify [Mgh mL, gL, hL] gE (bind_s (ret_s x) s) r.
Proof.
move=>T; rewrite -[hL]un0h; apply: bnd_do=>//=.
by move=>y gE' mL' gL' i' [[->->->]] ->; rewrite un0h; apply: T.
Qed.

Lemma val_read A x v mL gL gE i (r : cont A) : 
        (forall gE', 
           (owns mL -> gE' = gE) ->
           valid (gL \+ gE') ->
           def (x :-> v \+ i) ->
           r v gE' [Mgh mL, gL, x :-> v \+ i]) ->
        verify [Mgh mL, gL, x :-> v \+ i] gE (read_s A x) r.
Proof.
move=>T; apply: val_do=>/=; first by exists v.
by move=>y gE' mL' gL' i'; case/(_ v (erefl _))=>[[->->->]] ->; apply: T.
Qed.
 
Lemma bnd_read A B x v mL gL gE i s (r : cont B) : 
        (forall gE', 
           (owns mL -> gE' = gE) -> 
           verify [Mgh mL, gL, x :-> v \+ i] gE' (s v) r) -> 
        verify [Mgh mL, gL, x :-> v \+ i] gE (bind_s (read_s A x) s) r.
Proof.
move=>T; apply: bnd_do=>/=; first by exists v.
by move=>y gE' mL' gL' i'; case/(_ v (erefl _))=>[[->->->]] ->; apply: T.
Qed.

Lemma val_write A B x (v : A) (w : B) mL gL gE i (r : cont unit) : 
        (forall gE', 
           (owns mL -> gE' = gE) ->
           valid (gL \+ gE') ->
           def (x :-> v \+ i) ->
           r tt gE' [Mgh mL, gL, x :-> v \+ i]) ->
        verify [Mgh mL, gL, x :-> w \+ i] gE (write_s x v) r.
Proof.
move=>T; apply: val_do=>/=; first by exists B, w.
by move=>[] gE' mL' gL' i'; case/(_ _ _ (erefl _))=>[[->->->]]; apply: T.
Qed.
 
Lemma bnd_write A B C x (v : A) (w : B) mL gL gE i s (r : cont C) : 
        (forall gE', 
           (owns mL -> gE' = gE) -> 
           verify [Mgh mL, gL, x :-> v \+ i] gE' (s tt) r) -> 
        verify [Mgh mL, gL, x :-> w \+ i] gE (bind_s (write_s x v) s) r.
Proof.
move=>T; apply: bnd_do=>/=; first by exists B, w.
by move=>[] gE' mL' gL' i'; case/(_ _ _ (erefl _))=>[[->->->]]; apply: T.
Qed.

(* Q: do we need here that y is not equal to anything *)
(* in the shared resource? It seems not. In bnd_alloc *)
(* the inner verify takes care of that. In val_alloc  *)
(* r is independent of sh, so it doesn't matter.      *)

Lemma val_alloc A (v : A) mL gL i gE (r : cont ptr) : 
        (forall y gE', 
           (owns mL -> gE' = gE) ->
           valid (gL \+ gE') ->
           def (y :-> v \+ i) ->
           r y gE' [Mgh mL, gL, y :-> v \+ i]) -> 
        verify [Mgh mL, gL, i] gE (alloc_s v) r.
Proof.
move=>T; rewrite -[i]un0h; apply: val_do=>//=.
by move=>y gE' mL' gL' i'; case=>[[->->->]]; apply: T.
Qed.

Lemma bnd_alloc A (v : A) C mL gL i gE s (r : cont C) : 
        (forall y gE', 
           (owns mL -> gE' = gE) ->
           verify [Mgh mL, gL, y :-> v \+ i] gE' (s y) r) ->
        verify [Mgh mL, gL, i] gE (bind_s (alloc_s v) s) r.
Proof.
move=>T; rewrite -[i]un0h; apply: bnd_do=>//=.
by move=>y gE' mL' gL' i'; case=>[[->->->]]; apply: T.
Qed.

Lemma val_dealloc A x (v : A) mL gL i gE (r : cont unit) : 
        (forall gE',
           (owns mL -> gE' = gE) ->
           valid (gL \+ gE') ->
           def i ->
           r tt gE' [Mgh mL, gL, i]) -> 
        verify [Mgh mL, gL, x :-> v \+ i] gE (dealloc_s x) r.
Proof.
move=>T; apply: val_do=>/=; first by exists A, v.
move=>[] gE' mL' gL' i'; case/(_ _ _ (erefl _))=>[[->->->]].
by rewrite un0h; apply: T.
Qed.

Lemma bnd_dealloc A B x (v : B) mL gL i gE s (r : cont A) : 
        (forall gE',
           (owns mL -> gE' = gE) ->
           verify [Mgh mL, gL, i] gE' (s tt) r) ->
        verify [Mgh mL, gL, x :-> v \+ i] gE (bind_s (dealloc_s x) s) r.
Proof.
move=>T; apply: bnd_do=>/=; first by exists B, v.
move=>[] gE' mL' gL' i'; case/(_ _ _ (erefl _))=>[[->->->]].
by rewrite un0h; apply: T.
Qed.

Lemma val_cas gL i ep (r : cont bool) : 
        (forall gE' inv,
           Inv (gL \+ gE') inv ->
           valid (gL \+ gE') ->
           def (inv \+ i) ->
           r true gE' [mgh true, gL, inv \+ i]) ->
        (forall gE', 
           valid (gL \+ gE') -> 
           def i ->
           r false gE' [mgh false, gL, i]) ->
        verify [mgh false, gL, i] ep cas_s r.
Proof.
move=>T1 T2; rewrite -[i]un0h; apply: val_do=>//.
case=>gE' mL' gL' i' /=.
- by case=>inv [[->->->]]; rewrite unh0; apply: T1.
by case=>[[->->->]]; rewrite un0h; apply: T2.
Qed.

Lemma bnd_cas A gL i gE s (r : cont A) : 
        (forall gE' inv,
           Inv (gL \+ gE') inv ->
           verify [mgh true, gL, inv \+ i] gE' (s true) r) ->
        (forall gE', 
           verify [mgh false, gL, i] gE' (s false) r) ->
        verify [mgh false, gL, i] gE (bind_s cas_s s) r.
Proof.
move=>T1 T2; rewrite -[i]un0h; apply: bnd_do=>//.
case=>gE' mL' gL' i' /=.
- by case=>inv [[->->->]]; rewrite unh0; apply: T1.
by case=>[[->->->]]; rewrite un0h; apply: T2.
Qed.

Lemma val_unlock gL gL' i inv gE (r : cont unit) G : 
        assert G gL = Some gL' ->
        valid (gL' \+ gE) -> Inv (gL' \+ gE) inv ->
        (forall gE', 
           valid (gL' \+ gE') ->
           def i ->
           r tt gE' [mgh false, gL', i]) ->
        verify [mgh true, gL, inv \+ i] gE (unlock_s G) r.
Proof.
move=>/= Asrt V I T; apply: val_do=>/=; first by exists inv, gL'.
move=>[] gE' mL' gL'' i'; case/(_ _ _ (erefl _) Asrt V I)=>->->->.
by rewrite un0h; apply: T.
Qed.

Lemma bnd_unlock A gL gL' i inv gE s (r : cont A) G : 
        assert G gL = Some gL' ->
        valid (gL' \+ gE) -> Inv (gL' \+ gE) inv ->
        (forall gE', 
           verify [mgh false, gL', i] gE' (s tt) r) ->
        verify [mgh true, gL, inv \+ i] gE (bind_s (unlock_s G) s) r.
Proof.
move=>/= Asrt V I T; apply: bnd_do=>/=; first by exists inv, gL'.
move=>[] gE' mL' gL'' i'; case/(_ _ _ (erefl _) Asrt V I)=>->->->.
by rewrite un0h; apply: T.
Qed.


Section CondList.
Variables (A B : Type) (xs : seq A) (s1 : spec B) (s2 : A -> seq A -> spec B).

Program
Definition Match_list xs (e1 : progsep s1) (e2 : forall x' xs', progsep (s2 x' xs'))
  : progsep (match xs with nil => s1 | cons x' xs' => s2 x' xs' end) :=
  match xs with nil => Do e1 | cons x' xs' => Do (e2 x' xs') end.

End CondList.

(*********************************************************)
(* rules for pulling out the spec vars from binary posts *)
(*********************************************************)

Lemma vrf_mono A lp ep s (r1 r2 : cont A) : 
        r1 <== r2 -> verify lp ep s r1 -> verify lp ep s r2.
Proof.
move=>T H1 sh mE hE C; case: (H1 sh _ _ C)=>S1 T1.
by split=>// x sh' lp' mE' gE' hE' C'; move/(T1 _ _ _ _ _ _ C'); apply: T.
Qed.

Lemma vrfV A lp gE s (r : cont A) : 
        (valid (gh lp \+ gE) -> verify lp gE s r) ->
        verify lp gE s r.
Proof. by move=>H sh mE hE C; apply: H (C); move: (cohVld C); case/and3P. Qed.

(* rule of conjunction *)

Lemma vrf_and A lp gE s (r1 r2 : cont A) : 
        verify lp gE s r1 -> verify lp gE s r2 ->
        verify lp gE s (fun y gE lp => r1 y gE lp /\ r2 y gE lp).
Proof.
move=>H1 H2 sh1 mE1 hE1 C1.
split=>[|x sh2 lp2 mE2 gE2 hE2 C2 Q]; first by case: (H1 _ _ _ C1).
split; first by move/proj2: (H1 _ _ _ C1)=>T; apply: T C2 Q.
by move/proj2: (H2 _ _ _ C1)=>T; apply: T C2 Q.
Qed.
 

Section Ghosts.
Variables (A : Type) (s : spec A) (p : pre).

Lemma ghA (B1 B2 : Type) (q : B1 -> B2 -> post A) :
       conseq s p (fun y gE1 gE2 lp1 lp2 => 
                     forall x1 x2, q x1 x2 y gE1 gE2 lp1 lp2) <->
       conseq s p (fun y gE1 gE2 lp1 lp2 => 
                     forall x, q x.1 x.2 y gE1 gE2 lp1 lp2).
Proof.
split=>H1 lp1 ep1 H2; move: (H1 lp1 ep1 H2);
apply: vrf_mono=>x' lp' ep' H; first by case; apply: H.
by move=>x1 x2; apply: (H (x1, x2)).
Qed.
   
Lemma ghAI B (q1 q2 : B -> pre) (r : B -> post A) :
        conseq s p (fun y gE1 gE2 lp1 lp2 => 
                      forall x, q1 x gE1 lp1 -> q2 x gE1 lp1 -> 
                                r x y gE1 gE2 lp1 lp2) <->
        conseq s p (fun y gE1 gE2 lp1 lp2 => 
                      forall x, q1 x gE1 lp1 /\ q2 x gE1 lp1 -> 
                                r x y gE1 gE2 lp1 lp2).
Proof.
split=>H1 lp1 ep1 H2; move: (H1 lp1 ep1 H2);
apply: vrf_mono=>x' lp' ep' H x; first by case; apply: H.
by move=>Q1 Q2; apply: H.
Qed.

Lemma ghE (B : Type) (q : B -> pre) (r : B -> post A) : 
        (forall lp1 gE1, p gE1 lp1 -> exists x, q x gE1 lp1) ->
        (forall lp1 gE1 x, 
           q x gE1 lp1 -> p gE1 lp1 -> valid (gh lp1 \+ gE1) ->
           verify lp1 gE1 s (fun y gE2 lp2 => r x y gE1 gE2 lp1 lp2)) ->
        conseq s p (fun y gE1 gE2 lp1 lp2 => 
                      forall x, q x gE1 lp1 -> r x y gE1 gE2 lp1 lp2).
Proof.
move=>H1 T1 lp1 ep1 S1 sh mE hE C; split.
- case: (H1 _ _ S1)=>x Q1; move: (T1 _ _ _ Q1 S1).
  by case/and3P: (cohVld C)=>_ C' _; case/(_ C' _ _ _ C).
move=>y' sh' lp' mE' gE' hE' C' T' x Q1'.
case/and3P: (cohVld C)=>_ V' _.
by move/proj2/(_ _ _ _ _ _ _ C' T'): (T1 _ _ _ Q1' S1 V' _ _ _ C).
Qed.

End Ghosts.

Definition ghs := (ghA, ghAI).

End Separation.

Implicit Arguments prog [R].
Implicit Arguments spec [R].
Implicit Arguments ret [R A].
Implicit Arguments read [R].
Implicit Arguments write [R A].
Implicit Arguments alloc [R A].
Implicit Arguments dealloc [R].
Implicit Arguments cas [R].
Implicit Arguments unlock [R].

Implicit Arguments par_do [R A B s1 s2 m1 m2 g1 g2 h1 h2 gE r].

(* some notation *)

Notation "'Do' e" := (do' e _) (at level 80).
Notation "x '<--' c1 ';' c2" := (bind c1 (fun x => c2)) 
  (at level 78, right associativity) : proc_scope.
Notation "c1 ';;' c2" := (bind c1 (fun _ => c2)) 
  (at level 78, right associativity) : proc_scope.
Notation "'!' x" := (read _ x) (at level 50) : proc_scope.
Notation "e1 '::=' e2" := (write e1 e2) (at level 60) : proc_scope.

Notation verify lp gE s r := (@verify' _ _ lp gE (fr s) r).



