Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq rels ordtype.
Require Import Eqdep domains prelude.
Require Import resources8 proc9 step9.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(***********************)
(* Strong bisimulation *)
(***********************)

Definition sbisimulation U A (R : Pred (tree U A * tree U A)) : Prop :=
  forall p q, (p, q) \In R -> 
    [/\ forall t p', step p t p' -> exists q', (p', q') \In R /\ step q t q',
        forall t q', step q t q' -> exists p', (p', q') \In R /\ step p t p' &
        p =t q].

Inductive sbisim U A (p q : tree U A) : Prop := 
  Sbisim R of sbisimulation R & (p, q) \In R. 

Notation "t1 ==~ t2" := (sbisim t1 t2) (at level 50) : proc_scope.

(* ==~ is an equivalence *)

Lemma sb_refl U A (t : tree U A) : t ==~ t.
Proof.
exists [Pred p q | p = q]=>// p q [->].
by split=>// t' p'; exists p'.
Qed.

Hint Resolve sb_refl.

Lemma sb_sym U A (p1 p2 : tree U A) : p1 ==~ p2 -> p2 ==~ p1. 
Proof.
case=>R S H; exists [Pred p q | (q, p) \In R]=>// p q.
by move/S=>[S1 S2 T]; split=>//; apply: eqt_sym.
Qed.

Lemma sb_symE U A (p1 p2 : tree U A) : p1 ==~ p2 <-> p2 ==~ p1.
Proof. by split; move/sb_sym. Qed.

Lemma sb_trans' U A (p1 p2 p3 : tree U A) : p1 ==~ p2 -> p2 ==~ p3 -> p1 ==~ p3.
Proof.
case=>R1 S1 H1 [R2] S2 H2.
exists [Pred p q | exists r, (p, r) \In R1 /\ (r, q) \In R2]; last by exists p2.
move=>{p1 p2 p3 H1 H2} p q [r][]; move/S1=>[H1 T1 E1]; move/S2=>[H2 T2 E2].
(split=>//; last by apply:(eqt_trans E1 E2));
move=>t p'; [case/H1|case/T2]=>q' [] H; [case/H2|case/T1]=>r' [] H';
by exists r'; split=>//; exists q'.
Qed.

Definition sb_trans U A p2 p1 p3 := @sb_trans' U A p1 p2 p3.

Implicit Arguments sb_trans [U A].

Add Parametric Relation U A : (tree U A) (@sbisim U A) 
  reflexivity proved by (@sb_refl _ _)
  symmetry proved by (@sb_sym _ _)
  transitivity proved by (@sb_trans' _ _) as sbisim_rel.

(* Extension of ==~ to functions *)

Definition sbfun U A B (f1 f2 : A -> tree U B) := 
  forall v : A, f1 v ==~ f2 v.

Notation "f1 ==~1 f2" := (sbfun f1 f2) 
  (at level 70, no associativity) : proc_scope.

(* ==~1 is an equivalence *)

Lemma sbfun_refl U A B (f : A -> tree U B) : f ==~1 f. 
Proof. by []. Qed.

Lemma sbfun_sym U A B (f1 f2 : A -> tree U B) : f1 ==~1 f2 -> f2 ==~1 f1.
Proof. by move=>H x; rewrite H. Qed.

Lemma sbfun_trans U A B (f1 f2 f3 : A -> tree U B) : 
        f1 ==~1 f2 -> f2 ==~1 f3 -> f1 ==~1 f3.
Proof. by move=>H1 H2 x; rewrite H1 H2. Qed.

Add Parametric Relation U A B : (A -> tree U B) (@sbfun U A B) 
  reflexivity proved by (@sbfun_refl _ _ _)
  symmetry proved by (@sbfun_sym _ _ _)
  transitivity proved by (@sbfun_trans _ _ _) as sbfun_rel.

(* Alternative formulation of ==~ *)
Lemma sbP U A (p q : tree U A) : 
        p ==~ q <->  
        [/\ forall t p', step p t p' -> exists q', p' ==~ q' /\ step q t q',
            forall t q', step q t q' -> exists p', q' ==~ p' /\ step p t p' &
            p =t q].
Proof.
split. 
- by case=>R S; move/S=>[H1 H2 E]; split=>// t p';
  [case/H1 | case/H2]=>q' []; exists q'; 
  split=>//; [|apply: sb_sym]; exists R.
pose P (p q : tree U A) := 
  [/\ forall t p', step p t p' -> exists q', p' ==~ q' /\ step q t q',
      forall t q', step q t q' -> exists p', q' ==~ p' /\ step p t p' &
      p =t q].
move=>[H1 H2 E]; exists [Pred p q | P p q]; last by rewrite /P InE /=.
move=>{p q H1 H2 E} p q; rewrite {1}/P InE /=;
move=>[H1 H2 E]; split=>//t p'; [case/H1|case/H2]=>q' [];
[|move/sb_sym]; move=>[R] S H' SS; exists q'; split=>//;
rewrite /P InE /= {t p q H1 H2 E SS}; move/S:H'=>[H1 H2];
by split=>{p} // t p; [case/H1|case/H2]=>q [S' R'];
  exists q; split=>//; [| apply: sb_sym]; exists R.
Qed.

Lemma sb_dummyP U A (p : tree U A) : p ==~ Dummy -> p = Dummy.
Proof. by case/sbP=>_ _; rewrite /eqt /= andbF. Qed.

Lemma sb_retP U A (p : tree U A) x : p ==~ Ret x -> p = Ret x.
Proof. by case/sbP=>_ _; rewrite /eqt /= andbF. Qed.

Prenex Implicits sb_dummyP sb_retP.

(*****************************)
(* Stepping and bisimulation *)
(*****************************)

(* step *)

Lemma stbP U A t (p1 p2 q1 : tree U A) : 
        p1 ==~ p2 -> step p1 t q1 -> exists q2, q1 ==~ q2 /\ step p2 t q2.
Proof. by case/sbP=>[H1 _ _]; move/H1. Qed.

Lemma stb_consP U A B (a : action A) t (p : tree U B) p' q g : 
        p ==~ Cons a g p' -> step p t q -> 
          [/\ A = Tr.typ t, jmeq action a (Tr.act t) & 
              forall pf, p' (coerce (T:=id) (Tr.val t) pf) ==~ q].
Proof.
move=>H; case/(stbP H)=>r' [H1].
case/st_consP=>E; rewrite {A}E in a p' H *; move/jmE=>E.
by rewrite {a}E in H * => T; split=>// pf; rewrite (T pf) H1.
Qed.

Lemma stb_consPQ U A B v (a : action A) (p : tree U B) p' g : 
        p ==~ Cons a g p' -> exists q, p' v ==~ q /\ step p [tr a, v] q.
Proof. by case/sbP=>[H1 H2 E]; apply:H2. Qed.

Lemma stb_consPt U A B (a : action A) v (p : tree U B) p' q g : 
        p ==~ Cons a g p' -> step p [tr a, v] q -> q ==~ (p' v).
Proof. 
by move=>H; case/(stb_consP H)=>_ _ T; rewrite (T (erefl _)).
Qed.

Lemma stb_parP U A B C t p (r' : A * B -> tree U C) p' q' s : 
        p ==~ Par p' q' r' -> step p t s -> 
          ((exists p1, s ==~ Par p1 q' r' /\ step p' t p1) \/
           (exists q1, s ==~ Par p' q1 r' /\ step q' t q1)) \/
          t = [idle] /\
            exists x, exists y,
              p' = Ret x /\ q' = Ret y /\ s ==~ r' (x, y).
Proof.
move=>H; case/(stbP H)=>y' [H1] S.
case/st_parP: S H H1; first by case; case=>u [->]; eauto.
by case=>-> [x][y][->][->] ->; eauto 10.
Qed.

Lemma stb_retparP U A B C p x y (r' : A * B -> tree U C) : 
        p ==~ Par (Ret x) (Ret y) r' ->
          exists p', r' (x, y) ==~ p' /\ step p [idle] p'.
Proof.
case=>R SB H; move:(SB _ _ H)=>[_ X2 _].
case:(X2 _ _ (st_retparP x y r'))=>p' [H' T'].
by exists p'; split=>//; apply: sb_sym; exists R.
Qed.

Lemma stb_lparP U A B C t p1 p2 q1 r (s : A * B -> tree U C) :
        p1 ==~ p2 -> step p1 t q1 -> 
        exists q2, q1 ==~ q2 /\ step (Par p2 r s) t (Par q2 r s).
Proof.
move=>H; case/(stbP H)=>q2 [H1 H2]; exists q2.
by split=>//; apply: st_lparP.
Qed.

Lemma stb_rparP U A B C t p1 p2 q1 r (s : B * A -> tree U C) :
        p1 ==~ p2 -> step p1 t q1 -> 
        exists q2, q1 ==~ q2 /\ step (Par r p2 s) t (Par r q2 s).
Proof.
move=>H; case/(stbP H)=>q2 [H1 H2]; exists q2.
by split=>//; apply: st_rparP.
Qed.

(***********************************************)
(* ==~ is a congruence (not for fork-join par) *)
(***********************************************)

Lemma sb_comp U T A B C (r1 r2 : A * B -> tree U C) (f : T -> A * B) : 
        r1 ==~1 r2 -> r1 \o f ==~1 r2 \o f.
Proof. by move=>H x; apply: H. Qed.

Add Parametric Morphism U T A B C : (@funcomp (tree U C) (A * B) T tt) with signature 
      @sbfun U (A * B) C ===> @eq _ ===> @sbfun _ _ _ as sb_comp_morph.
Proof. by move=>r1 r2 H f; apply: sb_comp H. Qed. 

Lemma sb_consP U A B (a : action A) g1 g2 (p1 p2 : A -> tree U B) : 
        Cons a g1 p1 ==~ Cons a g2 p2 <-> p1 ==~1 p2.
Proof.
split=>[|T].
- move/sbP=>[_ T _] v; case: (T _ _ (st_consPQ a _ v g2))=>p1' [H3].
  by case/st_consP=>_ _ /= E; rewrite H3 (E (erefl _)).
apply/sbP; split=>// t p';
case/st_consP=>E; rewrite {A}E in a p1 p2 T *;
move: {T} (T (Tr.val t)); [|move/sb_sym]=>T; move/jmE=>->;
move/(_ (erefl _)); rewrite eqc => <-; 
[exists (p2 (Tr.val t)) | exists (p1 (Tr.val t))];
by split=>//; apply: st_cons.
Qed.

Add Parametric Morphism U A B (a : action A) g : (Cons a g) with signature 
    @sbfun U A B ===> @sbisim U B as sb_pcons_morph.
Proof. by move=>p1 p2 H; apply/sb_consP. Qed.

Lemma sb_consT U A1 A2 B (a1 : action A1) (a2 : action A2) 
               (p1 : A1 -> tree U B) (p2 : A2 -> tree U B) g1 g2 : 
        A1 + A2 -> Cons a1 g1 p1 ==~ Cons a2 g2 p2 -> 
          [/\ A1 = A2, jmeq action a1 a2 &
              forall pf v, p1 v ==~ p2 (coerce (T:=id) v pf)].
Proof.
case=>v;
move/sbP=>[H1 H2 _];
[case: (H1 _ _ (st_consPQ a1 _ v g1)) | 
 case: (H2 _ _ (st_consPQ a2 _ v g2))]=>p [H3];
case/st_consP=>/= E;
[rewrite {A2}E in a2 p2 H1 H2 * | rewrite {A1}E in a1 p1 H1 H2 *];
move/jmE=>E;
[rewrite {a2}E in H1 H2 * | 
 rewrite {a1}E in H1 H2 *];
move=>E; split=>// pf w; rewrite eqc;
move: w; [apply/(sb_consP a1 g1 g2) | apply/(sb_consP a2 g1 g2)];
by apply/sbP; split.
Qed.

(* commutativity of par *)
Lemma sb_parC U A B C p1 p2 (r : A * B -> tree U C) :  
         Par p1 p2 r ==~ Par p2 p1 (r \o swap).
Proof.
pose P A B C p1 p2 := 
  p1 = p2 \/ exists s1, exists s2, exists r : A * B -> tree U C, 
               p1 = Par s1 s2 r /\ p2 = Par s2 s1 (r \o swap). 
exists [Pred p1 p2 | P A B C p1 p2]; last first.
- by right; exists p1; exists p2; exists r.
have L: forall A B C p1 p2 (r : A * B -> tree U C) t p', 
          step (Par p1 p2 r) t p' -> 
          exists q', P A B C p' q' /\ step (Par p2 p1 (r \o swap)) t q'.
- move=>{A B C p1 p2 r} A B C p1 p2 r t p'; rewrite /P. 
  case/st_parP; last by case=>-> [x][y][->][->] ->; eauto. 
  case; case=>s [->] H;
  [exists ((Par p2 s (r \o swap)))|exists ((Par s p1 (r \o swap)))]; by eauto 8.
move=>{p1 p2 r} p q [->|[s1][s2][r][->][->] {p q} /=].
- by split=>// t p' H; exists p'; split=>//; left.
split=>// t p'; case/L=>q' [H1 H2]; exists q'; first by [].
rewrite compA swapI -compf1 in H2; split=>//.
case: H1=>[->|[p1][p2][r2][->][->]]; first by left. 
right; exists p2; exists p1; exists (r2 \o swap); split=>//.
by rewrite compA swapI -compf1.
Qed.

Lemma sb_par_cong U A B C p1 p2 q1 q2 (r1 r2 : A * B -> tree U C) : 
        p1 ==~ p2 -> q1 ==~ q2 -> r1 ==~1 r2 -> Par p1 q1 r1 ==~ Par p2 q2 r2. 
Proof.
suff L: forall A B C p1 p2 q (r1 r2 : A * B -> tree U C), 
          p1 ==~ p2 -> r1 ==~1 r2 -> Par p1 q r1 ==~ Par p2 q r2.
- move=>H1 H2 H3.
  apply: (sb_trans (Par p1 q2 r2)); last by apply: L.
  by rewrite 2!(sb_parC p1); apply: L H2 (sb_comp _ H3).
pose P A B C p1 p2 := 
  p1 ==~ p2 \/ 
  exists s1, exists s2, exists r1 : A * B -> tree U C, exists r2, exists q, 
    p1 = Par s1 q r1 /\ p2 = Par s2 q r2 /\ s1 ==~ s2 /\ r1 ==~1 r2.
have L: forall A B C p1 p2 (r1 r2 : A * B -> tree U C) q t p1', 
  p1 ==~ p2 -> r1 ==~1 r2 -> step (Par p1 q r1) t p1' -> 
  exists p2', P A B C p1' p2' /\ step (Par p2 q r2) t p2'.
- move=>{A B C p1 p2 q1 q2 r1 r2} A B C p1 p2 r1 r2 q t p1' H1 H2; rewrite /P.
  case/st_parP; last first.
  - move=>H; case: H H1=>-> [x][y][->][->] ->; move/sb_sym.
    by move/sb_retP=>->; eauto.
  case; case=>q' [->]; last by exists (Par p2 q' r2); split; [eauto 10| eauto].
  case/(stbP H1)=>p2' [H4 H5]; exists (Par p2' q r2); 
  by split; [eauto 10 | eauto].
move=>{A B C p1 p2 q1 q2 r1 r2} A B C p1 p2 q r1 r2 P1 R1. 
exists [Pred p1' p2' | P A B C p1' p2']; last by right; eauto 10.
move=>v1 v2 [|[s1][s2][t1][t2][q'][->][->][S1 T1] {v1 v2} /=]. 
- by move/sbP=>[H1 H2 E]; split=>// t p'; [case/H1|case/H2]=>q' []; 
     exists q'; split=>//; left; [|apply: sb_sym].
split=>// t p'; first by case/(L _ _ _ _ _ _ _ _ _ _ S1 T1)=>p2' []; exists p2'.
case/(L _ _ _ _ _ _ _ _ _ _ (sb_sym S1) (sbfun_sym T1))=>p2' [H1 H2].
exists p2'; split=>//.
case: H1; first by move/sb_sym; left.
case=>s3 [s4][r3][r4][q2'][->][->][S2 R2]. 
right; exists s4; exists s3; exists r4; exists r3; exists q2'.
by do !split=>//; rewrite 1?S2 1?R2.
Qed.

Add Parametric Morphism U A B C: (@Par U A B C) with signature 
    @sbisim _ _ ===> @sbisim _ _ ===> @sbfun _ _ _ ===> @sbisim _ _ as sb_par_morph.
Proof. by move=>*; apply: sb_par_cong. Qed.

(************************************)
(* parallel composition and mapping *)
(************************************)

Lemma sb_rparM U A B B' C p q (f : B' -> B) 
               (r : A * B -> tree U C) :
        Par p (tmap f q) r ==~ Par p q (r \o (id \* f)).
Proof.
pose F A B B' C q1 q2 := 
  q1 = q2 \/
  exists p, exists q, exists f : B' -> B, exists r : A * B -> tree U C,
    q1 = Par p (tmap f q) r /\ q2 = Par p q (r \o (id \* f)).
exists [Pred q1 q2 | F A B B' C q1 q2]; last by right; eauto 10.
move=>{p q f r} q1 q2 [].
- by move=>->; split=>// t p' H; exists p'; split=>//; left.
case=>p1 [q][f][r][->{q1}] ->{q2}; split=>//.
- move=>t p'; case/st_parP; last first.
  - case=>-> [y][z][->][]; case: q=>//= x [<-] ->. 
    by exists (r (y, f x)); split; [left | apply: st_retpar].
  case=>[[p2][->] H|[p2][->]].
  - exists (Par p2 q (r \o id \* f)).
    by split; [right; eauto 10 | apply: st_lparP H].
  case/st_tmapP=>q' [->] H; exists (Par p1 q' (r \o id \* f)).
  by split; [right; eauto 10 | apply: st_rparP].
move=>t p'; case/st_parP; last first.
- case=>-> [y][z][->][]; case: q=>//= x [<-] ->. 
  by exists (r (y, f x)); split; [left | apply: st_retpar].
case=>[[p2][->] H|[p2][->] H]. 
- exists (Par p2 (tmap f q) r).
  by split; [right; eauto 10 | apply: st_lparP H].
exists (Par p1 (tmap f p2) r).
by split; [right; eauto 10 | apply: st_rparP; apply: st_tmapPQ].
Qed.

Lemma sb_lparM U A A' B C p q (f : A' -> A) 
               (r : A * B -> tree U C) :
        Par (tmap f p) q r ==~ Par p q (r \o (f \* id)).
Proof. by rewrite 2!(sb_parC _ q) sb_rparM !compA swap_prod. Qed.

(***********************)
(* Monad laws for tcat *)
(***********************)

Lemma unit_l U A B (x : A) (f : A -> tree U B) :
        tcat (Ret x) f ==~ f x.
Proof. by []. Qed.

Lemma unit_r U A (q : tree U A) :
        tcat q (fun x => Ret x) ==~ q.
Proof.
elim:q=>A' //.
- by move=>B a g t IH; (*rewrite /tcat;*) apply/sb_consP.
by move=>B1 B2 t1 H1 t2 H2 r IH; (*rewrite /tcat;*) apply: sb_par_cong =>//.
Qed.

Lemma assoc U A B C (p : tree U A) (q : A -> tree U B) (r : B -> tree U C) :
        tcat (tcat p q) r ==~ tcat p (fun x => tcat (q x) r).
Proof.
elim:p q=>[||A' B' a g t IH|A' B1 B2 t1 H1 t2 H2 t' IH'] q //=.
- by apply/sb_consP=>b; apply: IH.
by apply: sb_par_cong =>// b12; apply: IH'.
Qed.

(*
Hint Resolve sb_refl.

Implicit Arguments sb_trans [U A].

Prenex Implicits sb_dummyP sb_retP.

Notation "t1 ==~ t2" := (sbisim t1 t2) (at level 50) : proc_scope.

Notation "f1 ==~1 f2" := (sbfun f1 f2) 
  (at level 70, no associativity) : proc_scope.
*)

