Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq rels ordtype.
Require Import Eqdep prelude heaps.
Require Import pcm resources8 proc9.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(************)
(* Stepping *)
(************)

Section Stepping.
Variables U : pcm.
Notation tree A := (@tree U A).
Notation Dummy A := (@Dummy U A).
Notation Ret x := (@Ret U _ x).

Inductive step A (p : tree A) (t : trans) (q : tree A) : Prop :=
| st_cons g1 p1 of p = Cons (Tr.act t) g1 p1 & q = p1 (Tr.val t)
| st_lpar A1 A2 (p1 : tree A1) (p2 : tree A2) (p1' : tree A1) r of 
    p = Par p1 p2 r & 
    q = Par p1' p2 r & @step A1 p1 t p1'
| st_rpar A1 A2 (p1 : tree A1) (p2 : tree A2) (p2' : tree A2) r of 
    p = Par p1 p2 r & 
    q = Par p1 p2' r & @step A2 p2 t p2'
| st_retpar A1 A2 (x1 : A1) (x2 : A2) r of 
    t = [idle] & p = Par (Ret x1) (Ret x2) r & q = r (x1, x2).

Hint Constructors step.

(* views for "evaluation by rewriting" *)

Lemma st_dummyP A t q : step (Dummy A) t q <-> False.
Proof. by split=>//; case. Qed.

Lemma st_retP A (x : A) t q : step (Ret x) t q <-> False.
Proof. by split=>//; case. Qed. 

Lemma st_consP A B (a : action A) t (p : A -> tree B) q g : 
        step (Cons a g p) t q <->
          [/\ A = Tr.typ t, jmeq action a (Tr.act t) & 
              forall pf, p (coerce (T:=id) (Tr.val t) pf) = q].
Proof.
split.
- case=>// g1 p1; case/consE=>E; rewrite {A}E in a p *.
  by move/jmE=>-> _; move/jmE=>->->; split=>// pf; rewrite eqc.
case=>E; rewrite {A}E in a p *; move/jmE=>-> H.
by apply: st_cons=>//; rewrite (H (erefl _)).
Qed.

Lemma st_consPQ A B (a : action A) (p : A -> tree B) v g : 
        step (Cons a g p) [tr a, v] (p v).
Proof. by apply/st_consP; split=>//= pf; rewrite eqc. Qed.

Hint Resolve st_consPQ.

Lemma st_consPT A B (a : action A) (p : A -> tree B) v q g : 
        step (Cons a g p) [tr a, v] q <-> q = p v.
Proof. by split=>[|->] //; case/st_consP=>/= _ _; move/(_ (erefl _))=>->. Qed.

Lemma st_parP A1 A2 B (p : tree A1) (q : tree A2) (r : _ -> tree B) t s : 
        step (Par p q r) t s <-> 
          ((exists p1, s = Par p1 q r /\ step p t p1) \/ 
           (exists q1, s = Par p q1 r /\ step q t q1)) \/
              t = [idle] /\ exists x, exists y, 
                            p = Ret x /\ q = Ret y /\ s = r (x, y).
Proof.
split; last first.
- by case=>[[[p1][->]|[q1][->]]|[->][x1][x2][->][->] ->];
     [apply: st_lpar | apply: st_rpar | apply: st_retpar].
case=>//.
- move=>A3 A4 p1 p2 p1' r'; case/parE=>E1 E2.
  rewrite -{A3}E1 -{A4}E2 in p1 p2 p1' r' *.
  move/jmE=>->; move/jmE=>->; move/jm2E=>->-> H.
  by left; left; exists p1'.
- move=>A3 A4 p1 p2 p2' r'; case/parE=>E1 E2.
  rewrite -{A3}E1 -{A4}E2 in p1 p2 p2' r' *.
  move/jmE=>->; move/jmE=>->; move/jm2E=>->-> H.
  by left; right; exists p2'.
move=>A3 A4 x1 x2 r' ->; case/parE=>E1 E2.
rewrite -{A3}E1 -{A4}E2 in x1 x2 r' *.
move/jmE=>->; move/jmE=>->; move/jm2E=>->->.
by right; split=>//; exists x1; exists x2.
Qed.

Lemma st_parPQ A1 A2 B (p1 p2 : tree A1) (q1 q2 : tree A2) 
                       (r1 r2 : _ -> tree B) t : 
        step (Par p1 q1 r1) t (Par p2 q2 r2) <-> 
        (r1 = r2 /\ (q2 = q1 /\ step p1 t p2 \/ p2 = p1 /\ step q1 t q2)) \/
        t = [idle] /\ exists x1, exists x2, 
                      p1 = Ret x1 /\ q1 = Ret x2 /\ r1 (x1, x2) = Par p2 q2 r2.
Proof.
rewrite st_parP; split; last first.
- case.
  - by case=>-> []; case=>-> H; left; [left; exists p2 | right; exists q2].
  by case=>-> [x1][x2][->][->] H; right; split=>//; exists x1; exists x2.
case; last first.
- by case=>-> [x][y][->][->]; right; split=>//; exists x; exists y.
by case; case=>p3 []; case/parE=>_ _; 
   move/jmE=>->; move/jmE=>->; move/jm2E=>->; left; split=>//;
   [left | right].
Qed.

Lemma st_lparP A B C (p1 p2 : tree A) (q : tree B) (r : _ -> tree C) t :
        step p1 t p2 -> step (Par p1 q r) t (Par p2 q r).
Proof. by rewrite st_parPQ; eauto. Qed.

Lemma st_rparP A B C (p : tree A) (q1 q2 : tree B) (r : _ -> tree C) t :
        step q1 t q2 -> step (Par p q1 r) t (Par p q2 r).
Proof. by rewrite st_parPQ; eauto. Qed.  

Lemma st_retparP A B C (x : A) (y : B) (r : _ -> tree C) :
        step (Par (Ret x) (Ret y) r) [idle] (r (x, y)).
Proof. by rewrite st_parP; right; eauto 10. Qed.

Hint Resolve st_retparP.

Lemma st_tmapP A B (f : A -> B) (p : tree A) t q : 
       step (tmap f p) t q <-> exists r, q = tmap f r /\ step p t r.
Proof.
case:p t q=>[|a|B' a g t'|B1 B2 t1 t2 t] r q /=.
- by split=>[|[s][->]]; move/st_dummyP.
- by split=>[|[s][->]]; move/st_retP.
- split=>[|[s][->]]; 
  case/st_consP=>E; rewrite {B'}E in a t' *;
  move/jmE=>->; move/(_ (erefl _))=><- /=;
  [exists (t' (Tr.val r)); split=>//|];
  by apply/st_consP; split=>// pf; rewrite eqc.
split=>[|[s][->]]; last first.
- case/st_parP=>[|[->][x][y][->][->] ->]; last by apply: st_retpar.
  by case; case=>p1 [->] H; [apply: st_lparP | by apply: st_rparP].
case/st_parP=>[|[->][x][y][->][->] ->]; last first.
- by exists (t (x, y)); split=>//; apply: st_retpar.
case; case=>p1 [->] H.
- by exists (Par p1 t2 t); split=>//; apply: st_lparP.
by exists (Par t1 p1 t); split=>//; apply: st_rparP.
Qed.

Lemma st_tmapPQ A B (f : A -> B) (p : tree A) t q : 
       step p t q -> step (tmap f p) t (tmap f q).
Proof. by move=>H; apply/st_tmapP; exists q. Qed.

Lemma st_tcatP A B (p : tree A) (q : A -> tree B) t r : 
        step (tcat p q) t r <-> 
          (exists p', r = tcat p' q /\ step p t p') \/ 
           exists x, p = Ret x /\ step (q x) t r.
Proof.
case:p q t r=>[|a|B' a g t'|B1 B2 t1 t2 t3] q t r /=.
- by split=>[|[][p][] // _]; case/st_dummyP. 
- split; first by right; exists a.
  by case; [case=>p [->]; case/st_retP | case=>x [[->]]].
- case: t=>T b v; split.
  - case/st_consP=>/= E; rewrite {B'}E in a t' *; move/jmE=>->.
    by move/(_ (erefl _))=>/= <-; left; exists (t' v).
  case=>[[t3][->]|[x][//]].
  case/st_consP=>/= E; rewrite {B'}E in a t' *; move/jmE=>->.
  by move/(_ (erefl _))=>/= <-; apply/st_consPT.
split; last first.
- case=>[[p][->]|[x][//]].
  case/st_parP; last by case=>-> [x][y][->][->] ->; apply: st_retparP.
  by case; case=>p' [->] S; [apply: st_lparP | apply: st_rparP].
case/st_parP; last by case=>-> [x][y][->][->] ->; eauto.
case; case=>p [->] S; left.
- by exists (Par p t2 t3); split=>//; apply: st_lparP.
by exists (Par t1 p t3); split=>//; apply: st_rparP.
Qed.

(******************************)
(* Transitive closure of step *)
(******************************)

Fixpoint stept A ts (p q : tree A) : Prop := 
  if ts is t::ts' then exists r, step p t r /\ stept ts' r q
  else p = q.

Lemma stt1 A t (p q : tree A) : step p t q -> stept [:: t] p q.
Proof. by move=>S; exists q. Qed.

Lemma stt_catE A (p : tree A) r ts1 ts2 : 
        stept (ts1 ++ ts2) p r <-> exists q, stept ts1 p q /\ stept ts2 q r.
Proof.
elim: ts1 p=>[|t ts1 IH] p /=.
- by split; [eauto | case=>[q][<-]].
split; first by move=>[s][H1]; move/IH=>[q][H2] H3; eauto.
by move=>[q][[s]][H1] H2 H3; exists s; rewrite IH; eauto.
Qed.

Lemma stt_catI A ts1 ts2 (p1 p2 p3 : tree A) :
        stept ts1 p1 p2 -> stept ts2 p2 p3 -> stept (ts1 ++ ts2) p1 p3.
Proof. by move=>H1 H2; apply/stt_catE; exists p2. Qed.

Lemma stt_dummyP A ts (p : tree A) : 
        stept ts (Dummy A) p <-> ts = [::] /\ p = Dummy A.
Proof.
elim:ts p=>[|t ts IH] p /=; first by split; case. 
by split; case=>// r []; case/st_dummyP.
Qed.

Lemma stt_retP A ts (v : A) p : 
        stept ts (Ret v) p <-> ts = [::] /\ p = Ret v.
Proof.
elim:ts v p=>[|t ts IH] p /=; first by split; case. 
by split; case=>// r []; case/st_retP.
Qed.

Lemma stt_consP A ts B (a : action B) p (q : tree A) g : 
        stept ts (Cons a g p) q <-> 
          if ts is t::ts' then exists v, t = [tr a, v] /\ stept ts' (p v) q
          else q = Cons a g p.
Proof.
case:ts B a p q=>[|t ts] B a p q //=.
split; last by case=>v [->]; eauto.
case=>r []; case/st_consP=>E; rewrite {B}E in a p *; move/jmE=>-> T.
by rewrite -(T (erefl _)) /=; exists (Tr.val t); rewrite -tr_eta.
Qed.

Lemma stt_consPQ A B (a : action B) p (q : tree A) v ts g : 
        stept ([tr a, v]::ts) (Cons a g p) q <-> stept ts (p v) q.
Proof.
rewrite stt_consP; split; last by eauto.
by case=>x []; rewrite trD; case=>_ ->.
Qed.

Lemma stt_lparP A B C (p1 p2 : tree A) (q : tree B) ts (r : A * B -> tree C) : 
        stept ts p1 p2 -> stept ts (Par p1 q r) (Par p2 q r).
Proof.
elim:ts p1 p2 r=>[|t ts IH] p1 p2 r /=.
- by move=>->.
case=>s [H1] H2; exists (Par s q r); split;
by [apply/st_parPQ; auto | apply: IH H2].
Qed.

Lemma stt_rparP A B C (p : tree A) (q1 q2 : tree B) ts (r : A * B -> tree C) : 
        stept ts q1 q2 -> stept ts (Par p q1 r) (Par p q2 r).
Proof.
elim: ts q1 q2 r=>[|t ts IH] q1 q2 r /=.
- by move=>->.
case=>s [H1] H2; exists (Par p s r); split;
by [apply/st_parPQ; auto | apply: IH H2].
Qed.

(* we often only care about the existence of a transition sequence *)
Inductive steps A (p q : tree A) : Prop := stepsI ts of stept ts p q.

Lemma sts_refl A (p : tree A) : steps p p.
Proof. by exists [::]. Qed.

Hint Resolve sts_refl.

Lemma sts_trans' A (p1 p2 p3 : tree A) : 
        steps p1 p2 -> steps p2 p3 -> steps p1 p3.
Proof. by case=>s1 H1[s2 H2]; exists (s1 ++ s2); apply/stt_catE; eauto. Qed.

Definition sts_trans A p2 p1 p3 := @sts_trans' A p1 p2 p3.

Implicit Arguments sts_trans [A p1 p3].

Add Parametric Relation A : (tree A) (@steps A)
  reflexivity proved by (@sts_refl A)
  transitivity proved by (@sts_trans' A) as steps_rel.

Lemma sts_dummyP A (p : tree A) : steps (Dummy A) p <-> p = Dummy A.
Proof. by split=>[|->] // [ts]; case/stt_dummyP. Qed.

Lemma sts_retP A v (p : tree A) : steps (Ret v) p <-> p = Ret v. 
Proof. by split=>[|->] // [ts]; case/stt_retP. Qed.

Lemma sts_consP A B (a : action B) p (q : tree A) g : 
        steps (Cons a g p) q <-> 
          q = Cons a g p \/ 
            exists v, exists r, step (Cons a g p) [tr a, v] r /\ steps r q.
Proof.
split.
- case=>ts; move/stt_consP; case: ts=>[|t ts']; first by left.
  by case=>v [_] S; right; exists v; exists (p v); split=>//; exists ts'.
case=>[->//|[v][r][]]; case/st_consP=>_ _; move/(_ (erefl _))=><- /= [ts S].
by exists ([tr a, v]::ts); exists (p v).
Qed.

Lemma sts_lparP A B C (p1 p2 : tree A) (q : tree B) (r : A * B -> tree C) : 
        steps p1 p2 -> steps (Par p1 q r) (Par p2 q r). 
Proof. by case=>ts H; exists ts; apply: stt_lparP H. Qed.

Lemma sts_rparP A B C (p : tree A) (q1 q2 : tree B) (r : A * B -> tree C) : 
        steps q1 q2 -> steps (Par p q1 r) (Par p q2 r). 
Proof. by case=>ts H; exists ts; apply: stt_rparP H. Qed.

Lemma sts_parP A B C (p1 : tree A) (p2 : tree B) (q : A * B -> tree C) r :
        steps (Par p1 p2 q) r <->
        (exists s1, exists s2, 
          [/\ r = Par s1 s2 q, steps p1 s1 & steps p2 s2]) \/
         exists x1, exists x2, 
          [/\ steps p1 (Ret x1), steps p2 (Ret x2) & steps (q (x1, x2)) r].
Proof.
split; last first.
- case=>[[s1][s2][->]H1 H2|[x1][x2][H1] H2 H3].
  - apply: (sts_trans (Par s1 p2 q)); 
    by [apply: sts_lparP | apply: sts_rparP].
  apply: sts_trans H3.
  apply: (sts_trans (Par (Ret x1) p2 q)); first by apply: sts_lparP.
  apply: (sts_trans (Par (Ret x1) (Ret x2) q)); first by apply: sts_rparP.
  by exists [:: [idle]]; exists (q (x1, x2)); split=>//. 
  (* by apply: st_retpar. *)
case=>ts; elim: ts p1 p2 q r=>[|t ts IH] p1 p2 q r /=.
- by move=><-; left; exists p1; exists p2.
case=>s []; case/st_parP=>[|[_][x][y][->][->] -> H1]; last first.
- by right; exists x; exists y; do !split=>//; exists ts.
case; case=>p [->] H1.
- case/IH=>[[s1][s2][->][rs H2] H3|[x1][x2][[rs H2]] H3 H4].
  - left; exists s1; exists s2; do !split=>//.
    by exists (t :: rs); exists p.
  right; exists x1; exists x2; do !split=>//.
  by exists (t :: rs); exists p.
case/IH=>[[s1][s2][->] H2 [rs H3]|[x1][x2][H2][rs H3] H4].
- left; exists s1; exists s2; do !split=>//.
  by exists (t :: rs); exists p.
right; exists x1; exists x2; do !split=>//.
by exists (t :: rs); exists p.
Qed. 

(*****************)
(* Weak stepping *)
(*****************)

(* Weak stepping modes out by finitely many idles. *)

Fixpoint stutter A n (p1 p2 : tree A) := 
  if n is n'.+1 then exists r, step p1 [idle] r /\ stutter n' r p2
  else p1 = p2.

Lemma stutter1 A (p1 p2 : tree A) : step p1 [idle] p2 <-> stutter 1 p1 p2.
Proof.
split=>//; first by move=>S; exists p2.
by move=>[p2'][S]S0; rewrite -S0.
Qed.

Lemma stutter_stt A n (p q : tree A) : 
        stutter n p q <-> stept (nseq n [idle]) p q. 
Proof. by elim:n p q=>[|n IH] p q //=; setoid_rewrite IH. Qed.

Lemma stutternS A n (p1 p2 : tree A) : 
        stutter n.+1 p1 p2 <-> exists r, stutter n p1 r /\ step r [idle] p2.
Proof.
elim:n p1 p2=>[|n /= IH] p1 p2 /=.
- by split; [case=>r [H] <- | case=>r [<-] H]; eauto.
split; first by case=>r [H]; rewrite IH=>[[r']][]; eauto.
by case=>r [[r']][H1] H2 H3; exists r'; rewrite IH; eauto.
Qed.

Lemma stutter_add A n1 n2 (p1 p2 : tree A) : 
        stutter (n1 + n2) p1 p2 <-> 
          exists q, stutter n1 p1 q /\ stutter n2 q p2.
Proof.
elim:n1 n2 p1 p2=>[|n1 IH] n2 t1 t3; rewrite ?add0n ?addSn /=.
- by split; [eauto | case=> t2 [->]].
split; first by case=>r [H1]; move/IH=>[t2][]; eauto.
by case=>[t2][[r]][H1] H2 H3; exists r; rewrite IH; eauto.
Qed.

Lemma stutter_sts A n (p q : tree A) : stutter n p q -> steps p q.
Proof. by move/stutter_stt; exists (nseq n [idle]). Qed.

Lemma stutter_dummyP A n (p : tree A) : 
        stutter n (Dummy A) p <-> n = 0 /\ p = Dummy A.
Proof.
case:n p=>[|n] p /=; first by split=>//; case.
by split=>[[r][]|[//]]; case/st_dummyP.
Qed.

Lemma stutter_retP A n v (p : tree A) : 
        stutter n (Ret v) p <-> n = 0 /\ p = Ret v.
Proof.
case:n v p=>[|n] p /=; first by split=>//; case.
by split=>[[r][]|[//]]; case/st_retP.
Qed.

Lemma stutter_consP A B n (a : action B) p (q : tree A) g : 
       stutter n (Cons a g p) q <->
         if n is n'.+1 then 
           [/\ B = unit, jmeq action a Idle & 
               forall pf, stutter n' (p (coerce (T:=id) tt pf)) q]
         else q = Cons a g p.
Proof.
case:n=>[|n]//=; split=>[[r][]|[E]];
[case/st_consP=>/= E|]; rewrite {B}E in a p *; move/jmE=>->.
- by move=>E; split=>// pf; rewrite E.
by move/(_ (erefl _)); rewrite eqc; eauto.
Qed.

Lemma stutter_lparP A B C q n p1 p2 (r : A * B -> tree C) :
        stutter n p1 p2 -> stutter n (Par p1 q r) (Par p2 q r).
Proof.
elim: n q p1 p2 r=>[|n IH] q p1 p2 r /=; first by move=>->.
move=>[p'][S1 Sn]; exists (Par p' q r).
by split=>//; [apply: st_lpar| apply: IH Sn].
Qed.

Lemma stutter_rparP A B C p n q1 q2 (r : A * B -> tree C) :
        stutter n q1 q2 -> stutter n (Par p q1 r) (Par p q2 r).
Proof.
elim: n p q1 q2 r=>[|n IH] p q1 q2 r /=; first by move=>->.
move=>[q'][S1 Sn]; exists (Par p q' r).
by split=>//; [apply: st_rpar | apply: IH Sn].
Qed.

Inductive stutt A (p1 p2 : tree A) : Prop := stuttI n of stutter n p1 p2.

Lemma stutt_stt A (p q : tree A) : 
        stutt p q <-> exists n, stept (nseq n [idle]) p q.
Proof.
by split; case=>n; [exists n; apply/stutter_stt | move/stutter_stt; exists n].
Qed.

Lemma stutt_refl A (p : tree A) : stutt p p.
Proof. by exists 0. Qed.

Hint Resolve stutt_refl.

Lemma stutt_trans' A (p1 p2 p3 : tree A) : 
        stutt p1 p2 -> stutt p2 p3 -> stutt p1 p3.
Proof. 
case=>m M[n N]; exists (m + n).
by apply/stutter_add; eauto.
Qed.

Definition stutt_trans A p2 p1 p3 := @stutt_trans' A p1 p2 p3.

Implicit Arguments stutt_trans [A p1 p3].

Add Parametric Relation A : (tree A) (@stutt A)
  reflexivity proved by (@stutt_refl A)
  transitivity proved by (@stutt_trans' A) as stutts.

Lemma stutt1 A (p1 p2 : tree A) : step p1 [idle] p2 -> stutt p1 p2.
Proof. by move=>H; exists 1; exists p2; split; [apply H|]. Qed.

Lemma stutt_sts A (p q : tree A) : stutt p q -> steps p q.
Proof. by case=>n; move/stutter_sts. Qed.

Lemma stutt_dummyP A (p : tree A) : stutt (Dummy A) p <-> p = Dummy A.
Proof. by split=>[[n]|->//]; case/stutter_dummyP. Qed.

Lemma stutt_retP A v (p : tree A) : stutt (Ret v) p <-> p = Ret v.
Proof. by split=>[[n]|->//]; case/stutter_retP. Qed.

Lemma stutt_consP A B (a : action B) p (q : tree A) g : 
        stutt (Cons a g p) q <-> 
        q = Cons a g p \/ 
        [/\ B = unit, jmeq action a Idle & 
            forall pf, stutt (p (coerce (T:=id) tt pf)) q].
Proof.
split. 
- case; case=>[|n /= [r][]]; first by move/stutter_consP; auto.
  case/st_consP=>/= H1 H2 H3 H4.
  by right; split=>// pf; exists n; rewrite H3.
case=>[->|[E]]; first by auto.
rewrite {B}E in a p *; move/jmE=>->; move/(_ (erefl _)); rewrite eqc. 
by case=>n H; exists n.+1; exists (p tt).
Qed.

Lemma stutt_lparP A B C p1 p2 q (r : A * B -> tree C) : 
        stutt p1 p2 -> stutt (Par p1 q r) (Par p2 q r).
Proof. by case=>n; move/stutter_lparP; exists n. Qed.

Lemma stutt_rparP A B C p q1 q2 (r : A * B -> tree C) : 
        stutt q1 q2 -> stutt (Par p q1 r) (Par p q2 r).
Proof. by case=>n; move/stutter_rparP; exists n. Qed.

Lemma stutt_parP A B C p1 p2 (q : A * B -> tree C) r :
        stutt (Par p1 p2 q) r <->
        (exists s1, exists s2, 
          [/\ r = Par s1 s2 q, stutt p1 s1 & stutt p2 s2]) \/
         exists x1, exists x2, 
          [/\ stutt p1 (Ret x1), stutt p2 (Ret x2) & stutt (q (x1, x2)) r].
Proof.
split; last first.
- case=>[[s1][s2][->]H1 H2|[x1][x2][H1] H2 H3].
  - apply: (stutt_trans (Par s1 p2 q)); 
    by [apply: stutt_lparP | apply: stutt_rparP].
  apply: stutt_trans H3.
  apply: (stutt_trans (Par (Ret x1) p2 q)); first by apply: stutt_lparP.
  apply: (stutt_trans (Par (Ret x1) (Ret x2) q)); first by apply: stutt_rparP.
  by exists 1; exists (q (x1, x2)); split=>//.
  (* by apply: st_retpar. *)
case=>n; elim: n p1 p2 q r=>[|n IH] p1 p2 q r /=.
- by move=><-; left; exists p1; exists p2.
case=>s []; case/st_parP=>[|[_][x][y][->][->] -> H1]; last first.
- by right; exists x; exists y; do !split=>//; exists n.
case; case=>p [->] H1.
- case/IH=>[[s1][s2][->][m H2] H3|[x1][x2][[m H2]] H3 H4].
  - left; exists s1; exists s2; do !split=>//.
    by exists m.+1; exists p.
  right; exists x1; exists x2; do !split=>//.
  by exists m.+1; exists p.
case/IH=>[[s1][s2][->] H2 [m H3]|[x1][x2][H2][m H3] H4].
- left; exists s1; exists s2; do !split=>//.
  by exists m.+1; exists p.
right; exists x1; exists x2; do !split=>//.
by exists m.+1; exists p.
Qed.

Lemma stutt_parPQ A B C (p1 p2 : tree A) (q1 q2 : tree B) (r : _ -> tree C) : 
        stutt p1 p2 -> stutt q1 q2 -> 
        stutt (Par p1 q1 r) (Par p2 q2 r).
Proof.
by move/(stutt_lparP q1 r)=>H1; move/(stutt_rparP p2 r); apply: stutt_trans.
Qed.

Add Parametric Morphism A B C : (@Par U A B C) with signature 
      @stutt _ ===> @stutt _ ===> @eq _ ===> @stutt _ as stutt_par_morhp'.
Proof. by move=>*; apply: stutt_parPQ. Qed.

(* Two stepping relations that may ignore [idle] transitions. The wstep *)
(* relation insists that the transition t is taken, but allows that a   *)
(* bunch of [idle] steps may be taken before or after t. The wwstep     *)
(* relations allows that t be ignored if t = [idle]. *) 

Inductive wstep A (p1 : tree A) (t : trans) (p2 : tree A) : Prop := 
  wstepI q1 q2 of stutt p1 q1 & step q1 t q2 & stutt q2 p2.

Inductive wwstep A (p1 : tree A) (t : trans) (p2 : tree A) : Prop := 
  wwstepI of wstep p1 t p2 \/ (t = [idle] /\ stutt p1 p2).

Lemma stw A (p1 p2 : tree A) t : step p1 t p2 -> wstep p1 t p2.
Proof. by move=>H; exists p1 p2; auto. Qed.

Lemma wstww A (p1 p2 : tree A) t : wstep p1 t p2 -> wwstep p1 t p2.
Proof. by move=>H; split; left. Qed.

Lemma stww A (p q : tree A) t : step p t q -> wwstep p t q.
Proof. by move/stw; move/wstww. Qed.

Lemma wst_sts A t (p q : tree A) : wstep p t q -> steps p q.
Proof.
case=>q1 q2; move/stutt_sts=>H1 H2; move/stutt_sts=>H3.
apply: sts_trans H3; apply: sts_trans H1 _.
by exists [:: t]; simpl; eauto.
Qed.

Lemma wwst_sts A t (p q : tree A) : wwstep p t q -> steps p q.
Proof. case=>[[|[_]]]; [apply: wst_sts | apply: stutt_sts]. Qed.

Hint Resolve stw wstww stww.

(* wstep extends with stuttering at front and back *)
Lemma wst_stuttw A (p p1 p2 : tree A) t :  
        wstep p1 t p2 -> stutt p p1 -> wstep p t p2.
Proof.
by case=>q1 q2 H1 H2 H3 H4; exists q1 q2; auto; apply: stutt_trans H1.
Qed.

Lemma wst_wstutt A (p p1 p2 : tree A) t : 
        wstep p1 t p2 -> stutt p2 p -> wstep p1 t p.
Proof.
by case=>q1 q2 H1 H2 H3 H4; exists q1 q2; auto; apply: stutt_trans H4.
Qed.

(* wstep by idle is stutter *)
Lemma wst_idle A (p1 p2 : tree A) : wstep p1 [idle] p2 -> stutt p1 p2.
Proof.
case=>q1 q2 H1 H2 H3; apply: stutt_trans H3; apply: stutt_trans H1 _.
by exists 1; simpl; eauto.
Qed. 

(* the same set of lemmas for wwstep *)
Lemma wwst_stuttw A (p p1 p2 : tree A) t : 
        wwstep p1 t p2 -> stutt p p1 -> wwstep p t p2.
Proof.
case=>[[H1 H2|[->] H1 H2]]; split. 
- by left; apply: wst_stuttw H2.
by right; split=>//; apply: stutt_trans H1.
Qed.

Lemma wwst_wstutt A (p p1 p2 : tree A) t : 
        wwstep p1 t p2 -> stutt p2 p -> wwstep p1 t p.
Proof.
case=>[[H1 H2|[->] H1 H2]]; split.
- by left; apply: wst_wstutt H2.
by right; split=>//; apply: stutt_trans H2.
Qed.

Lemma wwst_idleAll A (p1 p2 : tree A) t :
        wwstep p1 t p2 -> (forall ts p, stept ts p1 p -> all idleb ts) -> 
        t = [idle].
Proof.
case=>[[|[//]]]; case=>q1 q2 [n].
move/stutter_stt=>S1; move/stt1=>S2 _; move/(_ _ _ (stt_catI S1 S2)).
by rewrite all_cat /= andbC; case: idleP.
Qed.

(* if we have > 0 stutters, we can conclude wstep, not just wwstep *)
Lemma wwst_stutterw A n (p p1 p2 : tree A) t : 
        wwstep p1 t p2 -> stutter n.+1 p p1 -> wstep p t p2.
Proof.
case=>[[H1 H2|[->][m] H1 /= [r][H2] H3]].
- by apply: (wst_stuttw H1); exists n.+1.
by exists p r; auto; exists (n + m); rewrite stutter_add; eauto.
Qed.

Lemma wwst_wstutter A n (p p1 p2 : tree A) t :  
        wwstep p1 t p2 -> stutter n.+1 p2 p -> wstep p1 t p.
Proof.
case=>[[H1 H2|[->] H1 /= [r][H2] H3]].
- by apply: (wst_wstutt H1); exists n.+1.
by exists p2 r=>//; exists n.
Qed.

Lemma wwst_idle A (p1 p2 : tree A) : wwstep p1 [idle] p2 <-> stutt p1 p2.
Proof.
by split=>[[[H|[_ H]]] //|H]; [apply: wst_idle | split; auto].
Qed.

Lemma wst_stt A (p : tree A) t q : 
        wstep p t q <-> 
        exists m, exists n, stept (nseq m [idle] ++ t :: nseq n [idle]) p q.
Proof.
split; last first.
- case=>m [n]; case/stt_catE=>q1 [H1] /= [q2][H2] H3.
  by exists q1 q2=>//; apply/stutt_stt; [exists m | exists n].
case=>q1 q2; case/stutt_stt=>m H1 H2; case/stutt_stt=>n H3.
exists m; exists n; apply/stt_catE.
by exists q1; split=>//; exists q2.
Qed.

Lemma wwst_stt A (p : tree A) t q : 
        wwstep p t q <-> 
        (exists m, exists n, 
           stept (nseq m [idle] ++ t :: nseq n [idle]) p q) \/
        t = [idle] /\ exists n, stept (nseq n [idle]) p q. 
Proof.
split; last first.
- case=>[[m][n] H|[->][n] H].
  - by apply: wstww; apply/wst_stt; exists m; exists n.
  by split; right; split=>//; apply/stutt_stt; exists n.
case; case=>[|[->]].
- by case/wst_stt=>m [n] H; left; exists m; exists n. 
by case/stutt_stt=>n H; right; eauto.
Qed.

(* expressing the above two lemmas in terms of filter *)
(* ocassionally useful *)

(* first a few helper lemmas *)
Lemma filter_nseq A n f (x : A) : 
        filter f (nseq n x) = if f x then nseq n x else [::].
Proof.
elim:n f x=>[|n IH] f x /=; first by case: ifP.  
by case: ifP=>E; rewrite IH E.
Qed.

Lemma filter_split A f s (x : A) : 
        filter f s = [:: x] -> 
        exists s1, exists s2, 
          [/\ s = s1 ++ x :: s2, filter f s1 = [::] & filter f s2 = [::]]. 
Proof.
elim:s x=>[|y s IH] //= x.
case: ifP=>E; first by case=>-> E1; exists [::]; exists s.
by case/IH=>s1 [s2][->][E1] E2; exists (y::s1); exists s2; rewrite /= E.
Qed.

Prenex Implicits filter_split.

Lemma filter_idle ts : 
        filter (predC idleb) ts = [::] <-> ts = nseq (size ts) [idle].
Proof.
elim:ts=>[//|t ts IH] /=.
case: idleP=>E /=; last by split=>//; case.
rewrite IH E; split; last by case.
by move=>->; rewrite size_nseq.
Qed.

(* now the lemmas *)
Lemma wst_sttf A (p : tree A) t q : 
        wstep p t q <-> 
        exists ts, 
          [/\ size ts > 0, 
              filter (predC idleb) ts = filter (predC idleb) [:: t] &
              stept ts p q]. 
Proof.
split.
- case/wst_stt=>m [n] H; exists (nseq m [idle] ++ t :: nseq n [idle]).
  split=>//; first by case: m {H}.
  by rewrite filter_cat filter_nseq /= filter_nseq. 
case=>ts [H1] /= H2 H3; apply/wst_stt.
case: (idleP t) H2 H1 H3=>[->|_] /=; last first.
- by case/filter_split=>s1 [s2][->]; do 2![move/filter_idle=>->]; eauto.
move/filter_idle=>->; case: ts=>[//|t' ts] /= _ [r][H1] H2.
by exists 0; exists (size ts); exists r.
Qed.

Lemma wwst_sttf A (p : tree A) t q : 
        wwstep p t q <-> 
        exists ts, 
          filter (predC idleb) ts = filter (predC idleb) [:: t] /\ 
          stept ts p q.
Proof.
split.
- case; case=>[|[->]]; first by case/wst_sttf=>ts [H] H1 H2; eauto.
  by case/stutt_stt=>n H; exists (nseq n [idle]); rewrite filter_nseq.
case=>[ts][H1] H2.
case E: (size ts > 0); first by split; left; apply/wst_sttf; exists ts.
case: ts H1 H2 E=>[|s ts] /=; last by rewrite ltnS.
by case/trP: t=>//= _ -> _; apply/wwst_idle.
Qed.

(* we need some evaluation views as well *)

Lemma wst_dummyP A t (q : tree A) : wstep (Dummy A) t q <-> False.
Proof. by split=>// [[??]]; move/stutt_dummyP=>->; move/st_dummyP. Qed.

Lemma wwst_dummyP A t (q : tree A) : 
        wwstep (Dummy A) t q <-> t = [idle] /\ q = Dummy A.
Proof.
split=>[[]|[->->]]; last by split; auto.
by case; [move/wst_dummyP | rewrite stutt_dummyP].
Qed.

Lemma wst_retP A t v (q : tree A) : wstep (Ret v) t q <-> False.
Proof. by split=>// [[??]]; move/stutt_retP=>->; move/st_retP. Qed.

Lemma wwst_retP A t v (q : tree A) : 
        wwstep (Ret v) t q <-> t = [idle] /\ q = Ret v.
Proof.
split=>[[]|[->->]]; last by split; auto.
by case; [move/wst_retP | rewrite stutt_retP].
Qed.

Lemma wst_consP A B (a : action B) t p (q : tree A) g : 
        wstep (Cons a g p) t q <-> 
          [/\ B = Tr.typ t, jmeq action a (Tr.act t) & 
              forall pf, stutt (p (coerce (T:=id) (Tr.val t) pf)) q] \/
          [/\ B = unit, jmeq action a Idle &
              forall pf, wstep (p (coerce (T:=id) tt pf)) t q].
Proof.
split.
- move=>[q1 q2]; case/stutt_consP=>[->|[E]]; [case/st_consP=>E|]; 
  rewrite {B}E in a p *; move/jmE=>->{a}; move/(_ (erefl _))=>/=;
  [move=><- H; left | move=>H1 H2 H3; right]; split=>// pf; rewrite eqc //.
  by apply: wstepI H2 H3.
case=>[[E]|[E]]; rewrite {B}E in a p *; move/jmE=>->{a}; move/(_ (erefl _))=>/=.
- move=>H; exists (Cons (Tr.act t) g p) (p (Tr.val t)); auto.
  by rewrite {3}(tr_eta t); auto.
case=>q1 q2 [n H1] H2 H3; exists q1 q2=>//; exists n.+1; exists (p tt).
by split=>//; auto.
Qed.

Lemma wwst_consP A B (a : action B) t p (q : tree A) g : 
        wwstep (Cons a g p) t q <-> 
          [\/ [/\ B = Tr.typ t, jmeq action a (Tr.act t) & 
                  forall pf, stutt (p (coerce (T:=id) (Tr.val t) pf)) q], 
              [/\ B = unit, jmeq action a Idle &
                  forall pf, wstep (p (coerce (T:=id) tt pf)) t q] |
              t = [idle] /\ stutt (Cons a g p) q].
Proof.
split=>[[]|H]; last first.
by case: H; split; [left | left | right=>//]; apply/wst_consP; auto.
case; last by apply: Or33.
by case/wst_consP; [apply: Or31| apply: Or32].
Qed.

Lemma wst_consPQ A B (a : action B) (v : B) (p : B -> tree A) g : 
        wstep (Cons a g p) [tr a, v] (p v). 
Proof. by apply: stw; auto. Qed.

Lemma wwst_consPQ A B (a : action B) (v : B) (p : B -> tree A) g : 
        wwstep (Cons a g p) [tr a, v] (p v).
Proof. by apply: stww; auto. Qed. 

Hint Resolve wst_consPQ wwst_consPQ.

Lemma wst_lparP A B C p1 p2 q (r : A * B -> tree C) t : 
        wstep p1 t p2 -> wstep (Par p1 q r) t (Par p2 q r).
Proof.
case=>q1 q2; move/(stutt_lparP q r)=>H1.
move/(st_lparP q r)=>H2; move/(stutt_lparP q r)=>H3.
by exists (Par q1 q r) (Par q2 q r).
Qed.

Lemma wwst_lparP A B C p1 p2 q (r : A * B -> tree C) t :
        wwstep p1 t p2 -> wwstep (Par p1 q r) t (Par p2 q r).
Proof.
case; case=>[|[->] H]; first by move/(wst_lparP q r)=>H; apply: wstww.
by split; right; split=>//; apply: stutt_lparP H.
Qed.

Lemma wst_rparP A B C p q1 q2 (r : A * B -> tree C) t : 
        wstep q1 t q2 -> wstep (Par p q1 r) t (Par p q2 r).
Proof.
case=>r1 r2; move/(stutt_rparP p r)=>H1.
move/(st_rparP p r)=>H2; move/(stutt_rparP p r)=>H3.
by exists (Par p r1 r) (Par p r2 r).
Qed.

Lemma wwst_rparP A B C p q1 q2 (r : A * B -> tree C) t :
        wwstep q1 t q2 -> wwstep (Par p q1 r) t (Par p q2 r).
Proof.
case; case=>[|[->] H]; first by move/(wst_rparP p r)=>H; apply: wstww.
by split; right; split=>//; apply: stutt_rparP H.
Qed.

(**********)
(* active *)
(**********)

Definition active A (p : tree A) : bool := 
  if p is (Cons _ _ _ _ | Par _ _ _ _ _) then true else false.

Lemma active_parP A1 A2 C p1 p2 (p : A1 * A2 -> tree C) : active (Par p1 p2 p).
Proof. by []. Qed.

Lemma active_consP A C a (p : A -> tree C) g : active (Cons a g p).
Proof. by []. Qed.

Lemma st_actN A (p : tree A) t q : ~~ active p -> (step p t q <-> False).
Proof. 
by case:p t q=>[|a||] // t q _; [apply: st_dummyP | apply: st_retP]. 
Qed.

Implicit Arguments st_actN [A p t q].

Lemma st_actP A (p : tree A) t q : step p t q -> active p.
Proof. by case E: (active p)=>//; case/(st_actN (negbT E)). Qed.

Lemma stt_actN A ts (p q : tree A) : 
        ~~ active p -> stept ts p q -> ts = [::] /\ p = q.
Proof. by elim:ts=>[//|t ts IH] H [r][]; move/(st_actN H). Qed.

Lemma sts_actN A (p q : tree A) : ~~ active p -> steps p q -> p = q.
Proof. by move=>H [ts]; case/(stt_actN H). Qed. 

Lemma stutter_actN A n (p q : tree A) : 
        ~~ active p -> stutter n p q -> n = 0 /\ p = q.
Proof. by case:n=>[|n] // H; move/stutter_stt; case/(stt_actN H). Qed.

Lemma stutt_actN A (p q : tree A) : ~~ active p -> stutt p q -> p = q.
Proof. by move=>H; move/stutt_sts; apply: sts_actN. Qed.

Lemma wst_actN A (p : tree A) t q : 
        ~~ active p -> (wstep p t q <-> False).
Proof.
by case:p=>[|a||] // _; [apply: wst_dummyP | apply: wst_retP].
Qed.

Implicit Arguments wst_actN [A p t q].
 
Lemma wst_actP A (p : tree A) t q : wstep p t q -> active p.
Proof. by case E: (active p)=>//; move/(wst_actN (negbT E)). Qed.

Lemma wwst_actN A (p : tree A) t q : 
        ~~ active p -> wwstep p t q -> t = [idle] /\ p = q.
Proof.
by move=>H []; case=>[|[->]]; [move/(wst_actN H) | move/(stutt_actN H)].
Qed.

(* the following two lemmas may not be needed *)
(* but i proved them just for fun *)

(* stuttering is deterministic, in the following sense *)
Lemma stutt_idle A (p q : tree A) : 
        ~~ active q -> stutt p q -> 
        forall ts r, stept ts p r -> all idleb ts /\ stutt r q.
Proof.
elim:p q=>[B q H|B a q H|C B a g t IH1 q H|]. 
- by move/stutt_dummyP=>-> ts r; case/stt_dummyP=>-> ->.
- by move/stutt_retP=>-> ts r; case/stt_retP=>->->.
- case/stutt_consP=>[E|[E]]; first by rewrite E in H.
  rewrite {B}E in a t IH1 *; move/jmE=>->; move/(_ (erefl _))=>/= H1 ts r.
  move/stt_consP; case: ts=>[|t1 ts]; last by case=>[[]][->]; apply: IH1.
  move=>->; split=>//; apply/stutt_consP; right.
  by split=>//; move=>pf; rewrite eqc.
move=>C B1 B2 t IH1 q IH2 r1 IH3 r2 H1.
case/stutt_parP=>[[s1][s2][E]|[x1][x2][H2] H3 H4 ts r].
- by rewrite E in H1.
elim: ts t q IH1 IH2 H2 H3=>[|t1 ts IH] t q IH1 IH2 H2 H3 /=. 
- by move=><-; split=>//; apply/stutt_parP; right; exists x1; exists x2.
case=>r3 []; case/st_parP=>[|[->][y1][y2][E1][E2] ->] /=; last first.
- rewrite E1 E2 in H2 H3.
  case/stutt_retP: H2=><-; case/stutt_retP: H3=><-.
  by apply: (IH3 _ r2).
case; case=>p1 [->]; move/stt1=>H5.
- have L1: ~~ active (Ret x1) by [].
  case: (IH1 _ L1 H2 _ _ H5); rewrite /= andbT.
  case: idleP=>// E _; rewrite E /= in H5 *.
  case: H5=>x [H5] <-{p1} H6.
  apply: IH=>// q0 H0 S0 ts0 r0 H8.  
  have L2: stutt t q0 by apply: stutt_trans S0; exists 1; exists x.
  by apply: (IH1 q0 _ _ ([idle] :: ts0))=>//; exists x.
have L1: ~~ active (Ret x2) by [].
case: (IH2 _ L1 H3 _ _ H5); rewrite /= andbT.
case: idleP=>// E _; rewrite E /= in H5 *.
case: H5=>x [H5] <-{p1} H6.
apply: IH=>// q0 H0 S0 ts0 r0 H8.  
have L2: stutt q q0 by apply: stutt_trans S0; exists 1; exists x.
by apply: (IH2 q0 _ _ ([idle] :: ts0))=>//; exists x.
Qed.

Lemma stutt_inj A (p q1 q2 : tree A) : 
        ~~ active q1 -> ~~ active q2 ->
        stutt p q1 -> stutt p q2 -> q1 = q2. 
Proof.
elim:p q1 q2=>[B q1 q2 H1 H2|B a q1 q2 H1 H2|B C a g t IH q1 q2 H1 H2|].
- by do 2![move/stutt_dummyP=>->].
- by do 2![move/stutt_retP=>->].
- case/stutt_consP=>[E|[E]]; first by rewrite E in H1.
  rewrite {C}E in a t IH *; move/jmE=>->.
  move/(_ (erefl _))=>/= H3.
  case/stutt_consP=>[E|[_ _]]; first by rewrite E in H2.
  by move/(_ (erefl _))=>/=; apply: IH.
move=>B C D t IH1 q IH2 r IH3 p1 p2 H1 H2.
case/stutt_parP=>[[s1][s2][E]|[x1][x2][H3] H4 H5]; first by rewrite E in H1.
case/stutt_parP=>[[s1][s2][E]|[x3][x4][H6] H7 H8]; first by rewrite E in H2.
have [[E1][E2]] : Ret x1 = Ret x3 /\ Ret x2 = Ret x4.
- by split; [apply: IH1 | apply: IH2].
by rewrite -{x3}E1 -{x4}E2 in H6 H7 H8; apply: IH3 H5 H8.
Qed.

(*********************)
(* terminal equality *)
(*********************)

(* if one tree is terminal, the other tree is equal to it *)

Definition eqt A (p q : tree A) : Prop :=
  if active p && active q then True else p = q.

Notation "p =t q" := (eqt p q) (at level 70). 

Lemma eqt_refl A (p : tree A) : p =t p.
Proof. by rewrite /eqt; case: ifP. Qed.

Hint Resolve eqt_refl.

Lemma eqt_sym A (p q : tree A) : p =t q -> q =t p.
Proof. rewrite /eqt andbC; case:ifP=>//. Qed.

Lemma eqt_trans' A (p q r : tree A) : p =t q -> q =t r -> p =t r. 
Proof.
rewrite /eqt.
case H1: (active p); case H2: (active q); case H3: (active r)=>//=; 
try by apply: etrans.
- by move=>_ E; rewrite E H3 in H2.
by move=>E; rewrite -E H1 in H2.
Qed.

Definition eqt_trans A q p r := @eqt_trans' A p q r. 

Add Parametric Relation A : (tree A) (@eqt A)
  reflexivity proved by (@eqt_refl A)
  symmetry proved by (@eqt_sym A)
  transitivity proved by (@eqt_trans' A) as eqt_rel.

(********************************)
(* terminal stutter equivalence *)
(********************************)

(* if one tree is terminal, the other stutters to it *)
(* by stutt_idle, the other can only idle *)

Local Notation stutteqt1 := 
  (fun A p q => if ~~ active p then stutt q p else True).

Definition stutteqt A (p q : tree A) := stutteqt1 A p q /\ stutteqt1 A q p.

Notation "p ~~t q" := (stutteqt p q) (at level 70).

Lemma stutteqt_refl A (p : tree A) : p ~~t p.
Proof.
split; case: ifP=>// H; split=>//;
by case=>[//|t s r][q][]; move/(st_actN H).
Qed.

Hint Resolve stutteqt_refl.

Lemma stutteqt_sym A (p q : tree A) : p ~~t q -> q ~~t p.
Proof. by case. Qed.

Add Parametric Relation A : (tree A) (@stutteqt A) 
  reflexivity proved by (@stutteqt_refl A)
  symmetry proved by (@stutteqt_sym A) as stutteqt_rel.

(* this relation is not transitive, e.g. if p q r, and p is terminal *)
(* but q, r are not, then no relation between q and r exists *)

Lemma stutteqt_idle_consP A (p : tree A) : Idle @@ p ~~t p.
Proof.
split=>//; case: ifP=>// _.
by exists 1; exists p; split=>//; apply/st_consP.
Qed.

Hint Resolve stutteqt_idle_consP.

End Stepping.

Notation "p =t q" := (@eqt _ _ p q) (at level 70). 
Notation "p ~~t q" := (@stutteqt _ _ p q) (at level 70).

Hint Constructors step.
Hint Resolve st_consPQ st_retparP sts_refl stutt_refl stw wstww stww.
Hint Resolve wst_consPQ wwst_consPQ eqt_refl stutteqt_refl stutteqt_idle_consP.

Implicit Arguments sts_trans [U A p1 p3].
Implicit Arguments stutt_trans [U A p1 p3].
Implicit Arguments st_actN [U A p t q].
Implicit Arguments wst_actN [U A p t q].

Prenex Implicits filter_split.

