Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq rels ordtype.
Require Import Eqdep prelude domains.
Require Import resources8 proc9 step9 sbisim9.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(*********************)
(* Weak bisimulation *)
(*********************)

Definition wbisimulation U A (R : Pred (tree U A * tree U A)) : Prop :=
  forall p q, (p, q) \In R -> 
    p ~~t q /\
    (forall t p', step p t p' -> exists q', (p', q') \In R /\ wwstep q t q') /\
    (forall t q', step q t q' -> exists p', (p', q') \In R /\ wwstep p t p'). 

Inductive wbisim U A (p q : tree U A) : Prop := 
  Wbisim R of wbisimulation R & (p, q) \In R.

Notation "t1 ==~~ t2" := (wbisim t1 t2) (at level 50) : proc_scope.

(* ==~~ is an equivalence *)

Lemma wb_refl U A (p : tree U A) : p ==~~ p.
Proof.
exists [Pred p q | p = q]=>//; move=>{p} p q [->]. 
by split=>//; split=>t p' H; exists p'; split=>//; apply: stww.
Qed.

Hint Resolve wb_refl.

Lemma wb_sym U A (p q : tree U A) : p ==~~ q -> q ==~~ p.
Proof.
case=>R W H; exists [Pred p q | (q, p) \In R]=>{p q H} // p q.
by case/W=>S1 [S2 S3]; split=>//; apply: stutteqt_sym.
Qed.

Lemma wb_symE U A (p q : tree U A) : p ==~~ q <-> q ==~~ p.
Proof. by split; move/wb_sym. Qed.

Lemma wb_trans' U A (p1 p2 p3 : tree U A) : 
        p1 ==~~ p2 -> p2 ==~~ p3 -> p1 ==~~ p3.
Proof.
have wstutt: forall A R (p q : tree U A), wbisimulation R -> (p, q) \In R -> 
  (forall p', stutt p p' -> exists q', (p', q') \In R /\ stutt q q') /\
  (forall q', stutt q q' -> exists p', (p', q') \In R /\ stutt p p').
- move=>{A p1 p2 p3} A R p q W S; split=>p' [n];
  (elim: n p q p' S=>[|n IH] p q p' /=; first by move=>S <-; eauto);
  case/W=>T3 [T1 T2] [r][]; [case/T1 | case/T2]=>{T1 T2 T3} q' [];
  move/IH=>X; move/wwst_idle=>S; case/X=>r' [] H; 
  by move/(stutt_trans _ S); eauto.
have wstep: forall A R (p q : tree U A), wbisimulation R -> (p, q) \In R -> 
  (forall t p', wwstep p t p' -> exists q', (p', q') \In R /\ wwstep q t q') /\
  (forall t q', wwstep q t q' -> exists p', (p', q') \In R /\ wwstep p t p').
- move=>{A p1 p2 p3} A R p q W H; split=>t p' []; (case=>[[p1 r1]|[[->]]]; last first).
  - case/(wstutt _ _ _ _ W): H=>X _; case/X=>{X} q' [H1] H.
    by exists q'; rewrite wwst_idle.
  - case/(wstutt _ _ _ _ W): H=>X _; case/X=>{X} q1 [].
    case/W=>_ [X _] H1; case/X=>{X} s1 [].
    case/(wstutt _ _ _ _ W)=>X _ H2; case/X=>{X} q' [H H3].
    by exists q'; split=>//; apply: wwst_stuttw H1; apply: wwst_wstutt H3.
  - case/(wstutt _ _ _ _ W): H=>_ X; case/X=>{X} q' [H H1].
    by exists q'; rewrite wwst_idle.
  case/(wstutt _ _ _ _ W): H=>_ X; case/X=>{X} q1 [].
  case/W=>_ [_ X] H1; case/X=>{X} s1 [].
  case/(wstutt _ _ _ _ W)=>_ X H2; case/X=>{X} q' [H H3].
  by exists q'; split=>//; apply: wwst_stuttw H1; apply: wwst_wstutt H3.
have tstutt_actN : forall A (p1 p2 p3 : tree U A) R12 R23,
  ~~ active p1 ->
  wbisimulation R12 -> (p1, p2) \In R12 ->
  wbisimulation R23 -> (p2, p3) \In R23 ->
  stutt p3 p1.
- move=>{A p1 p2 p3} A p1 p2 p3 R12 R23 H W1 H12 W2 H23.
  - case/W1: H12=>[[]]; rewrite H=>[S2] _ _. 
    case/(wstutt _ _ _ _ W2): H23=>[]; case/(_ _ S2)=>p3' [H13] S3 _.
    case/W2: H13=>[[]]; rewrite H=>[S3'] _ _.
    by apply: stutt_trans S3 S3'.
case=>R12 W12 H12 [R23] W23 H23;
exists [Pred p q | exists r, (p, r) \In R12 /\ (r, q) \In R23]; 
  last by exists p2.
move=>{p1 p2 p3 H12 H23} p1 p3 [p2][H1] H2.
split; [|split=>t p1']. 
- split; case: ifP=>// H;
  [|move:(wb_sym (Wbisim W12 H1)) (wb_sym (Wbisim W23 H2))
  =>{R12 W12 H1 R23 W23 H2} [R12]W12 H2 [R23]W23 H1];
  by apply: (tstutt_actN _ _ _ _ _ _ H _ H1 _ H2).
- case/W12: H1=>_ [X _]; case/X=>{X} p2' [];
  case: (wstep _ _ _ _ W23 H2)=>X _ H1; case/X=>{X H2} q' [H H2];
  by exists q'; rewrite In_Simpl /=; eauto.
case/W23: H2=>_ [_ X]; case/X=>{X} p2' [];
case: (wstep _ _ _ _ W12 H1)=>_ X H2; case/X=>{X H1} q' [H H1];
by exists q'; rewrite In_Simpl /=; eauto.
Qed.

Definition wb_trans U A p2 p1 p3 := @wb_trans' U A p1 p2 p3.

Implicit Arguments wb_trans [U A].

Add Parametric Relation U A : (tree U A) (@wbisim U A)
  reflexivity proved by (@wb_refl _ _)
  symmetry proved by (@wb_sym _ _)
  transitivity proved by (@wb_trans' _ _ ) as wbisim_rel.

Definition wbfun U A B (f1 f2 : A -> tree U B) := 
  forall v : A, f1 v ==~~ f2 v.

Notation "f1 ==~~1 f2" := (wbfun f1 f2) 
  (at level 70, no associativity) : proc_scope.

(* ==~~1 is an equivalence *)

Lemma wbfun_refl U A B (f : A -> tree U B) : f ==~~1 f. 
Proof. by []. Qed.

Lemma wbfun_sym U A B (f1 f2 : A -> tree U B) : f1 ==~~1 f2 -> f2 ==~~1 f1.
Proof. by move=>H x; rewrite H. Qed.

Lemma wbfun_trans U A B (f1 f2 f3 : A -> tree U B) : 
        f1 ==~~1 f2 -> f2 ==~~1 f3 -> f1 ==~~1 f3.
Proof. by move=>H1 H2 x; rewrite H1 H2. Qed.

Add Parametric Relation U A B : (A -> tree U B) (@wbfun U A B) 
  reflexivity proved by (@wbfun_refl _ _ _)
  symmetry proved by (@wbfun_sym _ _ _)
  transitivity proved by (@wbfun_trans _ _ _) as wbfun_rel.

(* alternative formulation of ~~ *)
(* the proof is exactly the same as in the case of sbP *)
Lemma wbP U A (p q : tree U A) : 
        p ==~~ q <->
          p ~~t q /\
         (forall t p', step p t p' -> exists q', p' ==~~ q' /\ wwstep q t q') /\
         (forall t q', step q t q' -> exists p', q' ==~~ p' /\ wwstep p t p').
Proof.
split.
- by case=>R S; case/S=>T3 [T1 T2]; split=>//; split=>t p'; 
     [case/T1 | case/T2]=>q' []; exists q'; 
     split=>//; [|apply: wb_sym]; exists R.
pose P (p q : tree U A) := 
  p ~~t q /\
  (forall t p', step p t p' -> exists q', p' ==~~ q' /\ wwstep q t q') /\
  (forall t q', step q t q' -> exists p', q' ==~~ p' /\ wwstep p t p').
case=>T [H1 H2]; exists [Pred p q | P p q]=>{p q H1 H2 T} //.
move=>p q [T [H1 H2]]; split=>//; split=>t p'; 
[case/H1 | case/H2]=>q' []; [|move/wb_sym];
case=>R W; case/W=>T' [H1' H2'] S; exists q'; do 2![split=>//];
by split=>// t1 r1; [case/H1'|case/H2']=>r [S' H']; 
   exists r; split=>//; [|apply: wb_sym]; exists R.
Qed.

Lemma sbwb U A (p q : tree U A) : p ==~ q -> p ==~~ q.
Proof.
case=>R S H; exists R=>{p q H} // p q; case/S=>{S} [S1 S2 T]; 
split=>//; last first.
- split=>t p'; [case/S1 | case/S2]=>q' [S] H; exists q'; 
  by split=>//; apply: stww.
move: T; rewrite /stutteqt /eqt.
case P: (active p); first by case Q: (active q)=>// E; rewrite E Q in P.
move=><-; rewrite P /=; do !split=>//;
by move=>ts r; case/(stt_actN (negbT P))=>->.
Qed.

Lemma sbwb_cong U A (p1 p2 q1 q2 : tree U A) : 
        p1 ==~ p2 -> q1 ==~ q2 -> (p1 ==~~ q1 <-> p2 ==~~ q2).
Proof. by move=>H1 H2; rewrite (sbwb H1) (sbwb H2). Qed.

Add Parametric Morphism U A : (@wbisim U A) with signature 
    (@sbisim _ _) ===> (@sbisim _ _) ===> iff as sb_wbisim_morph.
Proof. by move=>*; apply: sbwb_cong. Qed.

(* the other direction should also hold here, but it's a pain to prove *)
Lemma wb_dummyP U A (p : tree U A) : p ==~~ Dummy -> stutt p Dummy.
Proof. by case/wbP=>[[]]. Qed.

Lemma wb_retP U A v (p : tree U A) : p ==~~ Ret v -> stutt p (Ret v). 
Proof. by case/wbP=>[[]]. Qed.

(*
Lemma wb_retparA A1 A2 A3 C v1 v2 p3 (r : (A1 * A2) * A3 -> tree C) :
        Par (Ret (v1, v2)) p3 r ==~~ 
        Par (Ret v1) (Par (Ret v2) p3 Ret) (r \o rA).
Proof.
admit.
Qed.
*)

Lemma wb_actN U A (p q : tree U A) : ~~ active q -> p ==~~ q -> stutt p q.
Proof. 
by case:q=>[_|a _||] //; [apply: wb_dummyP | apply: wb_retP]. 
Qed.
      
(*****************************)
(* Stepping and bisimulation *)
(*****************************)

(* step *)

Lemma stwP U A (p1 p2 q1 : tree U A) t : 
        p1 ==~~ p2 -> step p1 t q1 -> exists q2, q1 ==~~ q2 /\ wwstep p2 t q2.
Proof. by case/wbP=>_ [T] _; move/T. Qed.

Lemma sttwP U A ts (p1 p2 q1 : tree U A) : 
        p1 ==~~ p2 -> stept ts p1 q1 -> 
        exists q2, exists rs, 
          [/\ q1 ==~~ q2, 
              filter (predC idleb) ts = filter (predC idleb) rs & 
              stept rs p2 q2].
Proof.
elim:ts p1 p2 q1=>[|t ts IH] p1 p2 q1 /= H.
- by move=><-; exists p2; exists [::].
case=>r1 []; case/(stwP H)=>r2 [H1].
case/wwst_sttf=>rs /= [H2] H3.
case/(IH _ _ _ H1)=>q2 [ss] [H4] E H5.
exists q2; exists (rs ++ ss); split=>//; last by apply/stt_catE; eauto. 
by rewrite filter_cat H2 -E; case: idleP.
Qed.

Lemma stswP U A (p1 p2 q1 : tree U A) : 
        p1 ==~~ p2 -> steps p1 q1 -> exists q2, q1 ==~~ q2 /\ steps p2 q2.
Proof.
move=>H [ts]; case/(sttwP H)=>q2 [rs][E] _ S.
by exists q2; split=>//; exists rs.
Qed.

Lemma stuttwP U A (p1 p2 q1 : tree U A) : 
        p1 ==~~ p2 -> stutt p1 q1 -> exists q2, q1 ==~~ q2 /\ stutt p2 q2.
Proof.
move=>H [n]; elim: n p1 p2 q1 H=>[|n IH] p1 p2 q1 H /=.
- by move=><-; exists p2.
case=>[r][]; case/(stwP H)=>p1' [H1]; move/wwst_idle=>H2.
case/(IH _ _ _ H1)=>q2 [H3] H4; exists q2; split=>//.
by apply: stutt_trans H2 H4.
Qed.

Lemma wstwP U A (p1 p2 q1 : tree U A) t : 
        p1 ==~~ p2 -> wstep p1 t q1 -> exists q2, q1 ==~~ q2 /\ wwstep p2 t q2.
Proof.
move=>H; case/wst_sttf=>ts [_][S1] S2.
case/(sttwP H): S2 S1=>q2 [rs][H1] -> S2 S1; exists q2.
by rewrite wwst_sttf; eauto.
Qed.

Lemma wwstwP U A (p1 p2 q1 : tree U A) t : 
        p1 ==~~ p2 -> wwstep p1 t q1 -> exists q2, q1 ==~~ q2 /\ wwstep p2 t q2.
Proof.
move=>H; case/wwst_sttf=>ts [E] S.
case/(sttwP H): S E=>q2 [rs][H1] -> S E; exists q2.
by rewrite wwst_sttf; eauto.
Qed.

(* eventually, we probably need evaluation views for *)
(* all stepping relations and all constructors *)

(* but for now, just two *)

Lemma stw_dummyP U A (p q : tree U A) t : 
        p ==~~ Dummy -> step p t q -> t = [idle] /\ q ==~~ Dummy. 
Proof. by move=>H; case/(stwP H)=>r [H1]; case/wwst_dummyP=>-><-. Qed.

Lemma stw_retP U A v (p q : tree U A) t : 
        p ==~~ Ret v -> step p t q -> t = [idle] /\ q ==~~ Ret v.
Proof. by move=>H; case/(stwP H)=>r [H1]; case/wwst_retP=>-><-. Qed.

(* we should also have the following lemmas *)
(* about the non-active trees; notice that the non-active *)
(* tree appears as an ending object of stuttering, not the *)
(* initial one *)

Lemma wstutt_idle U A (p q1 q2 : tree U A) : 
        ~~ active q2 -> q1 ==~~ q2 -> stutt p q1 -> stutt p q2.
Proof. 
by move=>H; move/(wb_actN H)=>L1 L2; apply: stutt_trans L2 L1.
Qed.

(************************)
(* Algebraic properties *)
(************************)

(* Because ==~ implies ==~~, commutativity directly transfer to ==~~ *)
(* and need not be re-stated. *)

Lemma wb_comp U T A B C (r1 r2 : A * B -> tree U C) (f : T -> A * B) : 
        r1 ==~~1 r2 -> r1 \o f ==~~1 r2 \o f.
Proof. by move=>H x; apply: H. Qed.

Add Parametric Morphism U T A B C : (@funcomp (tree U C) (A * B) T tt) with signature 
      @wbfun U (A * B) C ===> @eq _ ===> @wbfun _ _ _ as wb_comp_morph.
Proof. by move=>r1 r2 H f; apply: wb_comp H. Qed. 

(* ==~~ mods out by idle actions *)
Lemma wb_idle_consP U A (p : unit -> tree U A) g : Cons Idle g p ==~~ p tt.
Proof.
apply/wbP; split; split=>//.
- by case: ifP=>// _; apply/stutt_consP; right; split=>// pf; rewrite eqc.
- case=>B a v p'; case/st_consP=>/= E; rewrite -{B}E in a v *; move/jmE=><- T.
  by case: v T=>T; exists (p tt); rewrite -(T (erefl _)) /= wwst_idle.
move=>t p' S; exists p'; split=>//; apply: wstww; exists (p tt) p'=>//.
by apply/stutt_consP; right; split=>// pf; rewrite eqc.
Qed.

(* the directions are not symmetric, so we provide the other one too *)
Lemma wb_idle_consPP U A (p : tree U A) : p ==~~ Idle @@ p.
Proof. by rewrite wb_idle_consP. Qed.

(* ==~~ is a congruence *)
Lemma wb_cons_cong U A B (a : action A) g1 g2 (p1 p2 : A -> tree U B) : 
         p1 ==~~1 p2 -> Cons a g1 p1 ==~~ Cons a g2 p2.
Proof.
move=>H; apply/wbP; split=>//; split=>t p'; 
case: t=>C c v; case/st_consP=>/= E;
rewrite -{C}E in c v *; move/jmE=><- T; [exists (p2 v) | exists (p1 v)];
by rewrite -(T (erefl _)) //= -(H v).
Qed.

Add Parametric Morphism U A B (a : action A) g : (Cons a g) with signature 
      @wbfun U A B ===> @wbisim U B as wb_cons_morph.
Proof. by move=>p1 p2 H; apply: wb_cons_cong. Qed.

(* the other direction also holds *)
(* so we can cancel actions *)
Lemma wb_consP U A B (a : action A) g1 g2 (p1 p2 : A -> tree U B) :
        Cons a g1 p1 ==~~ Cons a g2 p2 <-> p1 ==~~1 p2.
Proof.
split; last by apply: wb_cons_cong.
case/wbP=>T [T1 T2] v; apply/wbP; split.
- split; case: ifP=>// L; 
  [case: (T1 _ _ (st_consPQ a _ v g1))|case: (T2 _ _ (st_consPQ a _ v g2))]=>p [H1];
  (case/wwst_consP=>/=;
    [by case=>_ _; move/(_ (erefl _))=>/=; apply: wstutt_idle (wb_sym H1) | 
     case=>E; rewrite {A}E in a p1 p2 T T1 T2 v H1 L *;
     case: v L H1 => L H1; move/jmE=>->; move/(_ (erefl _)); move/wst_idle=>/=;
     by apply: wstutt_idle (wb_sym H1)|]);
  rewrite trD; case=>[[E]]; rewrite {A}E in a p1 p2 T T1 T2 v H1 L *;
  case: v L H1=> L H1; move/jmE=>->; move/(_ (erefl _))=>_ H2;
  (case/stutt_consP: H2 H1;
    [by move=>->; rewrite wb_idle_consP wb_symE => H; apply: wstutt_idle L H _|
     case=>_ _; move/(_ (erefl _))=>H1 H2;
     by apply: wstutt_idle (wb_sym H2) H1]).
split=>t p'; 
[case: (T1 _ _ (st_consPQ a _ v g1))|case: (T2 _ _ (st_consPQ a _ v g2))]=>p [H1];
(case/wwst_consP=>/=;
  [case=>_ _; move/(_ (erefl _))=>/= L; case/(stwP H1)=>q' [H2] W;
   by exists q'; split=>//; apply: wwst_stuttw W L | 
   case=>E; rewrite {A}E in a p1 p2 T1 T2 v T H1 *;
    case: v H1=>H1; move/jmE=>->; move/(_ (erefl _));
    move/wst_idle=>S1; case/(stwP H1)=>q' [S2] W;
    exists q'; split=>//; apply: wwst_stuttw W S1|]);
rewrite trD; case=>[[E]]; rewrite {A}E in a p1 p2 T1 T2 v T H1 *;
case: v H1=>H1; move/jmE=>->; move/(_ (erefl _))=>_;
(case/stutt_consP; 
  [move=>E; move: H1; rewrite {p}E wb_idle_consP=>H1; 
   by case/(stwP H1)=>q' [S2] W; exists q' |
   case=>_ _; move/(_ (erefl _))=>S1;
   case/(stwP H1)=>q' [S2] W; exists q'; split=>//;
   by apply: wwst_stuttw W S1]).
Qed.

(* even slightly more general, taking different types into account *)
(* the cancellation can be done, so long as at least one type is *)
(* inhabitted; otherwise, the two trees are deadlocked *)
(* so the trees are ==~~, but the actions are not equal *)
(* we also need to check that the actions are not Idle *)
(* if any one is Idle, it can be removed by a stutter *)
(* so it need not impose any restrictions on the other action *)
Lemma wb_consT U A1 A2 B (a1 : action A1) (a2 : action A2) g1 g2  
               (p1 p2 : _ -> tree U B) : 
        A1 + A2 -> ~~ Idleb a1 -> ~~ Idleb a2 ->
        Cons a1 g1 p1 ==~~ Cons a2 g2 p2 -> 
          [/\ A1 = A2, jmeq action a1 a2 & 
              forall pf v, p1 v ==~~ p2 (coerce (T:=id) v pf)].
Proof.
case=>v H1 H2; case/wbP=>T [T1 T2];
[case: (T1 _ _ (st_consPQ a1 _ v g1))|case: (T2 _ _ (st_consPQ a2 _ v g2))]=>p [H3];
case/wwst_consP=>/=.
- case=>E; rewrite {A2}E in a2 p2 T T1 T2 H1 H2 *; move/jmE=>E.
  rewrite {a2}E in T T1 T2 H1 H2 *=>S; split=>// pf w; rewrite eqc. 
  by move: w; apply/(wb_consP a1 g1 g2); apply/wbP.
- case=>E; rewrite {A2}E in a2 p2 T T1 T2 H1 H2 *; move/jmE=>E;
  by rewrite {a2}E in T T1 T2 H1 H2 *.
- rewrite trD; case=>[[E]]; rewrite {A1}E in a1 p1 v H1 H3 T T1 T2 *.
  by move/jmE=>E; rewrite E in H1.
- case=>E; rewrite {A1}E in a1 p1 H1 T T1 T2 *; move/jmE=>E.
  rewrite {a1 H1}E in T T1 T2 *=>L; split=>// pf w; rewrite eqc. 
  by move: w; apply/(wb_consP a2 g1 g2); apply/wbP. 
- case=>E; rewrite {A1}E in a1 p1 H1 T T1 T2 *; move/jmE=>E;
  by rewrite E in H1.
rewrite trD; case=>[[E]]; rewrite {A2}E in a2 p2 v H2 H3 T T1 T2 *.
by move/jmE=>E; rewrite E in H2.
Qed.
  
Lemma wb_par_cong U A B C p1 p2 q1 q2 (r1 r2 : A * B -> tree U C) : 
        p1 ==~~ p2 -> q1 ==~~ q2 -> r1 ==~~1 r2 -> 
        Par p1 q1 r1 ==~~ Par p2 q2 r2. 
Proof.
suff L: forall A B C p1 p2 q (r1 r2 : A * B -> tree U C), 
          p1 ==~~ p2 -> r1 ==~~1 r2 -> Par p1 q r1 ==~~ Par p2 q r2.
- move=>H1 H2 H3.
  apply: (wb_trans (Par p1 q2 r2)); last by apply: L.
  by rewrite 2!(sb_parC p1); apply: L H2 (wb_comp _ H3).
pose P A B C p1 p2 := 
  p1 ==~~ p2 \/ 
  exists s1, exists s2, exists r1 : A * B -> tree U C, exists r2, exists q, 
    p1 = Par s1 q r1 /\ p2 = Par s2 q r2 /\ s1 ==~~ s2 /\ r1 ==~~1 r2.
have L: forall A B C p1 p2 (r1 r2 : A * B -> tree U C) q t p1', 
  p1 ==~~ p2 -> r1 ==~~1 r2 -> step (Par p1 q r1) t p1' -> 
  exists p2', P A B C p1' p2' /\ wwstep (Par p2 q r2) t p2'.
- move=>{A B C p1 p2 q1 q2 r1 r2} A B C p1 p2 r1 r2 q t p1' H1 H2; rewrite /P.
  case/st_parP.
  - case; case=>q' [->]; last first.
    - by exists (Par p2 q' r2); split; [eauto 10 | eauto].
    case/(stwP H1)=>p2' [H4 H5]; exists (Par p2' q r2).
    by split; [eauto 10 | apply wwst_lparP].
  move=>H; case: H H1=>-> [x][y][->][->] ->; move/wb_sym.
  move=>H; exists (r2 (x, y)); split; first by left.
  apply: wstww; exists (Par (Ret x) (Ret y) r2) (r2 (x, y))=>//.
  - by apply: stutt_lparP; apply: wb_retP H.
  (* by apply: st_retpar. *)
move=>{A B C p1 p2 q1 q2 r1 r2} A B C p1 p2 q r1 r2 P1 R1. 
exists [Pred p1' p2' | P A B C p1' p2']; last by right; eauto 10.
move=>v1 v2 [|[s1][s2][t1][t2][q'][->][->][S1 T1] {v1 v2} /=]. 
- by case/wbP=>E [H1 H2]; split=>//; split=>t p'; [case/H1|case/H2]=>q' []; 
     exists q'; split=>//; left; [|apply: wb_sym].
split=>//; split=>t p'.
- by case/(L _ _ _ _ _ _ _ _ _ _ S1 T1)=>p; exists p.
case/(L _ _ _ _ _ _ _ _ _ _ (wb_sym S1) (wbfun_sym T1))=>p [H1 H2].
exists p; split=>//.
case: H1; first by move/wb_sym; left.
case=>s3 [s4][r3][r4][q2'][->][->][S2 R2]. 
right; exists s4; exists s3; exists r4; exists r3; exists q2'.
by do !split=>//; rewrite 1?S2 1?R2.
Qed.

Add Parametric Morphism U A B C: (@Par U A B C) with signature 
    @wbisim _ _ ===> @wbisim _ _ ===> 
    @wbfun _ _ _ ===> @wbisim _ _ as wb_par_morph.
Proof. by move=>*; apply: wb_par_cong. Qed.

(************************)
(* associativity of par *)
(************************)

(* will also need a lemma for dummy *)
(* and in general for inactive tail functions *)

Lemma wb_parCA U A1 A2 A3 C p1 p2 p3 (r : A1 * (A2 * A3) -> tree U C) :
        Par p1 (Par p2 p3 Ret) r ==~~ Par p2 (Par p1 p3 Ret) (r \o rCA).
Proof.
pose F A1 A2 A3 C p1 p2 := 
  [\/ p1 = p2, (* stutteq p1 p2 *)
      exists s1, exists s2, exists s3, 
      exists r : A1 * (A2 * A3) -> tree U C,
        p1 = Par s1 (Par s2 s3 Ret) r /\ 
        p2 = Par s2 (Par s1 s3 Ret) (r \o rCA),
      (* s2,s3 done *)
      exists s1, exists x2, exists x3, 
      exists r : A1 * (A2 * A3) -> tree U C,
        p1 = Par s1 (Ret (x2, x3)) r /\
        p2 = Par (Ret x2) (Par s1 (Ret x3) Ret) (r \o rCA),
      (* s1,s3 done *)
      exists x1, exists s2, exists x3,
      exists r : A1 * (A2 * A3) -> tree U C,
        p1 = Par (Ret x1) (Par s2 (Ret x3) Ret) r /\
        p2 = Par s2 (Ret (x1, x3)) (r \o rCA) |
      (* s1,s2,s3 done *)
      exists x1, exists x2, exists x3, 
      exists r : A1 * (A2 * A3) -> tree U C,
        p1 = Par (Ret x1) (Ret (x2, x3)) r /\
        p2 = Par (Ret x2) (Ret (x1, x3)) (r \o rCA)].
have Fsym: forall A1 A2 A3 C p1 p2,
  F A1 A2 A3 C p1 p2 -> F A2 A1 A3 C p2 p1.
- move=>{A1 A2 A3 C p1 p2 p3 r} A1 A2 A3 C p1 p2.
  case; first by [move=>->; apply: Or51];
  case=>s1 [s2][s3][r][->] ->;
  [apply: Or52 | apply: Or54 | apply: Or53 | apply: Or55];
  exists s2; exists s1; exists s3; exists (r \o rCA);
  by rewrite compA rCAI -compf1.
have L1: forall A1 A2 A3 C s1 s2 s3 (r : A1 * (A2 * A3) -> tree U C) t p',
  step (Par s1 (Par s2 s3 Ret) r) t p' ->
  exists q', wwstep (Par s2 (Par s1 s3 Ret) (r \o rCA)) t q' /\
             F A1 A2 A3 C p' q'.
- move=>{A1 A2 A3 C p1 p2 p3 r} A1 A2 A3 C s1 s2 s3 r t p'. 
  case/st_parP; last by case=>-> [x1][x23][->][].
  case=>[[s1'][->] H1|[s23][->]].   
  - exists (Par s2 (Par s1' s3 Ret) (r \o rCA)). 
    split; first by apply: stww; apply: st_rparP; apply: st_lparP H1. 
    by apply: Or52; exists s1'; exists s2; exists s3; exists r.
  case/st_parP; last first.
  - case=>-> [x2][x3][->][->] ->. 
    exists (Par (Ret x2) (Par s1 (Ret x3) Ret) (r \o rCA)). 
    split; first by split; right.
    by apply: Or53; exists s1; exists x2; exists x3; exists r.
  case=>[[s2'][->] H2|[s3'][->] H3].
  - exists (Par s2' (Par s1 s3 Ret) (r \o rCA)). 
    split; first by apply: stww; apply: st_lparP H2.
    by apply: Or52; exists s1; exists s2'; exists s3; exists r.
  exists (Par s2 (Par s1 s3' Ret) (r \o rCA)). 
  split; first by apply: stww; apply: st_rparP; apply: st_rparP H3.
  by apply: Or52; exists s1; exists s2; exists s3'; exists r.
have L2: forall A1 A2 A3 C s1 x2 x3 (r : A1 * (A2 * A3) -> tree U C) t p',
  step (Par s1 (Ret (x2, x3)) r) t p' ->
  exists q', wwstep (Par (Ret x2) (Par s1 (Ret x3) Ret)
                    (r \o rCA)) t q' /\ F A1 A2 A3 C p' q'.
- move=>{A1 A2 A3 C p1 p2 p3 r} A1 A2 A3 C s1 x2 x3 r t p'.
  case/st_parP=>[|[->][x1][_][->][[<-]] ->].
  - case=>[[s1'][->] H1|[s23][->]]; last by case/st_retP.
    exists (Par (Ret x2) (Par s1' (Ret x3) Ret) (r \o rCA)).
    split; first by apply: stww; apply: st_rparP; apply: st_lparP H1.
    by apply: Or53; exists s1'; exists x2; exists x3; exists r. 
  exists (r (x1, (x2, x3))); split; last by apply: Or51.
  split; right; split=>//; apply/stutt_parP; right.
  exists x2; exists (x1, x3); split=>//; apply/stutt_parP; right.
  by exists x1; exists x3.
have L3: forall A1 A2 A3 C x1 s2 x3 (r : A1 * (A2 * A3) -> tree U C) t p',
  step (Par (Ret x1) (Par s2 (Ret x3) Ret) r) t p' ->
  exists q', wwstep (Par s2 (Ret (x1, x3)) (r \o rCA)) t q' /\
                    F A1 A2 A3 C p' q'.
- move=>{A1 A2 A3 C p1 p2 p3 r} A1 A2 A3 C x1 s2 x3 r t p'.
  case/st_parP=>[|]; last by case=>->[_][?][[<-]][].
  case=>[[s2'][->]|[s13][->]]; first by case/st_retP.
  case/st_parP=>[|[->][x2][_][->][[<-]] ->]; last first.
  - exists (Par (Ret x2) (Ret (x1, x3)) (r \o rCA)).
    split; first by apply/wwst_idle.
    by apply: Or55; exists x1; exists x2; exists x3; exists r.
  case=>[[s1'][->] H1|[s3'][->]]; last by case/st_retP.
  exists (Par s1' (Ret (x1, x3)) (r \o rCA)).
  split; first by apply: stww; apply: st_lparP H1.
  by apply: Or54; exists x1; exists s1'; exists x3; exists r.
have L4: forall A1 A2 A3 C x1 x2 x3 (r : A1 * (A2 * A3) -> tree U C) t p',
  step (Par (Ret x1) (Ret (x2, x3)) r) t p' ->
  exists q', wwstep (Par (Ret x2) (Ret (x1, x3)) (r \o rCA)) t q' /\ 
             F A1 A2 A3 C p' q'.
- move=>{A1 A2 A3 C p1 p2 p3 r} A1 A2 A3 C x1 x2 x3 r t p'.
  case/st_parP=>[|[->][_][_][[<-]][[<-]] ->].
  - by case=>[[s1'][->]|[s23][->]]; move/st_retP.
  exists (r (x1, (x2, x3))). 
  by split; [apply: stww; apply: st_retpar | apply: Or51].
move:p1 p2 p3=>s1 s2 s3; exists [Pred p1 p2 | F A1 A2 A3 C p1 p2];
  last by apply: Or52; exists s1; exists s2; exists s3; exists r.
move=>{s1 s2 s3 r} p q. 
case.
- move=>->; split=>//; split=>t p' H;
  by exists p'; split; [apply: Or51 | apply: stww].
- case=>s1 [s2][s3][r][->] ->; split=>//; split=>t p' H; 
  case: (L1 _ _ _ _ _ _ _ _ _ _ H)=>q' [H1] H2; exists q'; 
  by [|rewrite compA rCAI -compf1 in H1; split=>//; apply: Fsym H2].
- case=>s1 [s2][s3][r][->] ->; split=>//; split=>t p' H.
  - by case: (L2 _ _ _ _ _ _ _ _ _ _ H)=>q' [H1] H2; exists q'.
  case: (L3 _ _ _ _ _ _ _ _ _ _ H)=>q' [H1] H2; exists q'.
  by rewrite compA rCAI -compf1 in H1; split=>//; apply: Fsym H2.
- case=>s1 [s2][s3][r][->] ->; split=>//; split=>t p' H.
  - by case: (L3 _ _ _ _ _ _ _ _ _ _ H)=>q' [H1] H2; exists q'.
  case: (L2 _ _ _ _ _ _ _ _ _ _ H)=>q' [H1] H2; exists q'.
  by rewrite compA rCAI -compf1 in H1; split=>//; apply: Fsym H2.
case=>s1 [s2][s3][r][->] ->; split=>//; split=>t p' H;
case: (L4 _ _ _ _ _ _ _ _ _ _ H)=>q' [H1] H2; exists q';
by [|rewrite compA rCAI -compf1 in H1; split=>//; apply: Fsym H2].
Qed.

Lemma wb_parA U A1 A2 A3 C p1 p2 p3 (r : (A1 * A2) * A3 -> tree U C) :
         Par (Par p1 p2 Ret) p3 r ==~~ Par p1 (Par p2 p3 Ret) (r \o rA).
Proof.
rewrite (sb_parC p2) par_compM sb_rparM wb_parCA sb_parC.
by apply: wb_par_cong=>//; case=>x [].
Qed.

Lemma wb_parAC U A1 A2 A3 C p1 p2 p3 (r : (A1 * A2) * A3 -> tree U C) :
         Par (Par p1 p2 Ret) p3 r ==~~ Par (Par p1 p3 Ret) p2 (r \o rAC).
Proof.
by rewrite sb_parC wb_parA wb_parCA !compA swap_rCA.
Qed.

(* because of composition r \o rAC, none of the associativiy lemmas
apply in inverse direction. this is not a problem for wb_parCA and
wb_parAC because they are idempotent, but it is a problem for
wb_parA. We need the inverse lemmas for wb_parA.  *)

Lemma wb_parAi U A1 A2 A3 C p1 p2 p3 (r : A1 * (A2 * A3) -> tree U C) :
        Par p1 (Par p2 p3 Ret) r ==~~ Par (Par p1 p2 Ret) p3 (r \o iA).
Proof.
rewrite wb_parCA sb_parC wb_parAC.
by apply: wb_par_cong=>//; case=>[[]]. 
Qed.

(******************************************)
(* parallel composition and concatenation *)
(******************************************)

Definition rlift A B B' (f : B -> option B') (x : A * B) : option (A * B') :=
  let: (x1,x2) := x in
  match f x2 with
  | None => None
  | Some x2' => Some (x1, x2')
  end.

Definition llift A A' B (f : A -> option A') (x : A * B) : option (A' * B) :=
  let: (x1,x2) := x in
  match f x1 with
  | None => None
  | Some x1' => Some (x1', x2)
  end.

(*
Lemma wb_rparC A B B' C p q (f : B' -> option B) (r : A * B -> tree C) :
        Par p (tcat q (to_tree1 f)) r
          ==~~ Par p q (fun x => tcat (to_tree (rlift f x)) r).
Proof.

pose F A B B' C q1 q2 := 
  [\/ q1 = q2,
      exists p, exists q, exists f : B' -> option B, exists r : A * B -> tree C,
        q1 = Par p (tcat q (to_tree1 f)) r
        /\ q2 = Par p q (fun x => tcat (to_tree (rlift f x)) r)
    | exists x, exists r : A * B -> tree C,
        q1 = Par (Ret x) Dummy r
        /\ q2 = Dummy].

move=>A B B' C p q f r.
exists [Pred q1 q2 | F A B B' C q1 q2]; last by apply: Or32; eauto 10.
move=>{p1 q f r} q1 q2 [].
- move=>->; split; first by apply: stutteqt_refl.
  by split=>// t p' H; (exists p'; split=>//; [apply: Or31|apply: stww]).
case=>p1 [q][f][r][->{q1}] ->{q2}; split=>//.
split.
  - move=>t p'; case/st_parP; last first.
    - case=>-> [x][y][->][]; case:q=>// y' E ->;
      exists (tcat (to_tree (rlift f (x,y'))) r).
      split; last by apply: stww; apply: st_retpar.
      by apply: Or31; move:E; rewrite /tcat /to_tree1 /rlift;
      case:(f y')=>//; move=>y0 [->] {y0}.
    case=>[[p2][->] H|[p2][->]].
    - exists (Par p2 q (fun x => tcat (to_tree (rlift f x)) r)).
      by split; [apply: Or32; eauto 10 | apply: stww; apply: st_lparP H].
    case/st_tcatP.
    - case=>q' [->] H; exists (Par p1 q' (fun x => tcat (to_tree (rlift f x)) r));
      by split; [apply: Or32; eauto 10 | apply: stww; apply: st_rparP].
    case=>x [_]; rewrite /to_tree1;
    by case:(f x); [move=>x'; move/st_retP|move/st_dummyP].

- move=>t p'; case/st_parP; last first.
  - case=>-> [x][y][->][]; case:q=>// y' E ->;
    exists (tcat (to_tree (rlift f (x,y'))) r).
    split
    - by apply: Or31; case:E=>-> {y'}.
    - by apply: stww; apply: st_retpar.

  rewrite /tcat /to_tree1 /rlift; case:(f y).
  - move=>y0; apply: stww; by apply: st_retpar.
  - 


  case=>[[p2][->] H|[p2][->]].
  - exists (Par p2 q (fun x => tcat (to_tree (rlift f x)) r)).
    by split; [apply: Or32; eauto 10 | apply: stww; apply: st_lparP H].
  case/st_tcatP.
  - case=>q' [->] H; exists (Par p1 q' (fun x => tcat (to_tree (rlift f x)) r));
    by split; [apply: Or32; eauto 10 | apply: stww; apply: st_rparP].
  case=>x [_]; rewrite /to_tree1;
  by case:(f x); [move=>x'; move/st_retP|move/st_dummyP].

move=>t p'; case/st_parP; last first.
- case=>-> [x][y][->][->]-> {p'}.

rewrite /tcat /rlift /to_tree1 /to_tree.
case:(f y);
first by move=>y'; exists (r (x, y')); split; [apply: Or31|apply: st_retpar].

exists Dummy.

=>//.

exists (r (Par p (tcat q (to_tree1 f)) r

exists (tcat (to_tree (rlift f (y, z))) r).
split.
  left.
  by apply: st_retpar.

case:p'=>//=.

 case: q=>//= x [<-] ->. 
  by exists (r (y, f x)); split; [left | apply: st_retpar].
case=>[[p2][->] H|[p2][->] H]. 
- exists (Par p2 (tmap f q) r).
  by split; [right; eauto 10 | apply: st_lparP H].
exists (Par p1 (tmap f p2) r).
by split; [right; eauto 10 | apply: st_rparP; apply: st_tmapPQ].
Abort.
*)
